<?php
/**
 * Stores device types and generates lists of properties, interfaces and actions for all device types.
 *
 * @author Wojciech Wichrowski
 */

require_once dirname(__FILE__) . '/../includes/config.php';
require_once dirname(__FILE__) . '/../includes/tools.php';
require_once dirname(__FILE__) . '/DeviceType.php';
require_once dirname(__FILE__) . '/DeviceInterface.php';
require_once dirname(__FILE__) . '/Property.php';
require_once dirname(__FILE__) . '/Action.php';

class DeviceHierarchy
{
    public $deviceTypes;
    public $deviceInterfaces;
    
    public function __construct()
    {
        $deviceTypes = array();
        $deviceInterfaces = array();
        
        if ($dh = opendir(TYPES_PATH))
        {
            while (($filename = readdir($dh)) !== false)
            {
                $ext = pathinfo($filename, PATHINFO_EXTENSION);
                if ($ext == 'xml' && is_file(TYPES_PATH . '/' . $filename))
                {
                    $path = TYPES_PATH . '/' . $filename;
                    $newType = $this->loadTypeFromFile($path);
                    if (!is_null($newType) && get_class($newType) == 'DeviceType')
                    {
                        $this->deviceTypes[$newType->getType()] = $newType;
                    }
                }
            }
            
            //hardcoded fix - will be fixed in devices hierarchy story
            $this->deviceTypes['remote_controller'] = $this->deviceTypes['com.fibaro.remoteController'];
            $this->deviceTypes['combo_device'] = $this->deviceTypes['com.fibaro.zwaveDevice'];
            
            
            closedir($dh);
        }
        
        if ($dh = opendir(INTERFACES_PATH))
        {
            while (($filename = readdir($dh)) !== false)
            {
                $ext = pathinfo($filename, PATHINFO_EXTENSION);
                if ($ext == 'xml' && is_file(INTERFACES_PATH . '/' . $filename))
                {
                    $path = INTERFACES_PATH . '/' . $filename;
                    $newInterface = $this->loadInterfaceFromFile($path);
                    if (!is_null($newInterface) && get_class($newInterface) == 'DeviceInterface')
                    {
                        $this->deviceInterfaces[$newInterface->getName()] = $newInterface;
                    }
                }
            }
            
            closedir($dh);
        }
        
    }
    
    private function loadTypeFromFile($path)
    {
        if (($xml = simplexml_load_file($path)) !== false)
        {
            $type = (string)$xml->attributes()->type;
            $baseType = (string)$xml->attributes()->baseType;
            $deviceType = new DeviceType($type, $baseType);
            
            foreach ($xml->interfaces->interface as $interface)
            {
                $deviceType->addInterface((string)$interface->attributes()->name);
            }
            
            foreach ($xml->properties->property as $property)
            {
                $propertyName = (string)$property->attributes()->name;
                $propertyType = (string)$property->attributes()->type;
                $defaultValue = NULL;
                if (!is_null($property->default))
                {
                    $defaultValue = (string)$property->default;
                }
                
                $deviceType->addProperty(new Property($propertyName, $propertyType, $defaultValue));
            }
            
            foreach ($xml->actions->action as $action)
            {
                $actionName = (string)$action->attributes()->name;
                $actionArgc = (int)$action->attributes()->argc;
                
                $deviceType->addAction(new Action($actionName, $actionArgc));
            }

            return $deviceType;
        }
    }
    
    private function loadInterfaceFromFile($path)
    {
        if (($xml = simplexml_load_file($path)) !== false)
        {
            $name = (string)$xml->attributes()->name;
            $deviceInterface = new DeviceInterface($name);
            
            foreach ($xml->properties->property as $property)
            {
                $propertyName = (string)$property->attributes()->name;
                $propertyType = (string)$property->attributes()->type;
                $defaultValue = NULL;
                if (!is_null($property->default))
                {
                    $defaultValue = (string)$property->default;
                }
                
                $deviceInterface->addProperty(new Property($propertyName, $propertyType, $defaultValue));
            }
            
            foreach ($xml->actions->action as $action)
            {
                $actionName = (string)$action->attributes()->name;
                $actionArgc = (int)$action->attributes()->argc;
                
                $deviceInterface->addAction(new Action($actionName, $actionArgc));
            }
            
            return $deviceInterface;
        }
        
    }
    
    public function getProperties($deviceType, $properties = array())
    {
        if (array_key_exists($deviceType, $this->deviceTypes) && get_class($this->deviceTypes[$deviceType]) == 'DeviceType')
        {
            if (count($this->deviceTypes[$deviceType]->getProperties()) > 0)
            {
                foreach ($this->deviceTypes[$deviceType]->getProperties() as $property)
                {
                    if ($property->valid)
                    {
                        $properties[$property->getName()] = $property->get(); //get COPY of the property
                    }
                }
            }
            
            if (count($this->deviceTypes[$deviceType]->getInterfaces()) > 0)
            {
                foreach ($this->deviceTypes[$deviceType]->getInterfaces() as $interface)
                {
                    $properties = $this->getInterfaceProperties($interface, $properties);
                }
            }
            
            $properties = $this->getProperties($this->deviceTypes[$deviceType]->getBaseType(), $properties);
        }
        
        return $properties;
    }
    
    public function getInterfaceProperties($deviceInterface, $properties = array())
    {
        if (array_key_exists($deviceInterface, $this->deviceInterfaces) && get_class($this->deviceInterfaces[$deviceInterface]) == 'DeviceInterface' && count($this->deviceInterfaces[$deviceInterface]->getProperties()) > 0)
        {
            foreach ($this->deviceInterfaces[$deviceInterface]->getProperties() as $property)
            {
                if ($property->valid)
                {
                    $properties[$property->getName()] = $property->get(); //get COPY of the property
                }
            }
        }
        
        return $properties;
    }
    
    public function getActions($deviceType, $actions = array())
    {
        if (array_key_exists($deviceType, $this->deviceTypes) && get_class($this->deviceTypes[$deviceType]) == 'DeviceType')
        {
            if (count($this->deviceTypes[$deviceType]->getActions()) > 0)
            {
                foreach ($this->deviceTypes[$deviceType]->getActions() as $action)
                {
                    if ($action->valid)
                    {
                        $actions[$action->getName()] = $action;
                    }
                }
            }
            
            if (count($this->deviceTypes[$deviceType]->getInterfaces()) > 0)
            {
                foreach ($this->deviceTypes[$deviceType]->getInterfaces() as $interface)
                {
                    $actions = $this->getInterfaceActions($interface, $actions);
                }
            }
            
            $actions = $this->getActions($this->deviceTypes[$deviceType]->getBaseType(), $actions);
        }
        
        return $actions;
    }
    
    public function getInterfaceActions($deviceInterface, $actions = array())
    {
        if (array_key_exists($deviceInterface, $this->deviceInterfaces) && get_class($this->deviceInterfaces[$deviceInterface]) == 'DeviceInterface' && count($this->deviceInterfaces[$deviceInterface]->getActions() > 0))
        {

            foreach ($this->deviceInterfaces[$deviceInterface]->getActions() as $action)
            {
                if ($action->valid)
                {
                    $actions[$action->getName()] = $action;
                }
            }
        }
        
        return $actions;
    }
    
    public function getInterfaces($deviceType, $interfaces = array())
    {
        if (array_key_exists($deviceType, $this->deviceTypes) && get_class($this->deviceTypes[$deviceType]) == 'DeviceType' && count($this->deviceTypes[$deviceType]->getInterfaces()) > 0)
        {
            foreach ($this->deviceTypes[$deviceType]->getInterfaces() as $interface)
            {
                if (!in_array($interface, $interfaces))
                {
                    $interfaces[] = $interface;
                }
            }
            
            $interfaces = $this->getInterfaces($this->deviceTypes[$deviceType]->getBaseType(), $interfaces);
        }
        
        return $interfaces;
    }
    
    public function getBaseType($deviceType)
    {
        if (is_string($deviceType) && array_key_exists($deviceType, $this->deviceTypes) && get_class($this->deviceTypes[$deviceType]) == 'DeviceType')
        {
            return $this->deviceTypes[$deviceType]->getBaseType();
        }
        
        return '';
    }
    
    public function get($deviceType)
    {
        if (array_key_exists($deviceType, $this->deviceTypes) && get_class($this->deviceTypes[$deviceType]) == 'DeviceType')
        {
            return $this->deviceTypes[$deviceType];
        }
        
        return $this->deviceTypes['com.fibaro.zwaveDevice'];
    }
    
    public function getChildTypes($parentType)
    {
        $childTypes = array();
        
        foreach ($this->deviceTypes as $deviceType)
        {
            if ($this->getBaseType($deviceType->getType()) == $parentType)
            {
                $childTypes[] = $deviceType->getType();
            }
        }
        
        return $childTypes;
    }
    
}
