--! includes
--! Device is the parent class for each plugin and must always be provided in the each new project.
require('common.device')
require('net.HTTPClient')
require('common.unitsConverter')

class 'WeatherProvider' (Device)

--! Initializes (constructor) WeatherProvider class plugin object.
--@param id: Id of the device.
function WeatherProvider:__init(id)
    Device.__init(self, id) -- Calls super class (Device) constructor.
    self.settingsInfo = api.get('/settings/info')
    self.settingsLocation = api.get('/settings/location')
    self.http = net.HTTPClient({ timeout = 40000 })  -- Creates an instance of HTTPClient class.
    if self.properties.selectedTemperatureSource == -1 then
        self:updateProperty('Temperature', 0)
    end
    if self.properties.selectedHumiditySource == -1 then
        self:updateProperty('Humidity', 0)
    end
    if self.properties.selectedWindSpeedSource == -1 then
        self:updateProperty('Wind', 0)
    end
    self:setCurentWeatherLabel()
    self:refreshState()
end

function WeatherProvider:getTemperatureSensorsList()
    local temperatureSensors = {}
    table.insert(temperatureSensors, {['value'] = -1, ['label'] = '---'})
    local devicesList = api.get('/devices')
    if devicesList and #devicesList > 0 then
        for _,device in pairs(devicesList or {}) do
            if type(device) == "table" and (device.baseType == "com.fibaro.temperatureSensor" or device.type == "com.fibaro.temperatureSensor" or device.baseType == "com.fibaro.weather" and device.type ~= "com.fibaro.WeatherProvider") then
                local sensorName = ''
                local id = ''
                for k2,v2 in pairs(device) do

                    if k2 == 'id' and v2 ~= '' then
                        id = v2
                    end
                    if k2 == 'name' then
                        sensorName = v2
                    end
                end
                if sensorName == '' then
                    sensorName = 'Device ID: ' .. id
                end
                if id ~= '' and sensorName ~= '' then
                    table.insert(temperatureSensors, {['value']=tonumber(id), ['label']=tostring(sensorName)})
                end
            end
        end
    end
    return temperatureSensors
end

function WeatherProvider:getHumiditySensorsList()
    local humiditySensors = {}
    table.insert(humiditySensors, {['value'] = -1, ['label'] = '---'})
    local devicesList = api.get('/devices')
    if devicesList and #devicesList > 0 then
        for _,device in pairs(devicesList or {}) do
            if type(device) == "table" and (device.baseType == "com.fibaro.humiditySensor" or device.type == "com.fibaro.humiditySensor" or device.baseType == "com.fibaro.weather" and device.type ~= "com.fibaro.WeatherProvider") then
                local sensorName = ''
                local id = ''
                for k2,v2 in pairs(device) do

                    if k2 == 'id' and v2 ~= '' then
                        id = v2
                    end
                    if k2 == 'name' then
                        sensorName = v2
                    end
                end
                if sensorName == '' then
                    sensorName = 'Device ID: ' .. id
                end
                if id ~= '' and sensorName ~= '' then
                    table.insert(humiditySensors, {['value']=tonumber(id), ['label']=tostring(sensorName)})
                end
            end
        end
    end
    return humiditySensors
end

function WeatherProvider:getWindSpeedSensorsList()
    local windSpeedSensors = {}
    table.insert(windSpeedSensors, {['value'] = -1, ['label'] = '---'})

    local devicesList = api.get('/devices')
    if devicesList and #devicesList > 0 then
        for _,device in pairs(devicesList or {}) do
            if type(device) == "table" and (device.baseType == "com.fibaro.windSensor" or device.type == "com.fibaro.windSensor" or device.baseType == "com.fibaro.weather" and device.type ~= "com.fibaro.WeatherProvider") then
                local sensorName = ''
                local id = ''
                for k2,v2 in pairs(device) do

                    if k2 == 'id' and v2 ~= '' then
                        id = v2
                    end
                    if k2 == 'name' then
                        sensorName = v2
                    end
                end
                if sensorName == '' then
                    sensorName = 'Device ID: ' .. id
                end
                if id ~= '' and sensorName ~= '' then
                    table.insert(windSpeedSensors, {['value']=tonumber(id), ['label']=tostring(sensorName)})
                end
            end
        end
    end
    return windSpeedSensors
end

--! Prepares HTTPClient object to do http request on specified target with request params.
--@param params: Table of options needed to invoke http request.
function WeatherProvider:httpRequest(params)
    self.http:request(params.url, {
        options = {
            method = params.method,
                },
                success = function(response)
                    if params.callback then
                        params.callback(response)
                    end
                end,
            error = function(error) self:onErorrFromRefreshState(error) end
        })
end

function WeatherProvider:refreshState()
    local params = {
        url = 'http://127.0.0.1:11111/api/refreshStates',
        callback =  function(response) self:onDataReceivedFromRefreshState(response) end,
        }
    if self.refreshStateLast ~= nil then
        params['url']='http://127.0.0.1:11111/api/refreshStates?last=' .. tostring(self.refreshStateLast)
    end
    self:httpRequest(params)
end

function WeatherProvider:onDataReceivedFromRefreshState(response)
    self.settingsInfo = api.get('/settings/info')
    self.settingsLocation = api.get('/settings/location')
    if tostring(type(response)) == 'table' and response.status and string.find(response.status, '200') then
        local responseJson = json.decode(response.data)
        self.refreshStateLast = responseJson.last
        if responseJson.changes then
            for i,v in pairs(responseJson.changes) do
                if v.id == self.properties.selectedTemperatureSource then
                    if v.value ~= nil then
                        self:updateTemperatureProperty(v.value)
                    end
                    if v.Temperature ~= nil then
                        self:updateTemperatureProperty(v.Temperature)
                    end
                end
                if v.id == self.properties.selectedHumiditySource then
                    if v.value ~= nil then
                        self:updateProperty('Humidity', tonumber(v.value))
                    end
                    if v.Humidity ~= nil then
                        self:updateProperty('Humidity', tonumber(v.Humidity))
                    end
                end
                if v.id == self.properties.selectedWindSpeedSource then
                    if v.value ~= nil then
                        self:updateWindProperty(v.value)
                    end
                    if v.Wind ~= nil then
                        self:updateWindProperty(v.Wind)
                    end
                end
            end
            self:setCurentWeatherLabel()
        end
    end
    self:refreshState()
end

function WeatherProvider:onErorrFromRefreshState(error)
    self:refreshState()
end

function WeatherProvider:updateTemperatureProperty(temp)
    if self.settingsLocation.temperatureUnit == "F" then
       temp = unitsConverter.farenheit2celsius(tonumber(temp))
    end
    self:updateProperty('Temperature', tonumber(temp))
end

function WeatherProvider:updateWindProperty(wind)
    if self.settingsLocation.windUnit == "mph" then
       wind = unitsConverter.mile2km(tonumber(wind))
    end
    self:updateProperty('Wind', tonumber(wind))
end

function WeatherProvider:round(num, idp)
    idp = idp or 1
    if idp~=0 and num%1~=0 then
        local mult = 10^idp
        r = math.floor(num * mult + 0.5) / mult
        if r%1==0 then
            return math.floor(r)
        else
            return r
        end
    else
        return math.floor(tonumber(num) + 0.5)
    end
end

function WeatherProvider:setCurentWeatherLabel()
    self.settingsLocation = api.get('/settings/location')
    local humidityValueLabel = ""
    local tempValueLabel = ""
    local windValueLabel = ""

    if self.properties.selectedTemperatureSource ~= -1 then
        if self.settingsLocation.temperatureUnit == "F" then
            tempValueLabel = tostring(self:round(unitsConverter.celsius2farenheit(self.properties.Temperature))) .. '°F'
        else
            tempValueLabel = tostring(self:round(self.properties.Temperature)) .. '°C'
        end
    else
        if self.settingsLocation.temperatureUnit == "F" then
            tempValueLabel = "---" .. '°F'
        else
            tempValueLabel = "---" .. '°C'
        end
    end

    if self.properties.selectedWindSpeedSource ~= -1 then
        if self.settingsLocation.windUnit == "mph" then
            windValueLabel = tostring(self:round(unitsConverter.km2mile(self.properties.Wind))) .. 'mph'
        else
            windValueLabel = tostring(self:round(self.properties.Wind)) .. 'km/h'
        end
    else
        if self.settingsLocation.windUnit == "mph" then
            windValueLabel = "---" .. 'mph'
        else
            windValueLabel = "---" .. 'km/h'
        end
    end

    if self.properties.selectedHumiditySource ~= -1 then
        humidityValueLabel = tostring(self:round(self.properties.Humidity)) .. "%"
    else
        humidityValueLabel = "---%"
    end

    if self.settingsInfo.defaultLanguage == 'pl' then
        self:updateProperty('ui.Current_Weather_Label.caption',  'Temp.: ' .. tempValueLabel .. '\n\nWilg.: '  .. humidityValueLabel .. '\n\nWiatr: '   .. windValueLabel)
    else
        self:updateProperty('ui.Current_Weather_Label.caption',  'Temp.: ' .. tempValueLabel .. '\n\nHumidity: '  .. humidityValueLabel .. '\n\nWind: '   .. windValueLabel)
    end
end
