require('common.device')
require('common.base64')
require('net.HTTPClient')
cosmo = require("cosmo.cosmo")
require('common.validator')
require('SyncQueue')

class 'AlphatechFarfisa' (Device)

--! Initializes AlphatechFarfisa plugin object.
--@param id: Id of the device.
function AlphatechFarfisa:__init(id)
	Device.__init(self, id)

	self.nextCommandTimeoutId = nil
	self.canExecuteCommandByLoop = true
	self.isPluginOnline = false
	self.availablePhoneBookEntries = 64
	self.savedKey = ''
	self.hcNetworkData = api.get('/settings/network')
	self.voipDevicesList = api.get('/voip')
	
	self.http = net.HTTPClient({ timeout = 20000 }) 

	self.url = {
		["openDoor1"] = '/cgi-bin/remctrl.sh?id=relay1',
		["openDoor2"] = '/cgi-bin/remctrl.sh?id=relay2',
		["configuration"] = '/cgi-bin/save_env.sh',
		["checkSavedKey"] = '/cgi-bin/sip.sh'
	}
	
	self:initSyncQueue()
	self:startPolling()
	self:updatePluginSettings()
end

--! Initialize process of syncQueue that is responsible for sync plugin data with device (AlphatechFarfisa).
function AlphatechFarfisa:initSyncQueue()
    self.syncQueue = SyncQueue()   
end

--! 2 seconds infinite polling loop for two purposes. 
--! 1) Tries execute command from syncQueue.
--! 2) Executes probing  method that checks if device (AlphatechFarfisa) is alive.
function AlphatechFarfisa:startPolling()
    if (self.nextCommandTimeoutId) then
  	 	clearTimeout(self.nextCommandTimeoutId)
	end

    self.nextCommandTimeoutId = setTimeout(function( ) self:startPolling(); end, 2000)

    self:isPluginAlive()
    self:executeCommandByLoopFromQueue()
end

--! Updates AlphatechFarfisa plugin.
function AlphatechFarfisa:updatePluginSettings()
	self.headers = {
		['Authorization'] = 'Basic '.. to_base64(self.properties.userName .. ':'.. self.properties.password),
		['Content-Type'] = 'application/x-www-form-urlencoded',
		['Accept']= 'text/html'
	}

	self:addVoipUserRowToJsonObjectForAlphatechFarfisaPhoneBook()
	self:runVoipUsersCreationFunctions()
end

--! Method responsible for cyclic check if plugin is online.
function AlphatechFarfisa:isPluginAlive()
	local params = {
			url = self.url.checkSavedKey,
			callback =  function(response) self:onDataReceivedFromIsPluginAlive(response) end,
		  } 

	self:getRequest(params, "isPluginAlive")
end

--! This function is responsible for managing data received from resources called from correspond method.
--!@param response: response body.
--!@param identifier: callback function identifier.
function AlphatechFarfisa:onDataReceivedFromIsPluginAlive(response, identifier)
	if (response.status and string.find(response.data, 'Registration successful')) then
		self.isPluginOnline = true
    else
		self.isPluginOnline = false
   	end	
end

--! Tries to run first command from queue if canExecuteCommandByLoop is true.
function AlphatechFarfisa:executeCommandByLoopFromQueue()
	if(self.syncQueue:getQueueTable() and self.canExecuteCommandByLoop) then
		local commandTalbe = self.syncQueue:getFrontElement()

		if commandTalbe  then
			for identifier,commdnFunction in pairs(commandTalbe) do
				commdnFunction(identifier)
			end	
		end
	end
end

--! Start point of run creation and injection of data for AlphatechFarfisa device
function AlphatechFarfisa:runVoipUsersCreationFunctions()
	self:getSavedKey("getSavedKey2", function() self:recreateViopUsersInAlphatechFarfisa("recreateViopUsersInAlphatechFarfisa") end )
	self:getSavedKey("getSavedKey", function() self:setAlphatechSipSettings("setAlphatechSipSettings") end )
end

--! Creates and manages http GET requests.
--!@param params - request options.
--!@param identifier - callback function identifier.
function AlphatechFarfisa:getRequest(params, identifier)
	if (identifier ~= 'openDoor1' and identifier ~= 'openDoor2' and identifier ~= 'isPluginAlive') then
		self.canExecuteCommandByLoop = false
	end

	local url = 'http://' .. self.properties.ip..params.url

	self.http:request(url, {
		options ={
			headers = self.headers,	
			method = 'GET',	
		},
		success = function(response)
			if (params.callback) then
				params.callback(response, identifier, params.afterGetKeyCallback)
			end
		end,
		error = function (error)
			if (params.callback) then
				params.callback(error, identifier)
			end
		end
	})
end

--! Creates and manages http POST requests.
--!@param params - request options.
--!@param identifier - callback function identifier.
function AlphatechFarfisa:postRequest(params, identifier)	
	self.canExecuteCommandByLoop = false
	local url = 'http://' .. self.properties.ip..params.url

	self.http:request(url, {	
		options = {
			headers = self.headers,
			method = 'POST',
			data = params.body,
		},
		success = function(response)
			if (params.callback) then
				params.callback(response, identifier)
			end
		end,
		error = function (error)
			if (params.callback) then
				params.callback(error, identifier)
			end
		end
	})
end

--! Method assigned to open door button1.
--! It is responsible to prepare appropriate params and send it to getRequest method.
function AlphatechFarfisa:openDoor1()
	local params = {
			url = self.url.openDoor1,
			callback =  function(response, identifier) self:onDataReceivedFromOpenDoor(response, identifier) end,
		  } 

	self:getRequest(params, "openDoor1")
end

--! Method assigned to open door button2.
--! It is responsible to prepare appropriate params and send it to getRequest method.
function AlphatechFarfisa:openDoor2()
	local params = {
			url = self.url.openDoor2,
			callback =  function(response, identifier) self:onDataReceivedFromOpenDoor(response, identifier) end,
		  } 

	self:getRequest(params, "openDoor2")
end

--! This function is responsible for managing data received from resources called from corresponded method. 
--!@param response: received data.
--!@param identifier: callback function identifier.
function AlphatechFarfisa:onDataReceivedFromOpenDoor(response, identifier)
	if (response.status == 200 and string.find(response.data, 'Relay'))then
		-- print('Succes STATUS: ', response.status)
		-- print('Succes DATA: ', response.data)
    else
  		-- print('*******************')
		-- print('* ERROR: ', response, response.status, response.data)
		-- print('*******************')
   	end	
end

--! Gets SAVED_KEY needet to POST correct SIP parameters.
--!@param identifier: callback function identifier.
--!@param afterGetKeyCallback: callback function called when SAVED_KEY has current (proper) value.
function AlphatechFarfisa:getSavedKey(identifier, afterGetKeyCallback)
	local params = {
			url = self.url.checkSavedKey,
			callback =  function(response, identifier, afterGetKeyCallback) self:onDataReceivedFromgetSavedKey(response, identifier, afterGetKeyCallback) end,
			afterGetKeyCallback = afterGetKeyCallback
		  } 

	local commandTable = {[identifier] = function(identifier)self:getRequest(params, identifier)end}		  
	self.syncQueue:pushBack(commandTable)	  
end

--! This function is responsible for managing data received from resources called from corresponded method. 
--!@param response: received data.
--!@param identifier: callback function identifier.
--!@param afterGetKeyCallback: callback function called when SAVED_KEY has current (proper) value.
function AlphatechFarfisa:onDataReceivedFromgetSavedKey(response, identifier, afterGetKeyCallback)
	if (response.status == 200 and string.find(response.data, 'SAVED_KEY'))then
		-- print('Succes STATUS: ', response.status)
		-- print('Succes DATA: ', response.data)

		self.savedKey = response.data:match("<input name='SAVED_KEY' value='(.-)' type='hidden'>")

	    if(self.syncQueue:getQueueTable()) then
			self.syncQueue:removeCommand(identifier)
			afterGetKeyCallback()
		end
    else
    	self.canExecuteCommandByLoop = true
  		-- print('*******************')
		-- print('* ERROR: ', response, response.status, response.data)
		-- print('*******************')
   	end	
end

--! Method responsible for create and inject suo data to Alphatech device.
--!@param  identifier: id for this method.
function AlphatechFarfisa:setAlphatechSipSettings(identifier)
	local voipSipAccount = 'ID=60&SAVED_KEY='..self.savedKey..'&SIP_SERVER_IP='..self.hcNetworkData.ip..'&SIP_SERVER_PORT=5060&SIP_REGISTRAR_IP='..self.hcNetworkData.ip..'&SIP_REGISTRAR_PORT=5060&SIP_ROUTE_IP=&SIP_ROUTE_PORT=5060&SIP_ACCOUNT_NAME='..self.properties.sipUserID..'&SIP_ACCOUNT_PASS='..self.properties.sipUserPassword..'&SIP_ACCOUNT_AUTHID=&SIP_ACCOUNT_EXPIRE=600&SIP_PROVIS_CODE=180'

	local params = {
				url = self.url.configuration,
				body = voipSipAccount,
				callback =  function(response, identifier) self:onDataReceivedFromSetAlphatechSipSettings(response, identifier) end,
			  } 

	self:postRequest(params, identifier)
end

--! This function is responsible for managing data received from resources called from corresponded method. 
--!@param response: received data.
function AlphatechFarfisa:onDataReceivedFromSetAlphatechSipSettings(response)
	self.canExecuteCommandByLoop = true
	if (response.status == 200) then
		-- print('Succes STATUS: ', response.status)
		-- print('Succes DATA: ', response.data)
    else
  		-- print('*******************')
		-- print('* ERROR: ', response, response.status, response.data)
		-- print('*******************')
   	end	
end

--! Method responsible for create and inject new data to AlphatechFarfisa device.
--!@param identifier: id of this method.
function AlphatechFarfisa:recreateViopUsersInAlphatechFarfisa(identifier)
	local buttonUsersList = 'SAVED_KEY='..self.savedKey..'&ID=7&'
	local voipUsersTable = self.properties.voipUsersForAlphatechFarfisaPhoneBook

	if (#voipUsersTable == 0 or voipUsersTable == nil) then
		self.canExecuteCommandByLoop = true
		return
	end

	for k,v in pairs(voipUsersTable) do
		for k2,v2 in pairs(v) do
			if (k2 == 'columns') then
				for k3,v3 in pairs(v2) do
					if (k3 == 1) then -- this iteration containg voipDevieceID						
						voipUserToInject = self:getVoipUserObjectByID(v3)	

						if (voipUserToInject == nil) then break end

					elseif (k3 == 3) then
						local code = tonumber(v3)
						voipUserToInject['keyCode'] = code  -- needed for mismatch with helios phoneBook start number (is 1 should be 0)
					end
				end

				if (voipUserToInject == nil) then break end

				buttonUsersList = buttonUsersList..'DAY_BUTTON'..voipUserToInject.keyCode..'='..voipUserToInject.sipUserID..'&NIGHT_BUTTON'..voipUserToInject.keyCode..'='..voipUserToInject.sipUserID..'&'
			end
		end
	end

	buttonUsersList = string.sub(buttonUsersList, 1, -2)  

	local params = {
				url = self.url.configuration,
				body = buttonUsersList,
				callback =  function(response, identifier) self:onDataReceivedFromRecreateViopUsersInAlphatechFarfisa(response, identifier) end,
			  } 

	self:postRequest(params, identifier)
end

--! This function is responsible for managing data received from resources called from corresponded method. 
--!@param response: received data.
function AlphatechFarfisa:onDataReceivedFromRecreateViopUsersInAlphatechFarfisa(response)
	self.canExecuteCommandByLoop= true
	if (response.status == 200) then
		-- print('Succes STATUS: ', response.status)
		-- print('Succes DATA: ', response.data)
    else
  		-- print('*******************')
		-- print('* ERROR: ', response, response.status, response.data)
		-- print('*******************')
   	end	
end

--! Set default properties for sip AlphatechFarfisa User.
function AlphatechFarfisa:setDefaultPropertyForSipAlphatechFarfisaUser()
	self:updateProperty('sipDisplayName', 'AlphatechFarfisa')
	self:updateProperty('sipUserID', '444444')
	self:updateProperty('sipUserPassword', '444444')
end

--! Helper method. Gets voip user name by ID from plugin voipUsers list.
--!@return name string or nil.
function AlphatechFarfisa:getVoipUserNameById(voipUserId)
	local sipDisplayName = ''
	local id = ''
	self.voipDevicesList = api.get('/voip')

	for k,v in pairs(self.voipDevicesList.voipDevices) do
		for k2,v2 in pairs(v) do
			if (k2 == 'sipDisplayName' and v2 ~= '' and v2 ~= self.properties.sipDisplayName) then
				sipDisplayName = v2
			end
			if (k2 == 'id' and tostring(v2) == voipUserId) then
				id = v2
			end
		end

		if (id ~= '' and sipDisplayName ~= '') then
			return sipDisplayName 
	    end
	end
				
	return nil
end

--! Helper method. Gets voip user by ID from HC2 api voip list.
--!@return voip user object or nil.
function AlphatechFarfisa:getVoipUserObjectByID(ID)
	self.voipDevicesList = api.get('/voip')
	for k,v in pairs(self.voipDevicesList.voipDevices) do
		for k2,v2 in pairs(v) do
			if (k2 == 'id' and tostring(v2) == tostring(ID)) then
				return v
			end
		end
	end
	
	return nil
end

--! Helper method. Checks if can add voip user.
--!@return true if can add voip user.
function AlphatechFarfisa:canAddVoipUser()
	local voipUsersTable = self.properties.voipUsersForAlphatechFarfisaPhoneBook
	if (self.properties.selectedVoipUser == '---' or
		self.properties.selectedKeyCode == '---' or 
		self.properties.selectedVoipUser == '' or 
		self.properties.selectedKeyCode == '' or
		self.properties.selectedVoipUser == 'null' or
		self.properties.selectedKeyCode == 'null') then

		return false
	end

	if (#voipUsersTable == 0 or voipUsersTable == nil) then return true end

	for k,v in pairs(voipUsersTable) do
		for k2,v2 in pairs(v) do
			if (k2 == 'columns') then
				for k3,v3 in pairs(v2) do
					if (v3 == self.properties.selectedKeyCode) then
						return false
					end
				end
			end	
		end
	end

	return true
end

--! Adds voip user to plugin property list.
function AlphatechFarfisa:addVoipUserRowToJsonObjectForAlphatechFarfisaPhoneBook()
	local rowToInsert = {}

	if (self:canAddVoipUser()) then
		local voipUsersTable = self.properties.voipUsersForAlphatechFarfisaPhoneBook
		rowToInsert = {["columns"]={self.properties.selectedVoipUser, self:getVoipUserNameById(self.properties.selectedVoipUser), self.properties.selectedKeyCode, "Delete"}}
		table.insert(voipUsersTable, rowToInsert)

		self:updateProperty('voipUsersForAlphatechFarfisaPhoneBook', voipUsersTable, true)
	end

	self:updateProperty('selectedVoipUser', '---')
	self:updateProperty('selectedKeyCode', '---')
end

--! Deletes voip user from plugin property list.
--!@return "delete" string to notify table control (connected with this method) of taken action.
function AlphatechFarfisa:deleteVoipUserRowFromJsonObjectForAlphatechFarfisaPhoneBook(tableRowId)	
	local voipUsersTable = self.properties.voipUsersForAlphatechFarfisaPhoneBook

	if (voipUsersTable == nil or voipUsersTable == '') then return end
	for k,v in pairs(voipUsersTable) do
		if ((k-1) == tableRowId) then
			table.remove(voipUsersTable, k)
		end
	end

	self:updateProperty('voipUsersForAlphatechFarfisaPhoneBook', voipUsersTable, true)

	return "delete"
end

--! Calculates voip users list and return it to select control that is connected with this method.
--!@return voip users list.
function AlphatechFarfisa:getVoipUsersList(arg)
	local usersList = {}
	table.insert(usersList, {["value"]='---', ["label"]='---'})
	local sipDisplayName = ''
	local id = ''
	self.voipDevicesList = api.get('/voip')

	for k,v in pairs(self.voipDevicesList.voipDevices) do
		for k2,v2 in pairs(v) do
			if (k2 == 'sipDisplayName' and v2 ~= '' and v2 ~= self.properties.sipDisplayName) then
				sipDisplayName = v2
			end
			if (k2 == 'id' and v2 ~= '') then
				id = v2
			end
		end

		if (id ~= '' and sipDisplayName ~= '') then
			table.insert(usersList, {["value"]=tostring(id), ["label"]=sipDisplayName})
	    end

		sipDisplayName = ''
		id = ''
	end

	return usersList
end

--! Calculate codes lists and return it to select control that is connected with this method.
--!@return codes list. 
function AlphatechFarfisa:getVoipCodesList(arg)
	local userCodes = {}
	table.insert(userCodes, {["value"]='---', ["label"]='---'})

	local iString

	for i=1,self.availablePhoneBookEntries do
		iString = tostring(i)
		table.insert(userCodes, {["value"] = iString, ["label"] = iString})
	end

	return userCodes
end

--!Temporary method to handle action method (open Lock) from mini plugin view;
function AlphatechFarfisa:openLock( ... )
	self:openDoor1()
end








