require('common.device')
require('commandQueue')
require('loopCommand')
require('testCommand')
require('verCommand')
require('eewrCommand')

require('common/TCPListener')

--! Class responsible for all communication with Davis weather station
--! Handles cyclical poling (LOOP command) 
class 'davis' (Device)


--! Initialises Davis object. Tries to open socket for communication
--!@param id device id
function davis:__init(id)
    Device.__init(self, id)

    --! Socket to connect with Davis weather station
    self.sock = nil
    --! Command that is currently executing
    self.currentCommand = nil
    --! Queue of commands
    self.commandQueue = nil
    --! Timeout, after which it is considered, that communication with Davis was disrupted. If executing command is still waiting for response, it is stopped
    self.commandTimeout = 5000
    --! Id of timeout handling canceling commands (after commandTimeout have been reached)
    self.clearCurrentCommandTimeoutId = nil
    --! Timeout after which it current polling status is cyclically checked 
    self.pollingTimeout = 1000
    --! Id of timeout for cyclical polling
    self.pollingTimeoutId = nil

    self.connectedTimeoutId = nil
    self.canSendCommand = false

    self.commandQueue = commandQueue()
    self.commandQueue:pushBack(loopCommand(self, 3))

    
    self:connectSocket()
    
    self:normalizeUI() 
end

--! Tries to connect to Davis station on socket
function davis:connectSocket()
    if (self.properties.ip ~= '' and self.properties.port ~= '' and 
        self.properties.ip ~= nil and self.properties.port ~= nil) then
        self.sock = TCPListener({delimiter = "\n\r", readTimeout = 10, connectTimeout=10})
        self.sock:registerEvent('dataReceived', function(sock, data) self:onDataReceived(data, id) end)
        self.sock:registerEvent('disconnected', function()  self:onClosed() end)
        self.sock:registerEvent('connected', function() self:onConnected() end)
        self.sock:registerEvent('error', function(sock, error) print("error", error) self.canSendCommand = false end)
        self.sock:connect(self.properties.ip, tonumber(self.properties.port))
    else
        print('No ip or port added')
    end
end

--! Sends message to socket
--!@param content message to be sent
function davis:send(content)
    self.sock:send(content)
end

--! Activates Davis console highlighting
function davis:turnConsoleHighlightOn()
    self:send("LAMPS 1\n")
end

--! Deactivates Davis console highlighting
function davis:turnConsoleHighlightOff()
    self:send("LAMPS 0\n")
end

--! Prints hex values of all charakters in passed parameter
--!@param data string that would be processed
function davis:printAsHex(data)
    local count = string.len(tostring(data))
    local i = 1
    local result = ""

    print("Got " .. count .. " bytes:")
    while count > 0 do
        local byte = string.byte(data, i)   

        result = result .. string.format("0x%x ", byte)    
        
        i = i+1
        count = count-1
    end
    print(result)
    print("")
end

--! Method handling receiving data from socket (from Davis). It is callback to socket object (self.sock)
--!@param data string with received data
function davis:onDataReceived(data)
    if self.currentCommand ~= nil then
        self.currentCommand:processResponse(data)
    end
end

--! Method invoked on succesfully connecting to Davis. It is callback to socket object (self.sock)
--! It also sends LOOP command to start polling Davis station for data and starts process of updating geo coordinates in Davis 
--!@param data string with received data
function davis:onConnected(data)
    self.canSendCommand = true
    self:poll()
    self:updateGeoCoordinates()
end

--! Method invoked on closing connection to Davis. It is callback to socket object (self.sock)
--! Tries to reconnect after 5 seconds
--!@param data string with received data
function davis:onClosed(data)
    self.canSendCommand = false

    if self.connectedTimeoutId then
        clearTimeout(self.connectedTimeoutId)
    end
    self.connectedTimeoutId = setTimeout(function() self:connectSocket() end, 5000)
end

--! Sends specified comand. Starts overlooking mechanism
--!@param command Command object
function davis:sendCommand(command)
    if self.canSendCommand then
        self.currentCommand = command
        self.currentCommand:send()
        self:currentCommandTimeout()
    end
end

--! Sends to socket <ACK> message
function davis:sendAck()
    self:send(string.char(0x06, 0x0a))
end

--! Sends to socket <NAK> message
function davis:sendNak()
    self:send(string.char(0x21, 0x0a))
end

--! Sends to socket <CANCEL> message
function davis:sendCancel()
    self:send(string.char(0x18, 0x0a))
end

--! Creates Test command and executes it
function davis:sendTest()
    self:sendCommand(testCommand(self))
end

--! Creates Ver command and executes it
function davis:sendVer()
    self:sendCommand(verCommand(self))
end

--! Creates Loop command and executes it
--!@param loopCount parameter to Loop command (how many data packet should be sent back)
function davis:sendLoop(loopCount)
    self:sendCommand(loopCommand(self, loopCount))
end

--! Refreshes timer on currentCommandTimeout
function davis:bumpCurrentCommand()
    clearTimeout(self.clearCurrentCommandTimeoutId)
    self.clearCurrentCommandTimeoutId = nil
    self:currentCommandTimeout()
end

--! Function overlooking command execution, so it does not wait more then self.commandTimeout seconds for response
function davis:currentCommandTimeout()
    self.clearCurrentCommandTimeoutId = setTimeout(function() self.currentCommand = nil; self.clearCurrentCommandTimeoutId=0; end, self.commandTimeout)
end

--! Resets self.currentCommand to nil
function davis:commandFinished()
    self.currentCommand = nil
end

--! Handles cyclical polling Davis station for data
function davis:poll()
    if self.pollingTimeoutId then
        clearTimeout(self.pollingTimeoutId)
    end


    self.pollingTimeoutId = setTimeout(function() self:poll(); end, self.pollingTimeout)
    if self.currentCommand == nil then
         local command = self.commandQueue:popFront()
         if command then
            self:sendCommand(command)
         end
     end
end

--! Changes longitude/latitude in davis station. Values are taken from api/settings/location
function davis:updateGeoCoordinates()
    local result = api.get("/settings/location") 
    local lon = result.longitude*10
    local lat = result.latitude*10
    lon = string.format("%04x", lon)
    lat = string.format("%04x", lat)
            
    local lon1 = string.upper(string.sub(lon, 1, 2))
    local lon2 = string.upper(string.sub(lon, 3))
    local lat1 = string.upper(string.sub(lat, 1, 2))
    local lat2 = string.upper(string.sub(lat, 3))

    self.commandQueue:pushBack(eewrCommand(self, "0B", lat2))
    self.commandQueue:pushBack(eewrCommand(self, "0C", lat1))
    self.commandQueue:pushBack(eewrCommand(self, "0D", lon2))
    self.commandQueue:pushBack(eewrCommand(self, "0E", lon1))            
end

--! Sets UI labels value to no value ("---")
function davis:normalizeUI()
    local value = "---"
    self:updateProperty("ui.barometerValueLabel.caption",  value)
    self:updateProperty("ui.humidityInsideValueLabel.caption", value)
    self:updateProperty("ui.humidityOutsideValueLabel.caption", value)
    self:updateProperty("ui.rainRateValueLabel.caption", value)
    self:updateProperty("ui.temperatureInsideValueLabel.caption", value)
    self:updateProperty("ui.temperatureOutsideValueLabel.caption", value)
    self:updateProperty("ui.windDirectionValueLabel.caption", value)
    self:updateProperty("ui.windStrengthValueLabel.caption", value)
    self:updateProperty("ui.sunriseValueLabel.caption", value)
    self:updateProperty("ui.sunsetValueLabel.caption", value)
    self:updateProperty("ui.soilTemperatureValueLabel.caption", value)
    self:updateProperty("ui.soilMoistureValueLabel.caption", value)
    self:updateProperty("ui.leafWetnessValueLabel.caption", value)
end

--! Inputs all weather values into UI properties
--!@param data weatherData object
function davis:updateUI(data)
    if (self.properties.units == '' or self.properties.units == nil) then self:updateProperty("units", "metric") end
    
    if (self.properties.units == "metric") then
        data:convertToMetric()
    end

    self:updateProperty("ui.barometerValueLabel.caption",  self:applyUnits(self:formatNumber(data.barometer), data.barometerUnit[self.properties.units]))
    self:updateProperty("ui.humidityInsideValueLabel.caption", self:applyUnits(data.insideHumidity, data.insideHumidityUnit[self.properties.units]))
    self:updateProperty("ui.humidityOutsideValueLabel.caption", self:applyUnits(data.outsideHumidity, data.outsideHumidityUnit[self.properties.units]))
    self:updateProperty("ui.rainRateValueLabel.caption", self:applyUnits(data.rainRate, data.rainRateUnit[self.properties.units]))
    self:updateProperty("ui.temperatureInsideValueLabel.caption", self:applyUnits(self:formatNumber(data.insideTemperature), data.insideTemperatureUnit[self.properties.units]))
    self:updateProperty("ui.temperatureOutsideValueLabel.caption", self:applyUnits(self:formatNumber(data.outsideTemperature) , data.outsideTemperatureUnit[self.properties.units]))
    self:updateProperty("ui.windDirectionValueLabel.caption", self:applyUnits(data.windDirection, data.windDirectionUnit[self.properties.units]))
    self:updateProperty("ui.windStrengthValueLabel.caption", self:applyUnits(self:formatNumber(data.windSpeed), data.windSpeedUnit[self.properties.units]))
    self:updateProperty("ui.sunriseValueLabel.caption", self:applyUnits(data.timeOfSunrise, data.timeOfSunriseUnit[self.properties.units]))
    self:updateProperty("ui.sunsetValueLabel.caption", self:applyUnits(data.timeOfSunset, data.timeOfSunsetUnit[self.properties.units]))

    local soilTemperature = "---";
    if (type(data.soilTemperatures) == 'table') then
        for k,v in pairs(data.soilTemperatures) do
            soilTemperature = v
            break
        end
    end
    self:updateProperty("ui.soilTemperatureValueLabel.caption", self:applyUnits(self:formatNumber(soilTemperature), data.soilTemperaturesUnit[self.properties.units]))

    local soilMoisture = "---";
    if (type(data.soilMoistures) == 'table') then
        for k,v in pairs(data.soilMoistures) do
            soilMoisture = v
            break
        end
    end
    self:updateProperty("ui.soilMoistureValueLabel.caption", self:applyUnits(soilMoisture, data.soilMoisturesUnit[self.properties.units]))

    local leafWetness = "---";
    if (type(data.leafWetnesses) == 'table') then
        for k,v in pairs(data.leafWetnesses) do
            leafWetness = v
            break
        end
    end

    self:updateProperty("ui.leafWetnessValueLabel.caption", self:applyUnits(leafWetness, data.leafWetnessesUnit[self.properties.units]))
end



--! If input is not nil, returns it with unit; else returns "---"
--!@param input value to be processed
--!@param unit measurement unit
--!@return "---" or value plus unit
function davis:applyUnits(input, unit)
    if (input == nil) then
        return "---"
    else
        return input .. unit
    end
end

--! Converts number to %.1f
--! If input is
function davis:formatNumber(input)
    if type(input) == "number" then
        return string.format("%.1f", input)
    else
        return nil
    end
end


