require('DenonHeosZone')
require('DenonHeosGroup')
require('net.UDPSocket')
require('common.device')
require('common.TCPListener')
require('ResponseHandler')
 

class 'DenonHeos' (Device) -- controller class
 

function DenonHeos:__init(id)
	Device.__init(self, id)	

	self.network = api.get('/settings/network')
	self.zones = {} -- the same as players in HEOS documentation nomenclature
	self.groups = {}
	self.port = 1255
	self.masterIp = nil
	self.delimiter = '\r\n'
	self.udpSocketTimeout = 4000
	self.responseHandler = ResponseHandler(self)
	self.closeTimeoutId = nil

	self:sync()
end

function DenonHeos:sync()
	self.retryCounter = 0 
	self:discover()
end

function DenonHeos:discover()
	self.retryCounter = self.retryCounter + 1

	self:updateProperty('ui.playersLabelInfo.caption', 'Players:\n---\n\nGroups:\n---')
	self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelDiscovering')

	local socket = net.UDPSocket({broadcast = true, timeout = self.udpSocketTimeout})
	local devicesTable = {}

	local data = table.concat({
		'M-SEARCH * HTTP/1.1',
	    'HOST: 239.255.255.250:1900',
        'MAN: "ssdp:discover"',
        'ST: urn:schemas-denon-com:device:ACT-Denon:1',
        'MX: 3',
        '',''},
        '\r\n'
        )
	socket:sendTo(data, "239.255.255.250", 1900,{
        success = function(data)
            self:discoverReadData(socket, devicesTable)
        end,
        error = function(error)
            print('Error:')
        end    
    })
end

-- read data from udp scocket 
-- on success read next data, on error call when all data is received - it means we can proceed
function DenonHeos:discoverReadData(socket, devicesTable)
	socket:receive({
        success = function(data)
            table.insert(devicesTable, data)
            self:discoverReadData(socket, devicesTable)
        end,
        error = function(error)
        	self:handleDiscoverResponse(devicesTable)
        end
    })
end

function DenonHeos:handleDiscoverResponse(devicesTable)
	local discoveredZonesDetails = {}

	for key, device in pairs(devicesTable) do
		if string.find(device,"schemas-denon-com:device:ACT-Denon:1", nil, true) then	
			local ip = device:match("http://(.*):%d*/")			
			table.insert(discoveredZonesDetails, ip)
		end
	end

	if self:countTableElements(discoveredZonesDetails) > 0 then
		self:updateCredentials(discoveredZonesDetails) 
		self:startSocket()
	else
		if self.retryCounter <= 5 then
			self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelDiscoveringRetry')
			setTimeout(function() self:discover() end, 3000)
		else
			self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelRetryExhausted')		
		end 
	end
end

function DenonHeos:updateCredentials(discoveredZonesDetails)
	self.masterIp = discoveredZonesDetails[1] -- need only one IP due to Heos instructions
end

function DenonHeos:startSocket()  
    self.sock = TCPListener({delimiter = string.char( 0x0D , 0x0A ), readTimeout = 10000, connectTimeout = 10000})
    self.sock:registerEvent('dataReceived', function(sock, data) self:onDataReceived(data) end)
    self.sock:registerEvent('disconnected', function()  self:onClosed() end)
    self.sock:registerEvent('connected', function() self:onConnected() end)
    self.sock:registerEvent('error', function(sock, error) print("error", error) end)
 
    self.sock:connect(self.masterIp, self.port)     
end

function DenonHeos:onConnected()
    self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelConnected')

    self:initializeHeosDriver()   
end

function DenonHeos:onDataReceived(data)
	self.responseHandler:handle(data)
end

function DenonHeos:onClosed()
    self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelConnectionError')

    if self.closeTimeoutID then
        clearTimeout(self.closeTimeoutID)
    end

    self.closeTimeoutID = setTimeout(function() self.sock:connect(self.masterIp, self.port) end, 3000)  
end

-- this procedure is written in Heos API documentation - 2.1.1 Driver Initialization
function DenonHeos:initializeHeosDriver() 
	self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelSynchronizing')

	self:unregisterForChangeEvents()
	setTimeout(function () self:getGroups() end, 15000) -- hack - we must wait 10 sec after socket connect to execute any of api command (if not the response will be incorrect) - warning - this hack not ensure the correct answer it only increase a success probability
    setTimeout(function () self:getZones() end, 17000)
    setTimeout(function () self:registerForChangeEvents() end, 18000)  
end

function DenonHeos:getZones()
	local command = 'heos://player/get_players' .. self.delimiter
	if self.sock then
		self.sock:send(command)
	end
end

function DenonHeos:getGroups()
	local command = 'heos://group/get_groups' .. self.delimiter
	if self.sock then
		self.sock:send(command)
	end
end

function DenonHeos:onDataReceivedFromGetZones(onlineZones)
	if type(onlineZones) == 'string' and onlineZones == 'error' then
		print('error on getZones')
	elseif type(onlineZones) == 'string' and onlineZones == 'empty' then
		print('no zones available in local network')
	else
		tprint(onlineZones)
		self.zones = self:getZonesFromExistingHC2Model() -- We doing this way in case if some one delete zone by cross button. Then we do not have notification (won't update internal model) but the model should be always updates if sync button is pressed.
		self:synchronizeZones(onlineZones, self.zones)
	end
end

function DenonHeos:onDataReceivedFromGetGroups(onlineGroups)
	if type(onlineGroups) == 'string' and onlineGroups == 'error' then
		print('error on getGroups')
	elseif type(onlineGroups) == 'string' and onlineGroups == 'empty' then
		print('no groups available in local network')
	else
		tprint(onlineGroups)
		self.groups = self:getGroupsFromExistingHC2Model() -- We doing this way in case if some one delete zone by cross button. Then we do not have notification (won't update internal model) but the model should be always updates if sync button is pressed.
		self:synchronizeGroups(onlineGroups, self.groups)
	end
end

function DenonHeos:onDataReceivedFromRegisterForChanges(data)
	if data.heos.message and data.heos.message ~= 'enable=off' then 
		self:updateSystemLabels()
	end
end

function DenonHeos:updateSystemLabels()
	self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelConnected')

	local labelText = ''

	if self:countTableElements(self.zones) > 0 then
		labelText = 'Players:\n'

		for _,zone in pairs(self.zones) do
			labelText = labelText .. zone.name ..'\n'
		end
	else
		labelText = 'Players:\n---\n'
	end

	if self:countTableElements(self.groups) > 0 then	

		labelText = labelText .. '\n\n Groups:\n'

		for _,group in pairs(self.groups) do
			labelText = labelText .. group.name .. '\n'
		end
	else
		labelText =  labelText .. '\n\nGroups:\n---'
	end

	self:updateProperty('ui.playersLabelInfo.caption', labelText)
end

function DenonHeos:synchronizeZones(onlineZones, hcZones)
	local onlineZonesVolume = self:countTableElements(onlineZones)
	local hcZonesVolume = self:countTableElements(hcZones)

	if onlineZonesVolume > 0 then
		for _,onlineZoneDeteils in pairs(onlineZones) do   
			local hcZone = self:getZoneByPidFromLocalZones(hcZones, onlineZoneDeteils.pid)       
        	if not hcZone then -- if onlne zone from LAN is not in local hcZones we will create it
    			self:createZoneDevice(onlineZoneDeteils)
    		end
    	end
    end

    if hcZonesVolume > 0 then 
    	for _,hcZone in pairs(hcZones) do   
			local cloudZoneDeteils = self:getZoneByPidFromOnlineZonesContainer(onlineZones, hcZone.properties.pid)       
        	if not cloudZoneDeteils then -- if zone from HC is not in online zones we will remove it
    			self:removeZoneDevice(hcZone) 
    		end
    	end
    end

    if onlineZonesVolume > 0 then
    	for _,onlineZoneDeteils in pairs(onlineZones) do   
			local hcZone = self:getZoneByPidFromLocalZones(hcZones, onlineZoneDeteils.pid)        
        	if hcZone then -- if zone from local LAN is in local zones we will update same local id zone with datra form online zone
    			hcZone:updateProperties(onlineZoneDeteils)
    		end
    	end
    end
end 

function DenonHeos:synchronizeGroups(onlineGroups, hcGroups)
	local onlineGroupsVolume = self:countTableElements(onlineGroups)
	local hcGroupsVolume = self:countTableElements(hcGroups)

	if onlineGroupsVolume > 0 then
		for _,onlineGroupDeteils in pairs(onlineGroups) do   
			local hcGroup = self:getGroupByGidFromLocalGroups(hcGroups, onlineGroupDeteils.gid)       
        	if not hcGroup then -- if onlne Group from LAN is not in local hcGroups we will create it
    			self:createGroupDevice(onlineGroupDeteils)
    		end
    	end
    end

    if hcGroupsVolume > 0 then 
    	for _,hcGroup in pairs(hcGroups) do   
			local cloudGroupDeteils = self:getGroupByGidFromOnlineGroupsContainer(onlineGroups, hcGroup.properties.gid)       
        	if not cloudGroupDeteils then -- if Group from HC is not in online Groups we will remove it
    			self:removeGroupDevice(hcGroup) 
    		end
    	end
    end

    if onlineGroupsVolume > 0 then
    	for _,onlineGroupDeteils in pairs(onlineGroups) do   
			local hcGroup = self:getGroupByGidFromLocalGroups(hcGroups, onlineGroupDeteils.gid)        
        	if hcGroup then -- if Group from local LAN is in local Groups we will update same local id Group with data form online Group
    			hcGroup:updateProperties(onlineGroupDeteils)
    		end
    	end
    end
end

function DenonHeos:getGroupByGidFromLocalGroups(hcGroups, gid)
	for _,groupDevice in pairs(hcGroups) do		
		if tostring(groupDevice.properties.gid) == tostring(gid) then
			return groupDevice
		end
	end
	
	return nil
end

function DenonHeos:getGroupByGidFromOnlineGroupsContainer(onlineGroups, gid)
	for _,groupDetails in pairs(onlineGroups) do
		if tostring(groupDetails.gid) == tostring(gid) then
			return groupDetails
		end
	end
	
	return nil
end

function DenonHeos:getZoneByPidFromLocalZones(hcZones, pid)
	if type(hcZones) == 'table' then
		for _,zoneDevice in pairs(hcZones) do
			if tostring(zoneDevice.properties.pid) == tostring(pid) then
				return zoneDevice
			end
		end
	end
	
	return nil
end

function DenonHeos:getZoneByPidFromOnlineZonesContainer(onlineZones, pid)
	if type(onlineZones) == 'table' then
		for _,zoneDetails in pairs(onlineZones) do
			if tostring(zoneDetails.pid) == tostring(pid) then
				return zoneDetails
			end
		end
	end
	
	return nil
end

function DenonHeos:createZoneDevice(onlineZoneDeteils)
	if onlineZoneDeteils.name and onlineZoneDeteils.pid then
		local zoneDevice = plugin.createChildDevice(self.id, 'com.fibaro.denonHeosZone', 'Player ' .. onlineZoneDeteils.name)
	    local zoneDeviceObject = DenonHeosZone(self, zoneDevice)
	    zoneDeviceObject:updateProperties(onlineZoneDeteils)
	    self.zones[zoneDevice.id] = zoneDeviceObject   
    end                     
end

function DenonHeos:removeZoneDevice(zone)
	self.zones[zone.id] = nil
    plugin.deleteDevice(zone.id)
end

function DenonHeos:createGroupDevice(onlineGroupDeteils)
	if onlineGroupDeteils.name and onlineGroupDeteils.gid then
		local groupDevice = plugin.createChildDevice(self.id, 'com.fibaro.denonHeosGroup', 'Group ' .. onlineGroupDeteils.name)
	    local groupDeviceObject = DenonHeosGroup(self, groupDevice)
	    groupDeviceObject:updateProperties(onlineGroupDeteils)
	    self.groups[groupDevice.id] = groupDeviceObject   
    end                     
end

function DenonHeos:removeGroupDevice(group)
	self.groups[group.id] = nil

    plugin.deleteDevice(group.id)
end

function DenonHeos:getZonesFromExistingHC2Model()
	local zones = {}

	local childDevices = plugin.getChildDevices(self.id)

    for _,dev in pairs(childDevices) do
        if (dev.type == 'com.fibaro.denonHeosZone') then
            zones[dev.id] = DenonHeosZone(self, dev)
        end
    end

    return zones
end

function DenonHeos:getGroupsFromExistingHC2Model()
	local groups = {}

	local childDevices = plugin.getChildDevices(self.id)

    for _,dev in pairs(childDevices) do
        if (dev.type == 'com.fibaro.denonHeosGroup') then
            groups[dev.id] = DenonHeosGroup(self, dev)
        end
    end

    return groups
end

function DenonHeos:unregisterForChangeEvents()
	if self.sock then
		local command = 'heos://system/register_for_change_events?enable=off' .. self.delimiter
		self.sock:send(command)		
	end
end 

function DenonHeos:registerForChangeEvents()
	if self.sock then
		local command = 'heos://system/register_for_change_events?enable=on' .. self.delimiter
		self.sock:send(command)		
	end
end

----------------------------------- HELPER METHODS 

function DenonHeos:countTableElements(table)
	local counter = 0

	if table then
		for _,_ in pairs(table) do
			counter = counter + 1
		end	
	end

	return counter
end

function tprint (tbl, indent)
	if type(tbl) == 'table' then
	  if not indent then indent = 0 end
	  for k, v in pairs(tbl) do
	    formatting = string.rep("  ", indent) .. k .. ": "
	    if type(v) == "table" then
	      print(formatting)
	      tprint(v, indent+1)
	    elseif type(v) == 'boolean' then
	      print(formatting .. tostring(v))      
	    else
	      print(formatting .. v)
	    end
	  end
	end
end






