require('common.device')
require('common/JsonRpcOverWebsocket')
require('common.Object')
require('IntercomEventsHandler')
require('net.HTTPClient')

class 'FibaroIntercom' (Device)

function FibaroIntercom:__init(id)
	Device.__init(self, id)

	logger.info("[FibaroIntercom:__init] plugin id: " .. id)
    self:resetProperties()
	
	logger.info("[FibaroIntercom:__init] create IntercomEventsHandler")
    self.intercomEventsHandler = IntercomEventsHandler(self)

	self.connectedTimeoutId = 0
	self.timeoutIdIsRecordingEnabled = 0
	self.timeoutIDsTable = {}

	self.notificationLabelLifeTime = 5000
	self.refreshTokenTimeInterval = 60000 -- 1 minute

	self.token = ''
	self:updateProperty('port', 8081)
	self.isConnected = false


	logger.info("[FibaroIntercom:__init] create net.HTTPClient")
	self.http = net.HTTPClient({ timeout = 10000 })

	logger.info("[FibaroIntercom:__init] create JsonRpcOverWebsocket")
    self.sock = JsonRpcOverWebsocket({callTimeout = 10000, secure = true}) -- secure - stets JsonRpcOverWebsocket to handle wss protocol
    self.sock:registerEvent('notificationFromServer', function(sock, data) self.intercomEventsHandler:handleEvents(data, id) end)
    self.sock:registerEvent('disconnected', function() 
			logger.error("JsonRpcOverWebsocket event disconnected")
			self:onClosed() 
		end)
    self.sock:registerEvent('connectionError', function()
			logger.error("JsonRpcOverWebsocket event connectionError")
			self:onClosed() 
		end)
    self.sock:registerEvent('connected', function() self:onConnected() end)

    self:connectSocket()
end

function FibaroIntercom:resetProperties()
	self:updateProperty("ui.notificationLabel.caption", "")
	self:updateProperty("ui.config.SwVersion.enabled", false)
	self:updateProperty("ui.config.SerialNumber.enabled", false)
	self:updateProperty('firstRelayIsOpen', false)
	self:updateProperty('secondRelayIsOpen', false)
	self:updateProperty('proxymityStateChanged', false)
	self:updateProperty('audioTriggerNoise', false)
    self:updateProperty('audioTriggerGlassBreak', false)
    self:updateProperty('audioTriggerSmoke', false)
    self:updateProperty('audioTriggerCarbonMonoxide', false)
    self:updateProperty('audioTriggerImpact', false)
    self:updateProperty('tamper', false)
end

function FibaroIntercom:connectSocket()
	logger.info("[FibaroIntercom:connectSocket] port:" .. tostring(self.properties.port) .. 
				" ip:" .. tostring(self.properties.ip) ..  
				" email:" .. tostring(self.properties.email) ..  
				" localPassword:" .. tostring(self.properties.localPassword))
	if self.properties.port == nil or
	   self.properties.ip == nil or
	   self.properties.ip == '' or
	   self.properties.email == nil or
	   self.properties.email == '' or
	   self.properties.localPassword == nil or
	   self.properties.localPassword == '' then

    	self:updateProperty('ui.statusLabel.caption', 'i18n_intercomStatusNotConfigured')
    	self.isConnected = false
    else
		local connectionString = 'wss://' .. self.properties.ip .. ':' .. self.properties.port .. '/wsock'
		logger.info("[FibaroIntercom:connectSocket] " .. connectionString)
    	self.sock:connect(connectionString)
    end
end

function FibaroIntercom:onConnected()
	self:login(self.properties.email, self.properties.localPassword)
end

function FibaroIntercom:getMjpg( ... )
	if self.properties.ip ~= nil and self.properties.ip ~= '' then
		self:updateProperty('ui.imageStream.source', 'http://' .. self.properties.ip .. ':' .. '8080' .. '/live/stream.jpg?token=' .. self.token)
		self:updateProperty('jpgPath', ':8080/live/image.jpg?token=' .. self.token)
		self:updateProperty('mjpgPath', ':8080/live/stream.jpg?token=' .. self.token)
	else
		self:updateProperty('ui.imageStream.source', 'novideo.png')
	end
end

function FibaroIntercom:login(user, pass)
	local methodName = 'com.fibaro.intercom.account.login'
	local params = {['user'] = user, ['pass'] = pass}
	local callback = function(response) self:onDataReceivedFromLogin(response) end

	self:sendRequest(methodName, params, callback)
end

function FibaroIntercom:onDataReceivedFromLogin(response)
	logger.info("[FibaroIntercom:onDataReceivedFromLogin] " .. json.encode(response))
	if response and response.result and response.result.token then
		self.token = response.result.token
	    self:initialMethodsAfterLogin()
	elseif response and response.error then
    	self:updateProperty('ui.statusLabel.caption', 'i18n_intercomStatusLoginError')
    	self.isConnected = false
	end
end

function FibaroIntercom:onClosed()
	self.isConnected = false
    self:updateProperty('ui.statusLabel.caption', 'i18n_intercomStatusConnectionError')

    if self.connectedTimeoutId then
        clearTimeout(self.connectedTimeoutId)
    end

	self.sock:close()
    self.connectedTimeoutId = setTimeout(function() self:connectSocket() end, 8000)
end

function FibaroIntercom:getRefreshToken()
	local methodName = 'com.fibaro.intercom.account.refreshToken'
	local params = {['token'] = self.token}
	local callback = function(response) self:onDataReceivedFromGetRefreshToken(response) end

	self:sendRequest(methodName, params, callback)

	if self.timeoutIDsTable then
		for _,timeoutId in pairs(self.timeoutIDsTable or {}) do
			clearTimeout(timeoutId)
		end
		self.timeoutIDsTable = {}
	end

	local id = setTimeout(function()self:getRefreshToken()end, self.refreshTokenTimeInterval)
	table.insert(self.timeoutIDsTable, id)
end

function FibaroIntercom:onDataReceivedFromGetRefreshToken(response)
	if response and response.result and response.result.token then
		self.token = response.result.token
		self:getMjpg()
	elseif response then
		logger.error("[FibaroIntercom:onDataReceivedFromGetRefreshToken] " .. json.encode(response))
		logger.info("[FibaroIntercom:onDataReceivedFromGetRefreshToken] reset connection")
		self:onClosed();
	end
end

function FibaroIntercom:initialMethodsAfterLogin()
	self:getRefreshToken()
	self:getMjpg()
	self:getAllUsers()
	self.isConnected = true
	self:isFTI()
	self:getSwVersion()
	self:getSerialNumber()
end

function FibaroIntercom:isFTI()
	local methodName = 'com.fibaro.intercom.device.isFTI'
	local params = {['token'] = self.token}
	local callback = function(response) self:onDataReceivedFromIsFti(response) end

	self:sendRequest(methodName, params, callback)
end

function FibaroIntercom:onDataReceivedFromIsFti(response)
	if response and response.result and response.result.state then
		self:updateProperty('ui.statusLabel.caption', 'i18n_intercomStatusFTI')
	else
		self:getMode()
	end
end

function FibaroIntercom:getAllUsers()
	local methodName = 'com.fibaro.intercom.account.users.getAll'
	local params = {['token'] = self.token}
	local callback = function(response) self:onDataReceivedFromGetAllUsers(response) end

	self:sendRequest(methodName, params, callback)
end

function FibaroIntercom:onDataReceivedFromGetAllUsers(response)
	local users = {}

	if response and response.result then
		for _,userTable in pairs(response.result or {}) do
			for k,v in pairs(userTable) do
				if k == 'user' then
					table.insert(users, v)
				end
			end
		end
	end

	self:updateProperty('users', json.encode(users))
end

function FibaroIntercom:getMode()
	local methodName = 'com.fibaro.intercom.device.getMode'
	local params = {['token'] = self.token}
	local callback = function(response) self:onDataReceivedFromGetMode(response) end

	self:sendRequest(methodName, params, callback)
end

function FibaroIntercom:getSwVersion()
	local methodName = 'com.fibaro.intercom.device.info.getSwVersion'
	local params = {['token'] = self.token}
	local callback = function(response)
		self:updateProperty("SwVersion", response.result.version, true)
	end

	self:sendRequest(methodName, params, callback)
end

function FibaroIntercom:getSerialNumber()
	local methodName = 'com.fibaro.intercom.device.info.getSerialNumber'
	local params = {['token'] = self.token}
	local callback = function(response)
		self:updateProperty("SerialNumber", response.result.serial, true)
	end

	self:sendRequest(methodName, params, callback)
end

function FibaroIntercom:onDataReceivedFromGetMode(response)
	if response and response.result and response.result.mode then
		self:updateProperty('currentMode', response.result.mode)
			self:adaptUiToCurrentMode(response.result.mode)

	end
end

function FibaroIntercom:adaptUiToCurrentMode(intercomMode)
	logger.info("[FibaroIntercom:adaptUiToCurrentMode] " .. intercomMode)
	if intercomMode == 'DOOR_STATION' then
		self:updateProperty('ui.statusLabel.caption', 'i18n_intercomStatusDoorStation')
	elseif intercomMode == 'NANNY' then
		self:updateProperty('ui.statusLabel.caption', 'i18n_intercomStatusNanny')
	elseif intercomMode == 'INTERCOM' then
		self:updateProperty('ui.statusLabel.caption', 'i18n_intercomStatusIntercom')
	end
end

function FibaroIntercom:sendRequest(methodName, params, callback)
    self.sock:sendJsonRpcCall(methodName, params,  {
        success = function(response)
            if callback then
                callback(response)
            end
         end,
        error = function (err)
			logger.error("[FibaroIntercom:sendRequest] error" .. tostring(err))
        	self:onClosed()
        end
    })
end

function FibaroIntercom:openDoor1()
	local methodName = 'com.fibaro.intercom.relay.open'
	local params = {['relay'] = 0, ['token'] = self.token}

	self:sendRequest(methodName, params)
end

function FibaroIntercom:openDoor2()
	local methodName = 'com.fibaro.intercom.relay.open'
	local params = {['relay'] = 1, ['token'] = self.token}

	self:sendRequest(methodName, params)
end

function FibaroIntercom:startStopRecord(event)
	if self.isConnected then
		if self.properties.isRecording then
			self:stopRecording()
		elseif not self.properties.isRecording then
			self:startRecording()
		end
	end
end

function FibaroIntercom:startRecording(event)
	self:isRecordingEnabled()

	local methodName = 'com.fibaro.intercom.recorder.start'
	local params = {['token'] = self.token}

	self:sendRequest(methodName, params)
	self:updateProperty('ui.statusLabel.caption', 'i18n_intercomStatusStartingRecording')
end

function FibaroIntercom:stopRecording(event)
	local methodName = 'com.fibaro.intercom.recorder.stop'
	local params = {['token'] = self.token}
	local callback =  function(response) 
		if response.error then
			logger.error('[FibaroIntercom:stopRecording] ' .. json.encode(response))
			logger.info('[FibaroIntercom:stopRecording] check recording status...')
			self:isRecording()
		end
	end

	self:sendRequest(methodName, params, callback)
	self:updateProperty('ui.statusLabel.caption', 'i18n_intercomStatusSavingRecording')
end

function FibaroIntercom:isRecording()
	local methodName = 'com.fibaro.intercom.recorder.isRecording'
	local params = {['token'] = self.token}
	local callback = function(response) 
		self:updateProperty('isRecording', response.in_progress == true) 
		self:isFTI()
	end

	self:sendRequest(methodName, params, callback)
end

function FibaroIntercom:isRecordingEnabled()
	local methodName = 'com.fibaro.intercom.recorder.isEnabled'
	local params = {['token'] = self.token}
	local callback = function(response) self:onDataReceivedFromIsRecordingEnabled(response) end

	self:sendRequest(methodName, params, callback)
end

function FibaroIntercom:onDataReceivedFromIsRecordingEnabled(response)
	if response and response.result then
		if not response.result.enabled then
			self:updateProperty('ui.statusLabel.caption', 'i18n_intercomStatusRecordingNotEnabled')
			if self.timeoutIdIsRecordingEnabled then
				clearTimeout(self.timeoutIdIsRecordingEnabled)
			end
			self.timeoutIdIsRecordingEnabled = setTimeout(function() self:getMode() end, self.notificationLabelLifeTime)
		end
	end
end

function FibaroIntercom:httpRequest(params)
    self.http:request(params.url, {
        options = {
            method = params.method,
            data = params.data,
            headers = params.headers
        },
        success = function(response)
            if params.callback then
                params.callback(response)
            end
        end,
        error = function (error)
            if params.callback then
                params.callback(error)
            end
        end
    })
end

function FibaroIntercom:openGate()
	self:openDoor2()
end

function FibaroIntercom:sendCameraImageToAdminEmail()
    api.post('/devices/' .. self.id .. '/action/sendPhotoToUser', {["args"]={2}} )
end

function FibaroIntercom:onSelectChange(selectValues)
	selectValues = json.decode(selectValues)

	local refreshEnabled = false

	if selectValues.value == '1' then
		refreshEnabled = true
	end

	self:updateProperty('ui.config.refreshTime.enabled', refreshEnabled)
end

function FibaroIntercom:getRefreshTimeValues()
	self:refreshRefreshSelectEnabled()

	local refreshValues = {
		{["value"]='1000', ["label"]='1'},
		{["value"]='2000', ["label"]='2'},
		{["value"]='3000', ["label"]='3'},
		{["value"]='4000', ["label"]='4'},
		{["value"]='5000', ["label"]='5'},
		{["value"]='6000', ["label"]='6'},
		{["value"]='7000', ["label"]='7'},
		{["value"]='8000', ["label"]='8'},
		{["value"]='9000', ["label"]='9'},
		{["value"]='10000', ["label"]='10'},
	}

	return refreshValues
end

function FibaroIntercom:refreshRefreshSelectEnabled()
	local refreshEnabled = false

	if self.properties.displayOnMainPage == 1 then
		refreshEnabled = true
	end

	self:updateProperty('ui.config.refreshTime.enabled', refreshEnabled)
end

function FibaroIntercom:forceUpdateProperty(name, value)
	self:updateProperty(name, '') -- hack for non ability to force change property in HC when property is the same
	os.execute("sleep " .. tonumber(1)) -- hack for non ability to force change property in HC when property is the same
	self:updateProperty(name, value)
end
