class 'IntercomEventsHandler'


function IntercomEventsHandler:__init(intercom)
    self.intercom = intercom
    self.notificationCenterUrl = 'http://127.0.0.1:11111/api/notificationCenter'
    self.publishEventUrl = 'http://127.0.0.1:11111/api/plugins/publishEvent'

    self.eventHandlers = {
        ['com.fibaro.intercom.audio.alarmDetected']              = function(params) self:intercomAudioTrigger(params) end, 
        ['com.fibaro.intercom.device.modeChanged']               = function(params) self:intercomModeChanged(params) end, 
        ['com.fibaro.intercom.relay.stateChanged']               = function(params) self:intercomRelayStateChanged(params) end, 
        ['com.fibaro.intercom.sensors.proximity.stateChanged']   = function(params) self:intercomSensorsProxymityStateChanged(params) end, 
        ['com.fibaro.intercom.sensors.tamper.stateChanged']      = function(params) self:intercomSensorsTamperStateChanged(params) end, 
        ['com.fibaro.intercom.audio.soundLevelChanged']          = function(params) self:intercomAudioLevelChanged(params) end, 
        ['com.fibaro.intercom.device.buttonStateChanged']        = function(params) self:intercomButtonStateChanged(params) end, 
        ['com.fibaro.intercom.device.group.completeChanged']     = function(params) self:intercomGroupCompleteChange(params) end, 
        ['com.fibaro.intercom.device.info.nameChanged']          = function(params) self:intercomNotifyIntercomNameChanged(params) end, 
        ['com.fibaro.intercom.account.accountChanged']           = function(params) self:intercomNotifyAccountChanged(params) end, 
        ['com.fibaro.intercom.recorder.recordings.stateChanged'] = function(params) self:intercomRecordingChanged(params) end, 
        ['com.fibaro.intercom.account.pinIncorrect']             = function(params) self:intercomPinIncorrect(params) end,
        ['com.fibaro.intercom.nonc.inputStateChanged']           = function(params) self:intercomInputStateChanged(params) end,
    } 
end 
 
function IntercomEventsHandler:handleEvents(eventJson)
    local ok, decodedEvent = pcall(json.decode, eventJson)
 
    if ok and decodedEvent.params and decodedEvent.method then
        if type(self.eventHandlers[decodedEvent.method]) == 'function' then 
			logger.info("[IntercomEventsHandler:handleEvents] handler:" .. decodedEvent.method)
            self.eventHandlers[decodedEvent.method](decodedEvent.params)    
        end
    end
end   

function IntercomEventsHandler:intercomAudioTrigger(params)
    if params.type == 'noise' then
        self.intercom:updateProperty('audioTriggerNoise', params.state)
        if params.state then
            self:sendNoiseNotification()
        end
    elseif params.type == 'glass-break' then
        self.intercom:updateProperty('audioTriggerGlassBreak', params.state)
        if params.state then
            self:sendGlassBreakNotification()
        end
    elseif params.type == 'smoke' then
        self.intercom:updateProperty('audioTriggerSmoke', params.state)
        if params.state then
            self:sendSmokeNotification()
        end    
    elseif params.type == 'carbon-monoxide' then
        self.intercom:updateProperty('audioTriggerCarbonMonoxide', params.state)
        if params.state then
            self:sendCarbonMonoxideNotification()
        end    
    elseif params.type == 'impact' then
        self.intercom:updateProperty('audioTriggerImpact', params.state)
        if params.state then
            self:sendImpactNotification()
        end
    end
end

function IntercomEventsHandler:intercomModeChanged(params)
    logger.info("[IntercomEventsHandler:intercomModeChanged] mode:" .. params.mode)
    self.intercom:updateProperty('currentMode', params.mode)
    self.intercom:adaptUiToCurrentMode(params.mode)
    self:sendModeChangedNotification(params.mode)
end

function IntercomEventsHandler:intercomRelayStateChanged(params)
    logger.info("[IntercomEventsHandler:intercomRelayStateChanged] " .. json.encode(params))
    if params.is_open then
        self.intercom:forceUpdateProperty('lastRelayStateChangedMethod', params.method)

        if params.method ~= 'TIMEOUT' and params.method ~= 'NO/NC' then
            self.intercom:forceUpdateProperty('recentlyCame', params.user)
        end
    end
    
    if params.relay == 0 then
        self.intercom:updateProperty('firstRelayIsOpen', params.is_open)
    elseif params.relay == 1 then
        self.intercom:updateProperty('secondRelayIsOpen', params.is_open)
    end
end

function IntercomEventsHandler:intercomSensorsProxymityStateChanged(params)
    logger.info("[IntercomEventsHandler:intercomSensorsProxymityStateChanged] state:" .. tostring(params.state))
    self.intercom:updateProperty('proxymityStateChanged', params.state)
end

function IntercomEventsHandler:intercomSensorsTamperStateChanged(params)
    if self.intercom.properties.tamper ~= params.state then
        logger.info("[IntercomEventsHandler:intercomSensorsTamperStateChanged] tamper changed:" .. tostring(params.state))
        self.intercom:updateProperty('tamper', params.state)
        if params.state then
            self.intercom:httpRequest(self:createNotificationParams('i18n_notificationTamperStatusChanged'))
        end
    end
end

function IntercomEventsHandler:intercomAudioLevelChanged(params)
    self.intercom:updateProperty('dBAudioLevelChanged', tonumber(params.level))
end

function IntercomEventsHandler:intercomButtonStateChanged(params)
    logger.info("[IntercomEventsHandler:intercomButtonStateChanged] button:" .. params.button .. " state:" .. tostring(params.state))
    if params.button == 1 then  -- ring
        self.intercom:updateProperty('buttonOneIsPressed', params.state)
    elseif params.button == 2 then -- call button
        self.intercom:updateProperty('buttonTwoIsPressed', params.state)
        if params.state then
            logger.info("[IntercomEventsHandler:intercomButtonStateChanged] incoming call")
            self.intercom:httpRequest(self:createNotificationParams('i18n_notificationButtonPressed'))
            self.intercom:httpRequest(self:createIncomingCallEvent())
        end
    elseif params.button == 3 then -- cancel button
        self.intercom:updateProperty('buttonThreeIsPressed', params.state)
    end
end

function IntercomEventsHandler:intercomGroupCompleteChange(params)
    self:sendGroupCompleteChangeNotification()
end

function IntercomEventsHandler:intercomNotifyIntercomNameChanged(params)
    self:sendIntercomNameChangedNotification(params)
end

function IntercomEventsHandler:intercomNotifyAccountChanged(params)
    self.intercom:getAllUsers()
    self:sendIntercomAccountChangedNotification(params)
end

function IntercomEventsHandler:intercomRecordingChanged(params)
    local isRecording = false
    local userRecording = false

    if params and params.event_type == 'on_demand' then
        userRecording = true
    end

    if params and params.state == 'recording' then
        isRecording = true
    end

    logger.info("[IntercomEventsHandler:intercomRecordingChanged] isRecording:" .. tostring(isRecording) .. " userRecording:" .. tostring(userRecording))
    if userRecording then
        self.intercom:updateProperty('isRecording', isRecording)  
        self.intercom:isFTI()
    end
end

function IntercomEventsHandler:intercomPinIncorrect(params)
    local timestamp = os.time()
    self.intercom:updateProperty('incorrectPinLastTyped', tonumber(timestamp))
end

function IntercomEventsHandler:intercomInputStateChanged(params)
    if params.input_id == 1 then
        self.intercom:updateProperty('input1', params.state > 0)
    elseif params.input_id == 2 then
        self.intercom:updateProperty('input2', params.state > 0)
    end
end

-- Events
function IntercomEventsHandler:createIncomingCallEvent()
    local params = {
        url = self.publishEventUrl,
        method = 'POST',
        data = json.encode({
            ['type'] = 'VideoGateIncomingCallEvent',
            ['data'] = {
                ['id'] = self.intercom.id
            }
        })
    }

    return params
end

-- Notifications 
function IntercomEventsHandler:createNotificationParams(text, priority)
    local params = {
        url = self.notificationCenterUrl,
        method = 'POST',
        data = json.encode({
            ['type'] = 'GenericDeviceNotification',
            ['canBeDeleted'] = true,
            ['priority'] = priority or 'alert',
            ['data'] = {
                ['deviceId'] = self.intercom.id,
                ['text'] = text or 'Unknown signal',
                ['title'] = text
            }
        })
    }

    return params
end

function IntercomEventsHandler:sendNoiseNotification()
    self.intercom:httpRequest(self:createNotificationParams('i18n_notificationBabayCrying'))
end

function IntercomEventsHandler:sendGlassBreakNotification()
    self.intercom:httpRequest(self:createNotificationParams('i18n_notificationGlassBreak'))
end

function IntercomEventsHandler:sendSmokeNotification()
    self.intercom:httpRequest(self:createNotificationParams('i18n_notificationSmokeAlarm'))
end

function IntercomEventsHandler:sendCarbonMonoxideNotification()
    self.intercom:httpRequest(self:createNotificationParams('i18n_notificationCarbon-Monoxide'))
end

function IntercomEventsHandler:sendImpactNotification()
    self.intercom:httpRequest(self:createNotificationParams('i18n_notificationGunshot'))
end

function IntercomEventsHandler:sendModeChangedNotification(mode)
    self.intercom:httpRequest(self:createNotificationParams('Intercom mode changed to: ' .. mode))
end

function IntercomEventsHandler:sendGroupCompleteChangeNotification()
    self.intercom:httpRequest(self:createNotificationParams('i18n_notificationGroupComplete'))
end

function IntercomEventsHandler:sendIntercomNameChangedNotification(params )
    if params and params.name then
        self.intercom:httpRequest(self:createNotificationParams('Intercom name changed to: ' .. params.name))
    end
end

function IntercomEventsHandler:sendIntercomAccountChangedNotification(params)    
    if params and params.change_type and params.user_name and params.change_type == 'Added' then
        self.intercom:httpRequest(self:createNotificationParams('User ' .. params.user_name .. ' added.' ))
    elseif params and params.change_type and params.user_name and params.change_type == 'Removed' then
        self.intercom:httpRequest(self:createNotificationParams('User ' .. params.user_name .. ' removed.' ))
    end
end


