require('common.device')
require('net.HTTPClient')
cosmo = require("cosmo.cosmo")
require('common.validator')

class 'Helios' (Device)

--! Initializes Helios plugin object.
--@param id: Id of the device.
function Helios:__init(id)
	Device.__init(self, id)
	self.hcNetworkData = api.get('/settings/network')
	self.voipDevicesList = api.get('/voip')
	self.heliosConnected = false
	self.syncStatusForHeliosSipUser = false
	self.syncStatusForUsersToCall = false
	self.availablePhoneBookEntries = 1999
	self.sid = ''
	
	self.http = net.HTTPClient({ timeout = 10000 }) 

	self.url = {
	["getSid"] = '/ajax?sid=',
	["saveConfig"] = '/upload?sid='
	}

	self.headersApplication = {
	['Content-Type'] = 'application/x-www-form-urlencoded'
	}

	self.headersApplicationJson = {
	['Content-Type'] = 'application/json'
	}

	self.headersMultipart = {
	['Content-Type'] = 'multipart/form-data; boundary=----WebKitFormBoundaryRW7n2B84eK5zGTX5'
	}
	
	self:initialMethods()
end

--! Initial methods after save button pressed
function Helios:initialMethods()
	if (self:credentialsExist()) then
		self.credentials = '[{"command":"system.login","user":"'..self:urlencode(self.properties.userName)..'","password":"'..self:urlencode(self.properties.password)..'"}]"'
	end

	if (not self:sipPropertiesExist()) then
		self:setDefaultPropertyForSipHeliosUser()
		setTimeout(function () self:addSipUserToHelios(); end, 2000)
	else
		self:addSipUserToHelios()
	end

	self:setDefaultsForHelios()
	self:clearHeliosPhoneBook()
	setTimeout(function () self:addSipUsersToCallFromHelios(self.properties.voipUsersForHeliosPhoneBook); end, 4000)
	
	if (self.timeoutId) then
		clearTimeout(self.timeoutId)
	end

	self:syncAndCheckHeliosStatus()
end

--! Helper method. Checks if sip properties exist.
--!@return true if sip properties exist exist or false if not.
function Helios:sipPropertiesExist()
	if (self.properties.sipDisplayName == nil or
		self.properties.sipDisplayName == '' or
		self.properties.sipUserID == nil or 
		self.properties.sipUserID == '' or
		self.properties.sipUserPassword == nil or
		self.properties.sipUserPassword == '') then

		return false
	else
		return true
	end
end

--! Helper method. Check if user and password exist.
--!@return true if credentials exist or false if not.
function Helios:credentialsExist()
	if (self.properties.userName ~= nil and self.properties.userName ~= '' and
   	    self.properties.password ~= nil and self.properties.password ~= '' ) then
   	  	return true
	else
		return false
   	end
end

--! Function that is needed to obtain an sid from Helios to give us ability for creating POST requests.
function Helios:getSid(paramCallback)
	if(self.credentials) then
		local params = {
				url = 'http://' .. self.properties.ip .. self.url.getSid,
				headers = self.headersApplicationJson,
				body = self.credentials,
				callbackAfterOnDataReceivedFromGetSid = paramCallback,
				callback =  function(response, callbackAfterOnDataReceivedFromGetSid) self:onDataReceivedFromGetSid(response, callbackAfterOnDataReceivedFromGetSid) end,
			  }

		self:postRequest(params)
	else
		self.heliosConnected = false
	end
end

--! Helper method. Creates url encoded data from given string.
--!@param str: String to be encoded. 
function Helios:urlencode(str)
  if (str) then
    str = string.gsub (str, "\n", "\r\n")
    str = string.gsub (str, "([^%w %-%_%.%~])",
        function (c) return string.format ("%%%02X", string.byte(c)) end)
    str = string.gsub (str, " ", "+")
  end

  return str	
end

--! Creates and manages http POST requests.
--!@param url: Url to resource.
function Helios:postRequest(params)	
	if (not isValidUrl(params.url)) then
		return
	end

	self.http:request(params.url, {	
		options={
			headers = params.headers,
			method = 'POST',
			data = params.body,
		}, 
		success = function(response)
			if (params.callback) then
				params.callback(response, params.callbackAfterOnDataReceivedFromGetSid)
			elseif  (params.updateSyncStatusForUsersToCall) then
				params.updateSyncStatusForUsersToCall(true)
			elseif  (params.updateSyncStatusForHeliosSipUser) then
				params.updateSyncStatusForHeliosSipUser(true)
			end
		end,
		error = function (error)
			self.heliosConnected = false

			print('*******************')
			print('* ERROR: '.. error)
			print('*******************')

			if  (params.updateSyncStatusForUsersToCall) then
				params.updateSyncStatusForUsersToCall(false)
			elseif  (params.updateSyncStatusForHeliosSipUser) then
				params.updateSyncStatusForHeliosSipUser(false)
			end
		end
	})
end

--! Method fired as callback to getSid function.
function Helios:onDataReceivedFromGetSid(response, callback)
 	self.heliosConnected = false

	if (not isJson(response.data)) then return end

	local decodedData = json.decode(response.data)

	self.sid = decodedData[1].sid
	if (self.sid ~= '' and self.sid)  then
		self.heliosConnected = true

		if (callback) then
			callback()
		end
	end
end

--! Start point for adding new sip user to Helios. Prepares payload and fire getSid function.
function Helios:addSipUserToHelios()
	local voipDataPayload = 
[[
------WebKitFormBoundaryRW7n2B84eK5zGTX5
Content-Disposition: form-data; name="cmd-options"

NSCP
------WebKitFormBoundaryRW7n2B84eK5zGTX5
Content-Disposition: form-data; name="file-xml-config"; 
Content-Type: text/xml

<?xml version="1.0" encoding="UTF-8"?>
<DeviceDatabase Version="0">
<Phone>
	<Sip>
		<User>
			<DisplayName>$DisplayName</DisplayName>
			<Id>$Id</Id>
			<AuthId>$AuthId</AuthId>
			<PasswordString>$PasswordString</PasswordString>
			<UseAuthId>0</UseAuthId>
			<CommunicationType>1</CommunicationType>
		</User>
		<Client>
			<Domain>$Domain</Domain>
			<Port>5060</Port>
		</Client>
		<Proxy>
			<Address>$Address</Address>
			<Port>5060</Port>
		</Proxy>
		<Registrar>
			<Enabled>1</Enabled>
			<Address>$Address</Address>
			<Port>5060</Port>
			<ExpireTime>120</ExpireTime>
		</Registrar>
		<Misc>
			<KeepAlive>1</KeepAlive>
			<Dscp>0</Dscp>
			<IpFilter>0</IpFilter>
		</Misc>
	</Sip>
</Phone>
</DeviceDatabase>

------WebKitFormBoundaryRW7n2B84eK5zGTX5--
]]

	voipDataPayload = cosmo.fill(voipDataPayload, {
							DisplayName = self:urlencode(self.properties.sipDisplayName), 
							Id = self:urlencode(self.properties.sipUserID),
							AuthId = self:urlencode(self.properties.sipUserID),
							PasswordString = self:urlencode(self.properties.sipUserPassword),
							Domain = self.hcNetworkData.ip,
							Address = self.hcNetworkData.ip									
						}) 

	self:updateSyncStatusForHeliosSipUser(false)
	self:getSid(function() self:injectNewDataToHelios(voipDataPayload, "addSipUserToHelios") end)
end

--! Start point for adding new sip user that we would like to call from Helios. Prepares payload and fire getSid function.
function Helios:addSipUsersToCallFromHelios(usersList)
	local voipUsersTable = usersList
	local voipUsersListToPaylod = ''
	local voipFinalDataPayload = ''

	if (#voipUsersTable == 0 or voipUsersTable == nil) then return end

	for k,v in pairs(voipUsersTable) do
		for k2,v2 in pairs(v) do
			if (k2 == 'columns') then
				for k3,v3 in pairs(v2) do
					if (k3 == 1) then -- this iteration containg voipDevieceID						
						voipUserToInject = self:getVoipUserObjectByID(v3)	

						if (voipUserToInject == nil) then break end

					elseif (k3 == 3) then
						local code = tonumber(v3)
						voipUserToInject['keyCode'] = code - 1 -- needed for mismatch with helios phoneBook start number (is 1 should be 0)
					end
				end

				if (voipUserToInject == nil) then break end

				voipUserXML = 
[[
<User At="$KeyCode">
	<Enabled>1</Enabled>
	<ShortName>$ShortName</ShortName>
	<Email></Email>
	<Number At="0">
		<Number>$Number</Number>
		<Profile>-1</Profile>
		<GroupCall>0</GroupCall>
		<StationName></StationName>
	</Number>
	<Number At="1">
		<Number></Number>
		<Profile>-1</Profile>
		<GroupCall>0</GroupCall>
		<StationName></StationName>
	</Number>
	<Number At="2">
		<Number></Number>
		<Profile>-1</Profile>
		<GroupCall>0</GroupCall>
		<StationName></StationName>
	</Number>
	<Deputy>-1</Deputy>
	<ActivationCode></ActivationCode>
	<DeactivationCode></DeactivationCode>
</User>
]]

				voipUserXML = cosmo.fill(voipUserXML, {
														KeyCode = voipUserToInject.keyCode,
														ShortName = voipUserToInject.sipDisplayName, 
														Number = 'sip:'..voipUserToInject.sipUserID..'@'..self.hcNetworkData.ip..':5060'																		
													}) 
			
				voipUsersListToPaylod = voipUsersListToPaylod .. voipUserXML
			end	
		end
	end

	voipFinalDataPayload = 
[[
------WebKitFormBoundaryRW7n2B84eK5zGTX5
Content-Disposition: form-data; name="cmd-options"

NSCP
------WebKitFormBoundaryRW7n2B84eK5zGTX5
Content-Disposition: form-data; name="file-xml-config"; 
Content-Type: text/xml

<?xml version="1.0" encoding="UTF-8"?>
<DeviceDatabase Version="0">
$VoipUsersListToPaylod
</DeviceDatabase>

------WebKitFormBoundaryRW7n2B84eK5zGTX5--
]]

	voipFinalDataPayload = cosmo.fill(voipFinalDataPayload, {
														VoipUsersListToPaylod = voipUsersListToPaylod 																	
													}) 

	self:updateSyncStatusForUsersToCall(false)
	self:getSid(function() self:injectNewDataToHelios(voipFinalDataPayload, "addSipUsersToCallFromHelios") end)
end

--! Sets all Helios phone book entries to disable. Prepares payload and fire getSid function.
function Helios:clearHeliosPhoneBook()
	local clearVoipUsersListToPaylod = ""
	local clearVoipFinalDataPayload = ""

	for i=1,self.availablePhoneBookEntries do

	 celearVoipUserXML = 
[[
<User At="$KeyCode"><Enabled>0</Enabled></User>
]]

				celearVoipUserXML = cosmo.fill(celearVoipUserXML, {
														KeyCode = i-1,																
													}) 
			
				clearVoipUsersListToPaylod = clearVoipUsersListToPaylod .. celearVoipUserXML
	end

	clearVoipFinalDataPayload = 
[[
------WebKitFormBoundaryRW7n2B84eK5zGTX5
Content-Disposition: form-data; name="cmd-options"

NSCP
------WebKitFormBoundaryRW7n2B84eK5zGTX5
Content-Disposition: form-data; name="file-xml-config"; 
Content-Type: text/xml

<?xml version="1.0" encoding="UTF-8"?>
<DeviceDatabase Version="0">
$VoipUsersListToPaylod
</DeviceDatabase>

------WebKitFormBoundaryRW7n2B84eK5zGTX5--
]]

	clearVoipFinalDataPayload = cosmo.fill(clearVoipFinalDataPayload, {
														VoipUsersListToPaylod = clearVoipUsersListToPaylod 																	
													}) 

	self:updateSyncStatusForUsersToCall(false)
	self:getSid(function() self:injectNewDataToHelios(clearVoipFinalDataPayload, "addSipUsersToCallFromHelios") end)
end

--! Sets default for Helios device. Prepares payload and fire getSid function.
function Helios:setDefaultsForHelios()
local dataPayload = 
[[
------WebKitFormBoundaryRW7n2B84eK5zGTX5
Content-Disposition: form-data; name="cmd-options"

NSCP
------WebKitFormBoundaryRW7n2B84eK5zGTX5
Content-Disposition: form-data; name="file-xml-config"; 
Content-Type: text/xml

<?xml version="1.0" encoding="UTF-8"?>
<DeviceDatabase Version="0">

<Picture>
<AnonymousAccess>1</AnonymousAccess>
</Picture>

<Switches>
<HttpControlEnabled>1</HttpControlEnabled>
<CallActivation>1</CallActivation>
</Switches>

<Phone>
<Calls>
<Incoming>
<AutoPickupEnabled>0</AutoPickupEnabled>
</Incoming>
</Calls>
</Phone>

<Keyboard>
<QuickDialByKeypad>1</QuickDialByKeypad>
</Keyboard>

</DeviceDatabase>

------WebKitFormBoundaryRW7n2B84eK5zGTX5--
]]

	self:getSid(function() self:injectNewDataToHelios(dataPayload, "setDefaultsForHelios") end)
end

--! Method responsible for setting new data in to Helios device.
--@param dataToInject: Data that will be set in to Helios. 
function Helios:injectNewDataToHelios(dataToInject, actionType)
	if(self.credentials) then
		local params = {}
		if (actionType == "addSipUserToHelios") then
			params = {
					url = 'http://' .. self.properties.ip .. self.url.saveConfig .. self.sid,
					headers = self.headersMultipart,
					body = dataToInject,
					updateSyncStatusForHeliosSipUser =  function(status) self:updateSyncStatusForHeliosSipUser(status) end,
				  }
		elseif (actionType == "addSipUsersToCallFromHelios") then
			params = {
					url = 'http://' .. self.properties.ip .. self.url.saveConfig .. self.sid,
					headers = self.headersMultipart,
					body = dataToInject,
					updateSyncStatusForUsersToCall =  function(status) self:updateSyncStatusForUsersToCall(status) end,
				  }
		else
			params = {
					url = 'http://' .. self.properties.ip .. self.url.saveConfig .. self.sid,
					headers = self.headersMultipart,
					body = dataToInject,		
					}		
		end

		self:postRequest(params)
	else
		self.heliosConnected = false
	end
end

--! Simple setter for syncStatusForHeliosSipUser field.
function Helios:updateSyncStatusForHeliosSipUser(status)
	self.syncStatusForHeliosSipUser = status
end

--! Simple setter for syncStatusForUsersToCall field.
function Helios:updateSyncStatusForUsersToCall(status)
	self.syncStatusForUsersToCall = status
end

--! Calculates and gets voip users list.
--!@return voip users list.
function Helios:getVoipUsersList(arg)
	local usersList = {}
	table.insert(usersList, {["value"]='---', ["label"]='---'})
	local sipDisplayName = ''
	local id = ''
	self.voipDevicesList = api.get('/voip')

	for k,v in pairs(self.voipDevicesList.voipDevices) do
		for k2,v2 in pairs(v) do
			if (k2 == 'sipDisplayName' and v2 ~= '' and v2 ~= self.properties.sipDisplayName) then
				sipDisplayName = v2
			end
			if (k2 == 'id' and v2 ~= '') then
				id = v2
			end
		end

		if (id ~= '' and sipDisplayName ~= '') then
			table.insert(usersList, {["value"]=tostring(id), ["label"]=sipDisplayName})
	    end

		sipDisplayName = ''
		id = ''
	end

	return usersList
end

--! Calculate codes lists and return it to select control that is connected with this method.
--!@return codes list. 
function Helios:getVoipCodesList(arg)
	local userCodes = {}
	table.insert(userCodes, {["value"]='---', ["label"]='---'})
	local iString

	for i=1,self.availablePhoneBookEntries do
		if (i < 10) then
			iString = tostring(i)
			table.insert(userCodes, {["value"]='0'..iString, ["label"]='0'..iString})
		else
			iString = tostring(i)
			table.insert(userCodes, {["value"]=iString, ["label"]=iString})
		end
	end

	return userCodes
end

--! Adds voip user to plugin list.
function Helios:addVoipUserRowToJsonObjectForHeliosPhoneBook()
	local rowToInsert = {}

	if (self:canAddVoipUser()) then
		local voipUsersTable = self.properties.voipUsersForHeliosPhoneBook
		rowToInsert = {["columns"]={self.properties.selectedVoipUser, self:getVoipUserNameById(self.properties.selectedVoipUser), self.properties.selectedKeyCode, "Delete"}}
		table.insert(voipUsersTable, rowToInsert)

		self:updateProperty('voipUsersForHeliosPhoneBook', voipUsersTable, true)
	end

	self:updateProperty('selectedVoipUser', '---')
	self:updateProperty('selectedKeyCode', '---')
end

--! Deletes voip user from plugin list and tries to sync it with Helios device list.
--!@return "delete" string to notify table control (connected with this method) of taken action.
function Helios:deleteVoipUserRowFromJsonObjectForHeliosPhoneBook(tableRowId)	
	local voipUsersTable = self.properties.voipUsersForHeliosPhoneBook

	for k,v in pairs(voipUsersTable) do
		if ((k-1) == tableRowId) then
			table.remove(voipUsersTable, k)
		end
	end

	self:updateProperty('voipUsersForHeliosPhoneBook', voipUsersTable, true)

	setTimeout(function () self:clearHeliosPhoneBook(); end, 1000)
	setTimeout(function () self:addSipUsersToCallFromHelios(self.properties.voipUsersForHeliosPhoneBook); end, 5000)

	if (self.timeoutId) then
		clearTimeout(self.timeoutId)
	end

	self:syncAndCheckHeliosStatus()

	return "delete"
end

--! Sets default properties for sip Helios User.
function Helios:setDefaultPropertyForSipHeliosUser()
	self:updateProperty('sipDisplayName', 'Helios Basic')
	self:updateProperty('sipUserID', '222222')
	self:updateProperty('sipUserPassword', '222222')
end

--! Helper method. Gets voip user name by ID from plugin voipUsers list.
--!@return name string or nil.
function Helios:getVoipUserNameById(voipUserId)
	local sipDisplayName = ''
	local id = ''
	self.voipDevicesList = api.get('/voip')

	for k,v in pairs(self.voipDevicesList.voipDevices) do
		for k2,v2 in pairs(v) do
			if (k2 == 'sipDisplayName' and v2 ~= '' and v2 ~= self.properties.sipDisplayName) then
				sipDisplayName = v2
			end
			if (k2 == 'id' and tostring(v2) == voipUserId) then
				id = v2
			end
		end

		if (id ~= '' and sipDisplayName ~= '') then
			return sipDisplayName 
	    end
	end
				
	return nil
end

--! Helper method. Gets voip user by ID from HC2 api voip list.
--!@return voip user object or nil.
function Helios:getVoipUserObjectByID(ID)
	self.voipDevicesList = api.get('/voip')
	
	for k,v in pairs(self.voipDevicesList.voipDevices) do
		for k2,v2 in pairs(v) do
			if (k2 == 'id' and tostring(v2) == tostring(ID)) then
				return v
			end
		end
	end
	
	return nil
end

--! Helper method. Check if can add voip user.
--!@return true if can add voip user.
function Helios:canAddVoipUser()
	local voipUsersTable = self.properties.voipUsersForHeliosPhoneBook
	if (self.properties.selectedVoipUser == '---' or
		self.properties.selectedKeyCode == '---' or 
		self.properties.selectedVoipUser == '' or 
		self.properties.selectedKeyCode == '' ) then
		return false
	end

	for k,v in pairs(voipUsersTable) do
		for k2,v2 in pairs(v) do
			if (k2 == 'columns') then
				for k3,v3 in pairs(v2) do
					if (v3 == self.properties.selectedKeyCode) then
						return false
					end
				end
			end	
		end
	end

	return true
end

--! 10 seconds infinite polling loop checking helios status and init sync if needed.
function Helios:syncAndCheckHeliosStatus()
    self.timeoutId = setTimeout(function( ) self:syncAndCheckHeliosStatus(); end, 10000)
    self:getSid()

    if (self.syncStatusForHeliosSipUser == false) then
    	if (not self:sipPropertiesExist()) then
			self:setDefaultPropertyForSipHeliosUser()
			setTimeout(function () self:addSipUserToHelios(); end, 2000)
		else
			self:addSipUserToHelios()
		end
	end

    if (self.syncStatusForUsersToCall == false) then
    		self:clearHeliosPhoneBook()
			setTimeout(function () self:addSipUsersToCallFromHelios(self.properties.voipUsersForHeliosPhoneBook); end, 4000)
    end
end







