require('common.device')
require('common.base64')
require('net.HTTPClient')
cosmo = require("cosmo.cosmo")
require('common.validator')
require('SyncQueue')

class 'Mobotix' (Device)

--! Initializes Mobotix plugin object.
--@param id: Id of the device.
function Mobotix:__init(id)
	Device.__init(self, id)

	self.queueCommandId = 0;
	self.nextCommandTimeoutId = nil
	self.canExecuteCommandByLoop = true
	self.isPluginOnline = false
	self.availablePhoneBookEntries = 200

	self.hcNetworkData = api.get('/settings/network')
	self.voipDevicesList = api.get('/voip')
	
	self.http = net.HTTPClient({ timeout = 20000 }) 

	self.url = {
		["openDoor"] = '/control/rcontrol?action=customfunction&action=sigout&profile=~Door',
		["configuration"] = '/admin/remoteconfig',
		["checkVoipUsers"] = '/admin/control?list&section=phoneprofiles',
		["checkPlugin"] = '/admin',
		["getConciergeSection"] = '/admin/concierge',
	}
	
	self:initSyncQueue()
	self:startPolling()
	self:updatePluginSettings()
end

--! Updates Mobotix plugin.
function Mobotix:updatePluginSettings()
	self.headers = {
		['Authorization'] = 'Basic '.. to_base64(self.properties.userName .. ':'.. self.properties.password),
		['Content-Type'] = 'application/x-www-form-urlencoded'
	}

	self:addVoipUserRowToJsonObjectForMobotixPhoneBook()
	self:runVoipUsersCreationFunctions()
end

--! Initialize process of syncQueue that is responsible for sync plugin data with device (Mobotix).
function Mobotix:initSyncQueue()
    self.syncQueue = SyncQueue()   
end

--! 2 seconds infinite polling loop for two purposes. 
--! 1) Tries execute command from syncQueue.
--! 2) Executes probing  method that checks if device (Mobotix) is alive.
function Mobotix:startPolling()
    if (self.nextCommandTimeoutId) then
  	 	clearTimeout(self.nextCommandTimeoutId)
	end

    self.nextCommandTimeoutId = setTimeout(function( ) self:startPolling(); end, 2000)

    self:isPluginAlive()
    self:executeCommandByLoopFromQueue()
end

--! Method responsible for cyclic check if plugin is online.
function Mobotix:isPluginAlive()
	local params = {
			url = self.url.checkVoipUsers,
			callback =  function(response) self:onDataReceivedFromIsPluginAlive(response) end,
		  } 

	self:getRequest(params)
end

--! This function is responsible for managing data received from resources called from correspond method.
--!@param response: response body.
--!@param identifier: callback function identifier.
function Mobotix:onDataReceivedFromIsPluginAlive(response, identifier)
	if (response.status and string.find(response.data, 'INF: Showing section:')) then
		self.isPluginOnline = true
    else
		self.isPluginOnline = false
   	end	
end

--! Tries to run first command from queue if canExecuteCommandByLoop is true.
function Mobotix:executeCommandByLoopFromQueue()
	if(self.syncQueue:getQueueTable() and self.canExecuteCommandByLoop) then
		local commandTalbe = self.syncQueue:getFrontElement()

		if commandTalbe  then
			for identifier,commdnFunction in pairs(commandTalbe) do
				commdnFunction(identifier)
			end	
		end
	end
end

--! Creates and manages http GET requests.
--!@param params - request options.
--!@param identifier - callback function identifier.
function Mobotix:getRequest(params, identifier)
	local url = 'http://' .. self.properties.ip..params.url
	self.http:request(url, {
		options ={
			headers = self.headers,	
			method = 'GET',	
		},
		success = function(response)
			if (params.callback) then
				params.callback(response, identifier)
			end
		end,
		error = function (error)
			if (params.callback) then
				params.callback(error, identifier)
			end
		end
	})
end

--! Creates and manages http POST requests.
--!@param params - request options.
--!@param identifier - callback function identifier.
function Mobotix:postRequest(params, identifier)	
	self.canExecuteCommandByLoop = false
	local url = 'http://' .. self.properties.ip..params.url

	self.http:request(url, {	
		options = {
			headers = self.headers,
			method = 'POST',
			data = params.body,
		},
		success = function(response)
			if (params.callback) then
				params.callback(response, identifier)
			end
		end,
		error = function (error)
			if (params.callback) then
				params.callback(error, identifier)
			end
		end
	})
end

--! Method assigned to open door button.
--! It is responsible to prepare appropriate params and send it to getRequest method.
function Mobotix:openDoor()
	local params = {
			url = self.url.openDoor,
			callback =  function(response, identifier) self:onDataReceivedFromOpenDoor(response, identifier) end,
		  } 

	self:getRequest(params)
end

--! This function is responsible for managing data received from resources called from corresponded method. 
--!@param response: received data.
--!@param identifier: callback function identifier.
function Mobotix:onDataReceivedFromOpenDoor(response, identifier)
	if (response.status == 200 and string.find(response.data, 'OK'))then
		print('Succes STATUS: ', response.status)
		print('Succes DATA: ', response.data)
    else
    	print('*******************')
		print('* ERROR: ', response, response.status, response.data)
		print('*******************')
   	end	
end

--! Method responsible for create and inject new data to mobotix device.
--!@param queue Identifier: id of command in queue object.
function Mobotix:recreateViopUsersInMobotix(queueIdentifier)
	local conciergePayload = ''
	local audionotifyPayload =
[[
AMSN=0
VOIPSPEAKERGAIN=350
VOIPSIG1MODE=
VOIPSIG2MODE=
VOIPSIG3MODE=
VOIPVOIP=on
VOIPSETUPMODE=expert
VOIPPEERDATATIMEOUT=5
VOIPUSERNAME=
VOIPHOSTNAME=
VOIPADDROFREC=sip:mx10-14-47-5@10.14.47.5
VOIPNAT=off
VOIPNATADDR=
VOIPNATREFRESH=60
VOIPSIPPORT=5060
VOIPAUDIORTPPORT=7078
VOIPCODECPCMA=on
VOIPCODECGSM=off
VOIPCODECPCMU=off
VOIPVIDEO=on
VOIPVIDEOSIZE=auto
VOIPVIDEOFRAMERATE=
VOIPVIDEOBITRATE=500000
VOIPVIDEOQUALITY=
VOIPVIDEORTPPORT=9078
VOIPVIDEOCODEC1=h264
VOIPVIDEOCODEC2=h263+
VOIPWELCOMEIN=on
VOIPWELCOMEOUT=on
VOIPDTMFKEYSIN=on
VOIPDTMFKEYSOUT=on
VOIPSTATUSONSPEAKER=on
VOIPMENUAUTOHIDE=off
VOIPMENUSETTINGS=on
VOIPHOMEPOSDSPMODE=none
VOIPHANGUPONDIALOUT=off
VOIPMULTICALL=off
VOIPWELCOMEMSGDELAY=2
$VOIPSIPACCOUNT
VOIPDEFAULTPROXY=
cin_mymsn=
cin_remotemsn=
cin_enabled=0
cin_duration=5
cin_mode=4
cin_pin=
cin_welcome=Default
cin_emailevent=
cin_isdnprofile=
cin_isdnidle=
mutelevel1=4
mutetime1=666
VOIPVOIPIN=on
VOIPMESSAGE=Default
VOIPHANGUP=300
VOIPINMODE=both
VOIPREMOTEURL=
VOIPALLOWEDVIAS=
$PROFILE
]]

	--audionotify section fields
	local voipSipAccount = 'VOIPSIPACCOUNT=0:userid='..self.properties.sipUserID..':domain='..self.hcNetworkData.ip..':server='..self.hcNetworkData.ip..':port=5060:register=1:defaultproxy=1:reginterval=900:authid='..self.properties.sipUserID..':authpwd='..self.properties.sipUserPassword
    local voipUsersTable = self.properties.voipUsersForMobotixPhoneBook
    local profile = ''


    --concierge.json section fields
	local usersList = {
		["addressees"]={},
		["meta"]={["id_timestamp"]="Written by concierge:serialize() at Fri May 30 14:54:45 2014\n", ["schema_version"]=0},
		["users_version"]=0,
		["ring_devs"]={}
	}

	if (#voipUsersTable ~= 0 or voipUsersTable ~= nil) then 
		for k,v in pairs(voipUsersTable) do
			for k2,v2 in pairs(v) do
				if (k2 == 'columns') then
					for k3,v3 in pairs(v2) do
						if (k3 == 1) then -- this iteration containg voipDevieceID						
							voipUserToInject = self:getVoipUserObjectByID(v3)	
							if (voipUserToInject == nil) then break end	
						elseif (k3 == 3) then

							local code = tonumber(v3)
							voipUserToInject['keyCode'] = code
						end
					end

					if (voipUserToInject == nil) then break end

					local addressees = {["id"]=k, ["ring"]={["selected"]=0, ["profiles"]={}}, ["name"]='adress'..voipUserToInject.sipUserID, ["contact_number"] = tostring(voipUserToInject.keyCode)}
					local profiles = {["name"]="callProfile"..voipUserToInject.sipUserID, ["list"]={}}
					local list = {["timeout"]=30, ["devices"]={}}
					local ringDevices = {["id"]=1700+k, ["type"]="phone", ["profile"]="call_"..voipUserToInject.sipUserID}

					table.insert(list.devices, 1700+k)
					table.insert(profiles.list, list)
					table.insert(addressees.ring.profiles, profiles)
					table.insert(usersList.addressees, addressees)
					table.insert(usersList.ring_devs, ringDevices)

					profile = profile ..'\n' .. 'profile=call_'..voipUserToInject.sipUserID..':_profilename='..voipUserToInject.sipDisplayName..voipUserToInject.sipUserID..':mode=both:message=Phonering:duration=0:destination='..voipUserToInject.sipUserID..'@'..self.hcNetworkData.ip..',5,1,-1;:pin=:dtmf=1:type=ipv'
				end
			end
		end
	else
		profile = ''
		conciergePayload = ''
	end

	audionotifyPayload = cosmo.fill(audionotifyPayload, {
				VOIPSIPACCOUNT = voipSipAccount,
				PROFILE = profile
		})

	conciergePayload = json.encode(usersList)

	local dataPayload =
[[

helo
view section concierge.json
write
SECTION audionotify
$AUDIONOTIFY_PAYLOAD
ENDSECTION audionotify
write
SECTION concierge.json
$CONCIERGE_PAYLOAD
ENDSECTION concierge.json
write
SECTION iowiring
dpc_input=
dpc_negate=1
bpc_input=
bpc_negate=1
door_output=mo,6
door_time=2
door_negate=0
light_output=
light_time=120
light_negate=0
ilight_output=
ilight_negate=0
bell=
profile=
ENDSECTION iowiring
view section concierge.json
update
store
quit

]]

	dataPayload = cosmo.fill(dataPayload, {
				CONCIERGE_PAYLOAD = conciergePayload,
				AUDIONOTIFY_PAYLOAD = audionotifyPayload,
		})

	local params = {
				url = self.url.configuration,
				body = dataPayload,
				callback =  function(response, identifier) self:onDataReceivedFromRecreateViopUsersInMobotix(response, identifier) end,
			  } 

	local commandTable = {[queueIdentifier] = function(identifier)self:postRequest(params, identifier)end}		  
	self.syncQueue:pushBack(commandTable)
end

--! This function is responsible for managing data received from resources called from corresponded method. 
--!@param response: received data.
--!@param identifier: callback function identifier. 
function Mobotix:onDataReceivedFromRecreateViopUsersInMobotix(response, identifier)
	if (response.status == 200 and string.find(response.data, '#Storing to flash...')) then
		print('Succes STATUS: ', response.status)
		print('Succes DATA: ', response.data)

		if(self.syncQueue:getQueueTable()) then
			self.syncQueue:removeCommand(identifier)
			local commandTalbe = self.syncQueue:getFrontElement()

			if commandTalbe then
				self.canExecuteCommandByLoop = false
				for identifier,commdnFunction in pairs(commandTalbe) do
					commdnFunction(identifier)
				end	
			else
				self.canExecuteCommandByLoop = true
			end
		end
    else
    	self.canExecuteCommandByLoop= true
    	print('*******************')
		print('* ERROR: ', response, response.status, response.data)
		print('*******************')
   	end	
end

--! Start point of run creation and injection of data for Mobotix device
function Mobotix:runVoipUsersCreationFunctions()
	self.queueCommandId = self.queueCommandId + 1
	self:recreateViopUsersInMobotix(self.queueCommandId)
end

--! Set default properties for sip Mobotix User.
function Mobotix:setDefaultPropertyForSipMobotixUser()
	self:updateProperty('sipDisplayName', 'Mobotix')
	self:updateProperty('sipUserID', '333333')
	self:updateProperty('sipUserPassword', '333333')
end

--! Helper method. Gets voip user name by ID from plugin voipUsers list.
--!@return name string or nil.
function Mobotix:getVoipUserNameById(voipUserId)
	local sipDisplayName = ''
	local id = ''
	self.voipDevicesList = api.get('/voip')

	for k,v in pairs(self.voipDevicesList.voipDevices) do
		for k2,v2 in pairs(v) do
			if (k2 == 'sipDisplayName' and v2 ~= '' and v2 ~= self.properties.sipDisplayName) then
				sipDisplayName = v2
			end
			if (k2 == 'id' and tostring(v2) == voipUserId) then
				id = v2
			end
		end

		if (id ~= '' and sipDisplayName ~= '') then
			return sipDisplayName 
	    end
	end
				
	return nil
end

--! Helper method. Gets voip user by ID from HC2 api voip list.
--!@return voip user object or nil.
function Mobotix:getVoipUserObjectByID(ID)
	self.voipDevicesList = api.get('/voip')
	for k,v in pairs(self.voipDevicesList.voipDevices) do
		for k2,v2 in pairs(v) do
			if (k2 == 'id' and tostring(v2) == tostring(ID)) then
				return v
			end
		end
	end
	
	return nil
end

--! Helper method. Check if can add voip user.
--!@return true if can add voip user.
function Mobotix:canAddVoipUser()
	local voipUsersTable = self.properties.voipUsersForMobotixPhoneBook
	if (self.properties.selectedVoipUser == '---' or
		self.properties.selectedKeyCode == '---' or 
		self.properties.selectedVoipUser == '' or 
		self.properties.selectedKeyCode == '' ) then

		return false
	end

	if (#voipUsersTable == 0 or voipUsersTable == nil) then return true end

	for k,v in pairs(voipUsersTable) do
		for k2,v2 in pairs(v) do
			if (k2 == 'columns') then
				for k3,v3 in pairs(v2) do
					if (v3 == self.properties.selectedKeyCode) then
						return false
					end
				end
			end	
		end
	end

	return true
end

--! Adds voip user to plugin property list.
function Mobotix:addVoipUserRowToJsonObjectForMobotixPhoneBook()
	local rowToInsert = {}

	if (self:canAddVoipUser()) then
		local voipUsersTable = self.properties.voipUsersForMobotixPhoneBook
		rowToInsert = {["columns"]={self.properties.selectedVoipUser, self:getVoipUserNameById(self.properties.selectedVoipUser), self.properties.selectedKeyCode, "Delete"}}
		table.insert(voipUsersTable, rowToInsert)

		self:updateProperty('voipUsersForMobotixPhoneBook', voipUsersTable, true)
	end

	self:updateProperty('selectedVoipUser', '---')
	self:updateProperty('selectedKeyCode', '---')
end

--! Deletes voip user from plugin property list.
--!@return "delete" string to notify table control (connected with this method) of taken action.
function Mobotix:deleteVoipUserRowFromJsonObjectForMobotixPhoneBook(tableRowId)	
	local voipUsersTable = self.properties.voipUsersForMobotixPhoneBook

	if (voipUsersTable == nil or voipUsersTable == '') then return end
	for k,v in pairs(voipUsersTable) do
		if ((k-1) == tableRowId) then
			table.remove(voipUsersTable, k)
		end
	end

	self:updateProperty('voipUsersForMobotixPhoneBook', voipUsersTable, true)

	return "delete"
end

--! Calculates voip users list and return it to select control that is connected with this method.
--!@return voip users list.
function Mobotix:getVoipUsersList(arg)
	local usersList = {}
	table.insert(usersList, {["value"]='---', ["label"]='---'})
	local sipDisplayName = ''
	local id = ''
	self.voipDevicesList = api.get('/voip')

	for k,v in pairs(self.voipDevicesList.voipDevices) do
		for k2,v2 in pairs(v) do
			if (k2 == 'sipDisplayName' and v2 ~= '' and v2 ~= self.properties.sipDisplayName) then
				sipDisplayName = v2
			end
			if (k2 == 'id' and v2 ~= '') then
				id = v2
			end
		end

		if (id ~= '' and sipDisplayName ~= '') then
			table.insert(usersList, {["value"]=tostring(id), ["label"]=sipDisplayName})
	    end

		sipDisplayName = ''
		id = ''
	end

	return usersList
end

--! Calculate codes lists and return it to select control that is connected with this method.
--!@return codes list. 
function Mobotix:getVoipCodesList(arg)
	local userCodes = {}
	table.insert(userCodes, {["value"]='---', ["label"]='---'})

	local iString

	for i=1,self.availablePhoneBookEntries do
		iString = tostring(i)
		table.insert(userCodes, {["value"]=iString, ["label"]=iString})
	end

	return userCodes
end

--!Temporary method to handle action method (open Lock) from mini plugin view;
function Mobotix:openLock( ... )
	self:openDoor()
end








