require('common.device')
require('netatmoWeatherStation')
require('netatmoDevice')
require('netatmoSensor')
require('net.HTTPClient')

class 'NetatmoController' (Device)

--! Communication header. Common for all requests.
local  headers =
                  {
                    ["Content-Type"] = "application/x-www-form-urlencoded"
                  }

--! List of available setpoint modes.
local setpoint_mode={
    [1]="program",
    [2]="away",
    [3]="hg",
    [4]="manual",
    [5]="off",
    [6]="max"
                    }

--! List of functions for device initialization.
local NetAtmoDeviceList={
    ["NAMain"]=function(self,param) self:prepareBaseStation(param)  end,                  -- the base station
    ["NAPlug"]=function(self,param) self:prepareRelayPlug(param)  end,               --for the thermostat relay/plug
}

--! List of functions triggering polling.
local NetAtmoDeviceListGET={
    ["NAMain"]=function(self,param) self:getBaseStationMeasures(param)  end,                  -- the base station
    ["NAPlug"]=function(self,param) self:getPlug(param)  end,               --for the thermostat relay/plug

    ["NAModule1"]=function(self,param) self:getOutdoorModuleMeasures(param)  end,                  --the outdoor module
    ["NAModule4"]=function(self,param) self:getAdditionalIndoor(param)  end,               --the additionnal indoor module
    ["NAModule3"]=function(self,param) self:getRainGauge(param)  end,               --the rain gauge module
    ["NATherm1"]=function(self,param) self:getThermostat(param)  end               --the thermostat module
}

--! Initializes Nest Account plugin object.
--!@param id: Id of the device.
function NetatmoController:__init(id)
  Device.__init(self, id)

        self.pollingTime = self.properties.polling_time * 1000
        self.weatherStations={}
        self.thermostats={}

  self.http = net.HTTPClient({ timeout = 3000 })

        self:connectToNetatmo()
        self.connected=false
end

--! Gets authorization tokens and establishes connection to Netatmo server.
function NetatmoController:connectToNetatmo()

    self.s_client_id = self.properties.client_id
    self.s_client_secret = self.properties.client_secret
    self.s_username = self.properties.username
    self.s_password = self.properties.password

        p_url="https://api.netatmo.net/oauth2/token"
        p_body="grant_type=password&client_id=" .. self.s_client_id .."&client_secret=" .. self.s_client_secret .. "&username=" .. self.s_username .. "&password=" .. self.s_password .."&scope=read_station read_thermostat write_thermostat"

    self.http:request(p_url, {
          options = {
                      headers= headers ,
                      method = 'POST',
                      data=p_body
                 },
          success = function(param)
                local ok, resp = pcall(json.decode, param.data)
                if ok and resp then
                    if resp.error ~= nil then
                      if(resp.error=='invalid_grant' or resp.error == 'invalid_client') then
                        self:updateProperty('ui.Connection_Status_Label.caption',"Invalid username or password")
                      elseif resp.error == 'invalid_request' then
                        self:updateProperty('ui.Connection_Status_Label.caption', "Invalid request")
                      end
    
                      self.connected=false  
                      if (self.timeoutId) then
                        clearTimeout(self.timeoutId)
                      end
    
                      self.timeoutId = setTimeout(function() self:connectToNetatmo() end, 60000) -- Connection lost! Wait a minute and try again.                 
                      return
                    end
    
                    if param.status==200 then
                      self.access_token = resp.access_token
                      self.refresh_token = resp.refresh_token
                      self.expires_in = resp.expires_in

                      if (self.refreshTimeoutId) then
                          clearTimeout(self.refreshTimeoutId)
                      end
                      self.refreshTimeoutId = setTimeout(function() self:refreshTokens() end, self.expires_in*1000)

                      self:updateProperty('ui.Connection_Status_Label.caption',"Your account is connected.")
                      self.connected=true
    
                      self:getStations()
                      self:getThermostats()
                    else
                      self:updateProperty('ui.Connection_Status_Label.caption', "Not connected")
                      self.connected=false
    
                      if (self.timeoutId) then
                        clearTimeout(self.timeoutId)
                      end
    
                      self.timeoutId = setTimeout(function() self:connectToNetatmo() end, 60000) -- Connection lost! Wait a minute and try again.
                    end
                else
                  print('Json parsing error.')
                  self:updateProperty('ui.Connection_Status_Label.caption',"Not connected.")
                  self.connected=false

                  if (self.timeoutId) then
                     clearTimeout(self.timeoutId)
                  end

                  self.timeoutId = setTimeout(function() self:connectToNetatmo() end, 60000) -- Connection lost! Wait a minute and try again.
                end            
          end,
          error = function(error)
              self:tryToReconnect()
          end
        })
end

--! Refreshes expired authorization token.
function NetatmoController:refreshTokens()
  print("Netatmo: Refresh token")
  local p_url='https://api.netatmo.net/oauth2/token'
  local p_body="grant_type=refresh_token&refresh_token=" .. self.refresh_token .. "&client_id=" .. self.s_client_id .. "&client_secret=" .. self.s_client_secret
        self.http:request(p_url, {
            options = {
                        headers= headers ,
                        method = 'POST',
                        data=p_body
                   },
            success = function(param)
                    local ok, resp = pcall(json.decode, param.data)
                    if ok and resp then
                        if param.status==200 and resp.error == nil then
                            print("Netatmo: Token successfuly refreshed")
                            self.access_token = resp.access_token
                            self.refresh_token = resp.refresh_token
                            self.expires_in = resp.expires_in
                            self:updateProperty('ui.Connection_Status_Label.caption',"Your account is connected.")
                            self.connected=true

                            if (self.refreshTimeoutId) then
                                clearTimeout(self.refreshTimeoutId)
                            end
                            self.refreshTimeoutId =  setTimeout(function() self:refreshTokens() end, self.expires_in*1000)   
                        end
                    else
                        print('Json parsing error.')
                        if (self.refreshTimeoutId) then
                            clearTimeout(self.refreshTimeoutId)
                        end
                        self.refreshTimeoutId = setTimeout(function() self:refreshTokens() end, self.pollingTime) -- Failed to refresh token, wait some time and try again                    
                    end             
            end,
            error = function(error)
                if (self.refreshTimeoutId) then
                    clearTimeout(self.refreshTimeoutId)
                end
                self.refreshTimeoutId = setTimeout(function() self:refreshTokens() end, self.pollingTime) -- Failed to refresh token, wait some time and try again
            end
          })
end

--! Gets a list of all available stations and their wireless modules.
function NetatmoController:getStations()
    local p_url="https://api.netatmo.net/api/devicelist"
    local p_body="access_token=" .. self.access_token .. "&app_type=app_station"

    self.http:request(p_url, {
            options = {
                        headers= headers ,
                        method = 'POST',
                        data=p_body
                   },
            success = function(param)
                        local ok, resp = pcall(json.decode, param.data)
                        if ok and resp then
                            if resp.error and resp.error.message and resp.error.message == 'User usage reached' then
                                self:updateProperty('ui.Connection_Status_Label.caption',"User usage reached")
                                setTimeout(function() self:getStations() end, 5*60*1000) -- Failed to communicate with netatmo API, wait 5 min and try again
                                return
                            end

                            if param.status ~= 403 then
                                for key,value in pairs(resp.body.devices) do
                                    local callback=NetAtmoDeviceList[resp.body.devices[key].type]
                                    if (callback) then
                                        callback(self,resp.body.devices[key])
                                    end
                                end

                                for key,value in pairs(resp.body.modules) do
                                    if (self.weatherStations[resp.body.modules[key].main_device]) then

                                        self.weatherStations[resp.body.modules[key].main_device]:addModule(resp.body.modules[key])

                                        local callback= NetAtmoDeviceListGET[resp.body.modules[key].type]
                                        if (callback) then
                                            callback(self,resp.body.modules[key])
                                        end
                                    end
                                end
                            end
                        else
                            print('Json parsing error.')                            
                            setTimeout(function() self:getStations() end, 5*60*1000) -- Failed to communicate with netatmo API wait 5 min and try again. 
                            return
                        end  
                    end,
                error = function(error)
                   setTimeout(function() self:getStations() end, 5*60*1000) -- Failed to communicate with netatmo API wait 5 min and try again. 
            end
          })
end

--! Gets a lis of all available thermostats and their wireless modules.
function NetatmoController:getThermostats()
    local p_url="https://api.netatmo.net/api/devicelist"
    local p_body="access_token=" .. self.access_token .. "&app_type=app_thermostat"

    self.http:request(p_url, {
            options = {
                        headers= headers ,
                        method = 'POST',
                        data=p_body
                   },
            success = function(param)
                    local ok, resp = pcall(json.decode, param.data)
                    if ok and resp then
                        if resp.error and resp.error.message and resp.error.message == 'User usage reached' then
                            self:updateProperty('ui.Connection_Status_Label.caption',"User usage reached")

                            setTimeout(function() self:getThermostats() end, 5*60*1000) -- Failed to communicate with netatmo API, wait 5 min and try again. 
                            return
                        end

                        if param.status == 200 then
                            for key,value in pairs(resp.body.devices) do
                                local callback=NetAtmoDeviceList[resp.body.devices[key].type]
                                if (callback) then
                                    callback(self,resp.body.devices[key])
                                end
                            end
                            for key,value in pairs(resp.body.modules) do
                                if (self.thermostats[resp.body.modules[key].main_device]) then

                                    self.thermostats[resp.body.modules[key].main_device]:addModule(resp.body.modules[key])

                                    local callback = NetAtmoDeviceListGET[resp.body.modules[key].type]
                                    if (callback) then
                                        callback(self,resp.body.modules[key])
                                    end
                                end
                            end
                        end                         
                    else
                        print('Json parsing error.')
                        setTimeout(function() self:getThermostats() end, 5*60*1000) -- Failed to communicate with netatmo API, wait 5 min and try again. 
                        return
                    end
            end,
                error = function(error)
                    setTimeout(function() self:getThermostats() end, 5*60*1000) -- Failed to communicate with netatmo API, wait 5 min and try again. 
            end
          })
end

--! Triggers initialization process for a Base Station of a Weather Station and starts polling.
--!@param device: Id of a device.
function NetatmoController:prepareBaseStation(device)

    local childDevices = plugin.getChildDevices(self.id)

    for _,dev in pairs(childDevices) do
        if (dev.properties.netatmoID == device._id) then
             self.weatherStations[device._id]=NetatmoWeatherStation(self, dev, device)
        end
    end

    if self.weatherStations[device._id]==nil then
        local dev = plugin.createChildDevice(self.id,'com.fibaro.netatmoWeatherStation', 'Netatmo Weather Sation')
        self.weatherStations[device._id]=NetatmoWeatherStation(self, dev, device)
    end


    self:getBaseStationMeasures(device)
end

--! Triggers initialization process for a Plug/Relay and starts polling.
--!@param device: Id of a device.
function NetatmoController:prepareRelayPlug(device)

    local childDevices = plugin.getChildDevices(self.id)

    for _,dev in pairs(childDevices) do
        if (dev.properties.netatmoID == device._id) then
            self.thermostats[device._id]=NetatmoDevice(self, dev)
        end
    end

    if self.thermostats[device._id]==nil then
        local dev = plugin.createChildDevice(self.id,'com.fibaro.netatmoDevice', 'netatmo Plug/Relay')
        self.thermostats[device._id]=NetatmoDevice(self, dev)
        self.thermostats[device._id]:updateProperty('netatmoID', device._id)
        self.thermostats[device._id]:updateProperty('netatmoType', device.type)
    end
end

--! Gets measurements of a Base Stations Indoor Module.
--!@param device: Device's defining structure.
function NetatmoController:getBaseStationMeasures(device)
    local p_url="https://api.netatmo.net/api/getmeasure"
    local p_body="access_token=" ..self.access_token .."&device_id=" .. device._id .. "&scale=max&type=Temperature,CO2,Humidity,Pressure,Noise&date_end=last"

    self.http:request(p_url, {
            options = {
                        headers= headers ,
                        method = 'POST',
                        data=p_body
                   },
            success = function(param)
                    local ok, resp = pcall(json.decode, param.data)
                    if ok and resp then
                        if resp.status == "ok" then
                            self.weatherStations[device._id]:updateBaseStationValues(resp)
                        else
                            print("Netatmo Error: getBaseStationMeasures response status: ",resp.status )
                        end
                    else
                        print('Json parsing error.')
                    end
                    if self.connected then
                        setTimeout(function() self:getBaseStationMeasures(device) end, self.pollingTime) 
                    else 
                        self:tryToReconnect() 
                    end
            end,
            error = function(error)
                setTimeout(function() self:getBaseStationMeasures(device) end, self.pollingTime) 
            end
          })
end

--! Gets measurements of a Base Stations Outdoor Module.
--!@param module: Module's defining structure.
function NetatmoController:getOutdoorModuleMeasures(module)
    local p_url="https://api.netatmo.net/api/getmeasure"
    local p_body="access_token=" ..self.access_token .."&device_id=" .. module.main_device .. "&module_id=" .. module._id .. "&scale=max&type=Temperature,Humidity&date_end=last"

    self.http:request(p_url, {
            options = {
                        headers= headers ,
                        method = 'POST',
                        data=p_body
                   },
            success = function(param)
                  local ok, resp = pcall(json.decode, param.data)
                  if ok and resp then
                      if resp.status == "ok" then
                          self.weatherStations[module.main_device]:OutdoorModuleMeasures(resp,module._id)
                      else
                          print("Netatmo Error: getOutdoorModuleMeasures response status: ",resp.status )
                      end
                  else
                      print('Json parsing error.')
                  end

                  if self.connected then
                      setTimeout(function() self:getOutdoorModuleMeasures(module) end, self.pollingTime)
                  else 
                      self:tryToReconnect() 
                  end
            end,
            error = function(error)
                setTimeout(function() self:getOutdoorModuleMeasures(module) end, self.pollingTime)
            end
          })
end

function NetatmoController:getAdditionalIndoor(module)
    local p_url="https://api.netatmo.net/api/getmeasure"
    local p_body="access_token=" ..self.access_token .."&device_id=" .. module.main_device .. "&module_id=" .. module._id .. "&scale=max&type=Temperature,Humidity,CO2&date_end=last"

    self.http:request(p_url, {
            options = {
                        headers= headers ,
                        method = 'POST',
                        data=p_body
                   },
            success = function(param)
                        local ok, resp = pcall(json.decode, param.data)
                        if ok and resp then
                            if resp.status == "ok" then
                                self.weatherStations[module.main_device]:AdditionalIndoorMeasures(resp,module._id)
                            else
                                print("Netatmo Error: getAdditionalIndoor response status: ",resp.status )
                            end    
                        else
                            print('Json parsing error.')  
                        end
                        if self.connected then
                            setTimeout(function() self:getAdditionalIndoor(module) end, self.pollingTime)
                        else 
                            self:tryToReconnect() 
                        end
                    end,
            error = function(error)
                setTimeout(function() self:getAdditionalIndoor(module) end, self.pollingTime)
            end
          })
end

function NetatmoController:getRainGauge(module)
    local p_url="https://api.netatmo.net/api/getmeasure"
    local p_body="access_token=" ..self.access_token .."&device_id=" .. module.main_device .. "&module_id=" .. module._id .. "&scale=max&type=Rain&date_end=last"

    self.http:request(p_url, {
            options = {
                        headers= headers ,
                        method = 'POST',
                        data=p_body
                   },
            success = function(param)
                    local ok, resp = pcall(json.decode, param.data)
                    if ok and resp then
                        if resp.status == "ok" then
                                  self.weatherStations[module.main_device]:RainGaugeMeasures(resp,module._id)
                        else
                            print("Netatmo Error: getRainGauge response status: ",resp.status )
                        end
                    else
                        print('Json parsing error.')
                    end
                    if self.connected then
                        setTimeout(function() self:getRainGauge(module) end, self.pollingTime)
                    else 
                        self:tryToReconnect() 
                    end
                  end,
            error = function(error)
                  setTimeout(function() self:getRainGauge(module) end, self.pollingTime)
            end
          })
end

--! Gets current state of a thermostat.
--!@param module: Module's defining structure.
function NetatmoController:getThermostat(module)
    local p_url='https://api.netatmo.net/api/getthermstate'
    local p_body="access_token=" ..self.access_token .."&device_id=".. module.main_device .. "&module_id="..module._id
             self.http:request(p_url, {
               options = {
                           headers= headers ,
                           method = 'POST',
                           data=p_body
                      },
               success = function(param)
                    local ok, resp = pcall(json.decode, param.data)
                    if ok and resp then
                        if resp.status == "ok" then
                            self.thermostats[module.main_device]:updateState(resp,module._id)  
                        else
                          print("Netatmo Error: getThermostat response status: ",resp.status )
                        end
                    else
                        print('Json parsing error.')
                    end
                    if self.connected then
                        setTimeout(function() self:getThermostat(module) end, self.pollingTime)
                    else 
                        self:tryToReconnect() 
                    end
               end,
               error = function(error)
                  setTimeout(function() self:getThermostat(module) end, self.pollingTime)
               end
             })
end

--! Modifies UI values in RT and if there are no changes for 2 seconds sends settings to a device.
--!@param increase: Boolean value determining weter to increase or decrease currently displayed target value.
--!@param devId: Id of a thermostat.
function NetatmoController:setTemperature(increase,devId)

    for _,dev in pairs(self.thermostats) do
         for _,mod in pairs(dev.childDevices) do
            if (mod.id == devId) then
                 main_device=dev.properties.netatmoID
                 module_id=mod.properties.netatmoID
            end
        end
    end

    if main_device~=nil and  module_id~=nil then
        local currentSetpoint = self.thermostats[main_device].childDevices[module_id].properties.currentSetpointTemp
        local Setpoint = self.thermostats[main_device].childDevices[module_id].properties.setpointTemp
        local setpointEndtime = self.thermostats[main_device].childDevices[module_id].properties.setpointEndtime

        if (currentSetpoint == '' or currentSetpoint == nil) then
            print 'must wait for Setpoint refresh'
            return
        end

        if (Setpoint == '' or Setpoint == nil or Setpoint == '---') then
            print 'updating Setpoint'
            Setpoint = currentSetpoint
        end

        if (setpointEndtime == '' or setpointEndtime == nil or setpointEndtime==0 or setpointEndtime<os.time(os.date("*t"))) then
            self.thermostats[main_device].childDevices[module_id]:updateProperty('ui.Setpoint_Endtime_Slider.value',1)
            setpointEndtime = os.time(os.date("*t"))+60*60
        end

        local temp = Setpoint


        if increase then
            temp = temp+0.5
        else
            temp = temp-0.5
        end

         if (temp < 5) then
            print("unsuported value of setPoint: "..temp)
            temp = 5
        elseif (temp > 30) then
            print("unsuported value of setPoint: "..temp)
            temp = 30
        end

        if self.timeoutId ~= nil then clearTimeout(self.timeoutId) end
        self.timeoutId = setTimeout(function() self:setThermostatTemperature( main_device, module_id, temp, setpointEndtime) end,2000)

    local location=api.get("/settings/location")
    if location.temperatureUnit == 'F' then
       self.thermostats[main_device].childDevices[module_id]:updateProperty('ui.Setpoint_Value_Label.caption', unitsConverter.celsius2farenheit(temp)..'°F')
    else
       self.thermostats[main_device].childDevices[module_id]:updateProperty('ui.Setpoint_Value_Label.caption', temp..'°C')
    end
       self.thermostats[main_device].childDevices[module_id]:updateProperty('setpointTemp', temp)
       self.thermostats[main_device].childDevices[module_id]:updateProperty('setpointEndtime',setpointEndtime)
    end
end

--! Sets end time for current target temperature.
--!@param devId: Id of a thermostat.
--!@param endtime: Timestamp defining end time of current target temperature.
function NetatmoController:setEndpoint(devId, endtime)

    for _,dev in pairs(self.thermostats) do
         for _,mod in pairs(dev.childDevices) do
            if (mod.id == devId) then
                 main_device=dev.properties.netatmoID
                 module_id=mod.properties.netatmoID
            end
        end
    end

    if endtime ~= nil then
        self.thermostats[main_device].childDevices[module_id]:updateProperty('ui.Setpoint_Endtime_Slider.value', endtime)
    else
        print("Empty response body. Polling limit reached or server malfunction.")
    end

    if main_device~=nil and  module_id~=nil then

        local currentSetpoint = self.thermostats[main_device].childDevices[module_id].properties.currentSetpointTemp
        local Setpoint = self.thermostats[main_device].childDevices[module_id].properties.setpointTemp

        if (currentSetpoint == '' or currentSetpoint == nil) then
            print 'must wait for Setpoint refresh'
            return
        end

        if (Setpoint == '' or Setpoint == nil or Setpoint == '---') then
            print 'updating coolSetpoint'
            Setpoint = currentSetpoint
        end

        if endtime == 0 then
            self:changeMode(devId,1)
            self.thermostats[main_device].childDevices[module_id]:updateProperty('setpointEndtime',0)
            return
        end

        temp=Setpoint
        endpoint=os.time(os.date("*t"))+endtime*60*60

        if self.timeoutId ~= nil then clearTimeout(self.timeoutId) end
        self.timeoutId = setTimeout(function() self:setThermostatTemperature( main_device, module_id, temp,endpoint) end,2000)

        self.thermostats[main_device].childDevices[module_id]:updateProperty('setpointEndtime',endpoint)
    end
end

--! Sets end time for current target temperature.
--!@param main_device: Id of a Plug/Relay.
--!@param module_id: Id of a Thermostat.
--!@param temp: Target temperature.
--!@param endpoint: Timestamp defining end time of current target temperature.
function NetatmoController:setThermostatTemperature( main_device, module_id, temp, endpoint)

        local p_url='https://api.netatmo.net/api/setthermpoint?access_token=' ..self.access_token .."&device_id=".. main_device .. "&module_id="..module_id .. "&setpoint_mode=manual&need_sync=true&&setpoint_temp="..temp.."&setpoint_endtime="..endpoint

                         self.http:request(p_url, {
                           options = {
                                       headers= headers ,
                                       method = 'GET'
                                  },
                           success = function(param)
                                   setTimeout(function() self:getSingleThermostatState( main_device, module_id) end,5000)
                           end,
                           error = function(error)
                           end
                         })
end

--! Gets current state of a thermostat (singe time).
--!@param main_device: Id of a Plug/Relay.
--!@param module_id: Id of a Thermostat.
function NetatmoController:getSingleThermostatState( main_device, module_id)
    local p_url='https://api.netatmo.net/api/getthermstate'
    local p_body="access_token=" ..self.access_token .."&device_id=".. main_device .. "&module_id="..module_id

        self.http:request(p_url, {
         options = {
                     headers= headers ,
                     method = 'POST',
                     data=p_body
                },
         success = function(param)
                local ok, resp = pcall(json.decode, param.data)
                if ok and resp then
                  self.thermostats[main_device]:updateState(resp,module_id)
                else
                  print('Json parsing error.')
                end
         end,
         error = function(error)
         end
        })
end

--! Changes currently selected mode of a thermostat.
--!@param devId: Id of a device.
--!@param mode: Integer value defining mode (see mode list at the top).
function NetatmoController:changeMode(devId,mode)

    for _,dev in pairs(self.thermostats) do
         for _,mod in pairs(dev.childDevices) do
            if (mod.id == devId) then
                 main_device=dev.properties.netatmoID
                 module_id=mod.properties.netatmoID
            end
        end
    end

    if main_device~=nil and  module_id~=nil then
        local p_url='https://api.netatmo.net/api/setthermpoint?setpoint_mode='..setpoint_mode[mode]..'&need_sync=true&device_id='.. main_device .. '&module_id='..module_id .. '&&access_token=' ..self.access_token
                 self.http:request(p_url, {
                   options = {
                               headers= headers ,
                               method = 'POST'
                          },
                   success = function(param)
                    if mode == 2 then
                        self.thermostats[main_device].childDevices[module_id]:updateProperty('ui.Away_Switch.value', true)
                        self.thermostats[main_device].childDevices[module_id]:updateProperty('ui.HG_Switch.value', false)
                    elseif mode == 3 then
                        self.thermostats[main_device].childDevices[module_id]:updateProperty('ui.HG_Switch.value', true)
                        self.thermostats[main_device].childDevices[module_id]:updateProperty('ui.Away_Switch.value', false)
                    else
                        self.thermostats[main_device].childDevices[module_id]:updateProperty('ui.HG_Switch.value', false)
                        self.thermostats[main_device].childDevices[module_id]:updateProperty('ui.Away_Switch.value', false)
                    end
                   end,
                   error = function(error)
                   end
                 })
    end
end

--! In case of error in communication, connection will be resumed after 10 min
function NetatmoController:tryToReconnect()
    self.connected = false
    if (self.timeoutId) then
        clearTimeout(self.timeoutId)
    end
    self.timeoutId = setTimeout(function() self:connectToNetatmo() end, 10*60*1000)
end

--! Reloads a list of devices and modules.
function NetatmoController:addNewDevices()
    if self.connected then
        self:getStations()
        self:getThermostats()
    else
        print("Not Connected!")
        self:updateProperty('ui.Connection_Status_Label.caption',"Not connected.")
    end
end