require('common.device')
require('common.unitsConverter')
require('netatmoSensor')

--! List of functions for specific module initialization.
local NetAtmoDeviceList={
    ["NAModule1"]=function(self,param) self:prepareOutdoor(param)  end,                  --the outdoor module
    ["NAModule4"]=function(self,param) self:prepareAdditionalIndoor(param)  end,               --the additionnal indoor module
    ["NAModule3"]=function(self,param) self:prepareRainGauge(param)  end,               --the rain gauge module
}

--! Computes number of elements in a table.
--!@param tab: A table.
function tableSize(tab)
        local count = 0
        for _,_ in pairs(tab) do
                count = count + 1
        end

        return count
end

class 'NetatmoWeatherStation' (Device)

--! Initializes Netatmo Weather Station object, finds and initializes children.
--!@param parent: Id of a parent.
--!@param id: Id of the device.
--!@param res: Current measures from a Weather Station.
function NetatmoWeatherStation:__init(parent, dev, device)
        Device.__init(self, dev)

    self.device=device
    self:updateProperty('netatmoID', device._id)
    self:updateProperty('netatmoType', device.type)
    self.childTable={}
    self.childDevices={}

    self:setVisible(false)

    local childDevices = plugin.getChildDevices(self.id)
    if tableSize(childDevices)>0 then
        self.childTable= self.properties.childTable
        for _,devs in pairs(childDevices) do
                if (devs.type == 'com.fibaro.multilevelSensor') or (devs.type == 'com.fibaro.temperatureSensor') or (devs.type == 'com.fibaro.humiditySensor') then
                    if self.childTable[tostring(devs.id)] == nil then
                        print("Device not recognized.")
                    else
                        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Adding Device")
                        self.childDevices[tostring(self.childTable[tostring(devs.id)])] = NetatmoSensor(self, devs)
                    end
                end
        end
    end
    if tableSize(self.childDevices)==0 then
        self:initializeParams(device)
    end
end

--! Initializes Netatmo Weather Station children sensors.
function NetatmoWeatherStation:initializeParams()

    local dev = plugin.createChildDevice(self.id,'com.fibaro.temperatureSensor', 'netatmo Indoor Temperature')
    self.childDevices[self.device._id..".Temp"]=NetatmoSensor(self, dev)
    self.childTable[self.childDevices[self.device._id..".Temp"].id]=self.device._id..".Temp"

    dev = plugin.createChildDevice(self.id,'com.fibaro.multilevelSensor', 'netatmo Indoor CO2')
    self.childDevices[self.device._id..".CO2"]=NetatmoSensor(self, dev)
    self.childTable[self.childDevices[self.device._id..".CO2"].id]=self.device._id..".CO2"

    dev = plugin.createChildDevice(self.id,'com.fibaro.humiditySensor', 'netatmo Indoor Humidity')
    self.childDevices[self.device._id..".Hum"]=NetatmoSensor(self, dev)
    self.childTable[self.childDevices[self.device._id..".Hum"].id]=self.device._id..".Hum"

    dev = plugin.createChildDevice(self.id,'com.fibaro.multilevelSensor', 'netatmo Indoor Pressure')
    self.childDevices[self.device._id..".Press"]=NetatmoSensor(self, dev)
    self.childTable[self.childDevices[self.device._id..".Press"].id]=self.device._id..".Press"

    dev = plugin.createChildDevice(self.id,'com.fibaro.multilevelSensor', 'netatmo Indoor Noise')
    self.childDevices[self.device._id..".Noise"]=NetatmoSensor(self, dev)
    self.childTable[self.childDevices[self.device._id..".Noise"].id]=self.device._id..".Noise"

    self:updateProperty('childTable', self.childTable)
end

--! Adds wireless module to a Netatmo Weather Station object.
--!@param module: Module's defining structure.
function NetatmoWeatherStation:addModule(module)
    local callback= NetAtmoDeviceList[module.type]
    if (callback) then
       callback(self, module)
    end

    print(json.encode(self.childTable))
    self:updateProperty('childTable', self.childTable)
end

--! Adds wireless outdoor module to a Netatmo Weather Station object.
--!@param module: Module's defining structure.
function NetatmoWeatherStation:prepareOutdoor(module)
    if self.childDevices[module._id..".Temp"]==nil then
        local dev = plugin.createChildDevice(self.id,'com.fibaro.temperatureSensor', 'netatmo Outdoor Temperature')
        self.childDevices[module._id..".Temp"]=NetatmoSensor(self, dev)
        self.childTable[self.childDevices[module._id..".Temp"].id]=module._id..".Temp"
    end

    if self.childDevices[module._id..".Hum"]==nil then
        local dev = plugin.createChildDevice(self.id,'com.fibaro.humiditySensor', 'netatmo Outdoor Humidity')
        self.childDevices[module._id..".Hum"]=NetatmoSensor(self, dev)
        self.childTable[self.childDevices[module._id..".Hum"].id]=module._id..".Hum"
    end
end

--! Adds additional wireless indoor module to a Netatmo Weather Station object.
--!@param module: Module's defining structure.
function NetatmoWeatherStation:prepareAdditionalIndoor(module)
    if self.childDevices[module._id..".Temp"]==nil then
        local dev = plugin.createChildDevice(self.id,'com.fibaro.temperatureSensor', 'netatmo Indoor Temperature')
        self.childDevices[module._id..".Temp"]=NetatmoSensor(self, dev)
        self.childTable[self.childDevices[module._id..".Temp"].id]=module._id..".Temp"
    end

    if self.childDevices[module._id..".Hum"]==nil then
        local dev = plugin.createChildDevice(self.id,'com.fibaro.humiditySensor', 'netatmo Indoor Humidity')
        self.childDevices[module._id..".Hum"]=NetatmoSensor(self, dev)
        self.childTable[self.childDevices[module._id..".Hum"].id]=module._id..".Hum"
    end

    if self.childDevices[module._id..".CO2"]==nil then
        local dev = plugin.createChildDevice(self.id,'com.fibaro.multilevelSensor', 'netatmo Indoor CO2')
        self.childDevices[module._id..".CO2"]=NetatmoSensor(self, dev)
        self.childTable[self.childDevices[module._id..".CO2"].id]=module._id..".CO2"
    end
end

--! Adds wireless rain gauge module to a Netatmo Weather Station object.
--!@param module: Module's defining structure.
function NetatmoWeatherStation:prepareRainGauge(module)
    if self.childDevices[module._id..".Rain"]==nil then
        local dev = plugin.createChildDevice(self.id,'com.fibaro.multilevelSensor', 'netatmo Rain Gauge')
        self.childDevices[module._id..".Rain"]=NetatmoSensor(self, dev)
        self.childTable[self.childDevices[module._id..".Rain"].id]=module._id..".Rain"
    end
end

--! Updates all Indoor module values.
--!@param reponsepost: Response from a server with all current measures.
function NetatmoWeatherStation:updateBaseStationValues(reponsepost)

    temperature_indoor =reponsepost.body[1].value[1][1]
    co2 =reponsepost.body[1].value[1][2]
    humidity_indoor =reponsepost.body[1].value[1][3]
    pressure =reponsepost.body[1].value[1][4]
    noise0 =reponsepost.body[1].value[1][5]

    self.childDevices[self.device._id..".Temp"]:setValue(temperature_indoor)
    self.childDevices[self.device._id..".CO2"]:setValue(co2)
    self.childDevices[self.device._id..".Hum"]:setValue(humidity_indoor)
    self.childDevices[self.device._id..".Press"]:setValue(pressure)
    self.childDevices[self.device._id..".Noise"]:setValue(noise0)
    self:updateProperty('lastUpdated', reponsepost.body[1].beg_time)

end

--! Updates all Outdoor module values.
--!@param reponsepost: Response from a server with all current measures.
--!@param device_id: Module's Id.
function NetatmoWeatherStation:OutdoorModuleMeasures(reponsepost,device_id)

    if reponsepost.body[1].value[1][1] ~= nil then
        temperature_outdoor =reponsepost.body[1].value[1][1]
        humidity_outdoor =reponsepost.body[1].value[1][2]

        self.childDevices[device_id..".Temp"]:setValue(temperature_outdoor)
        self.childDevices[device_id..".Hum"]:setValue(humidity_outdoor)
        self:updateProperty('lastOutdoorUpdated', reponsepost.body[1].beg_time)
    else
        print("Empty response body. Polling limit reached or server malfunction.")
    end

end

function NetatmoWeatherStation:AdditionalIndoorMeasures(reponsepost,device_id)

    if reponsepost.body[1].value[1][1] ~= nil then
        temperature_indoor =reponsepost.body[1].value[1][1]
        humidity_indoor =reponsepost.body[1].value[1][2]
        co2_indoor = reponsepost.body[1].value[1][3]

        self.childDevices[device_id..".Temp"]:setValue(temperature_indoor)
        self.childDevices[device_id..".Hum"]:setValue(humidity_indoor)
        self.childDevices[device_id..".CO2"]:setValue(co2_indoor)
    else
        print("Empty response body. Polling limit reached or server malfunction.")
    end

end

function NetatmoWeatherStation:RainGaugeMeasures(reponsepost,device_id)

    if reponsepost.body[1].value[1][1] ~= nil then
        rain_measure =reponsepost.body[1].value[1][1]

        self.childDevices[device_id..".Rain"]:setValue(rain_measure)
    else
        print("Empty response body. Polling limit reached or server malfunction.")
    end

end
