require('NuVoZone')
require('net.TCPSocket')
require('net.UDPSocket')
require('common.device')
require('json')

class 'Nuvo' (Device)

function Nuvo:__init(id)
	Device.__init(self, id)	

	self.uiBinding = {
		["play_Button"] =	function() self:playAll() end,
		["pause_Button"] =	function() self:pauseAll() end,
		["next_Button"] =	function() self:nextAll() end,
		["prev_Button"] =	function() self:previousAll() end,
		["mute_Switch"] = 	function(eventArgs)
								self:muteAll(eventArgs.values[1])
							end,
		["volume_Slider"] = function(eventArgs) self:setGlobalValume(eventArgs.values[1]) end,
	}

	self.zones = {}
    local childDevices = plugin.getChildDevices(self.id)

    for _,dev in pairs(childDevices) 
    do
        if (dev.type == 'com.fibaro.nuvoZone') 
        then
            self.zones[dev.id] = NuVoZone(self, dev, dev.name)
        end
    end

    self:synchronizeZones(true)
	self:discover()	
	self:refreshGlobalVolume()
end

function Nuvo:checkIfZoneExists(ip)
	for _, zone in pairs(self.zones)
	do
		if zone.properties.ip == ip
		then
			return true
		end
	end

	return false
end

function Nuvo:resetZonesTable()
	local zonesTable = {}

	for _, zone in pairs(self.zones) do
		local tuple = {columns = {zone.name, zone.properties.ip, 'offline'}}
		table.insert(zonesTable, tuple)
	end

	self:updateProperty('zones', zonesTable)
end

function Nuvo:updateZonesTable(onlineZones)
	self:resetZonesTable()

	local zonesTable = {}

	for _, zone in pairs(self.zones) do
		local tuple = {}
		if onlineZones then
			if self:isZoneInOnlineZones(onlineZones, zone) then
				tuple = {columns = {zone.name, zone.properties.ip, 'online'}}
			else
				tuple = {columns = {zone.name, zone.properties.ip, 'offline'}}
			end
		else
			tuple = {columns = {zone.name, zone.properties.ip, 'online'}}
		end

		table.insert(zonesTable, tuple)
	end

	self:updateProperty('zones', zonesTable)
end

function Nuvo:updateZonesTableWithZone(paramZone)
	local status = ''
	if paramZone.properties.dead == true then 
		status = 'offline'
	else
		status = 'online'
	end

    for id,zone in pairs(self.properties.zones) do
    	if tostring(zone.columns[2]) == tostring(paramZone.ip) then
	    	self.properties.zones[id].columns[3] = status
	    	self:updateProperty('zones', self.properties.zones)
    	end
    end
end

function Nuvo:isZoneInOnlineZones(onlineZones, zone)
	if onlineZones and zone then
		for _,onlineZone in pairs(onlineZones) do
			if onlineZone.ip == zone.ip then
				return true
			end
		end
	end

	return false
end

function Nuvo:synchronizeZones(loop)
    local changes = false

    for _,zone in pairs(self.zones) 
    do
    	if plugin.getDevice(zone.id) == nil
    	then
    		self.zones[zone.id] = nil
    		changes = true
    	end
    end	

    if changes
    then
    	self:updateZonesTable()
    end

    if loop
    then
    	setTimeout(function() self:synchronizeZones(true) end
    	, 10000)
	end
end

function Nuvo:printDevices(response)
	local discoveredZones = {}
	if response then
		for _, device in pairs(response) do
			if self:checkIfZoneExists(device.ip) ~= true then
				local dev = plugin.createChildDevice(self.id, 'com.fibaro.nuvoZone', 'new zone', {ip = device.ip})
				local devObject = NuVoZone(self, dev, dev.name)

				self.zones[dev.id] = devObject
			end
		end

		self:updateZonesTable(response)
	end
end

-- method to discover nuvo devices in local networkW
function Nuvo:discover()
	--create UDP socekt 
	local socket = net.UDPSocket({broadcast = true, timeout = 2000})
	devicesTable = {}

	--prepare payload for SSDP discover datagram 
	local data = table.concat({'M-SEARCH * HTTP/1.1',
	        'HOST: 239.255.255.250:1900',
	        'MAN: "ssdp:discover"',
	        'ST: urn:schemas-nuvotechnologies-com:device:Zone:1','MX: 3','',''}, '\r\n')
	--send discover
	socket:sendTo(data, "239.255.255.250", 1900,{
            success = function()
                self:discoverReadData(socket, devicesTable)
            end,
            error = function(error)
                print('Error:')
            end    
        })
end

-- read data from udp scocket 
-- on success read next data 
function Nuvo:discoverReadData(socket, devicesTable)
	socket:receive({
	            success = function(data)
	                table.insert(devicesTable, data)
	                self:discoverReadData(socket, devicesTable)
	            end,
	            error = function(error)
	            	local SSDPResponse = self:parseDiscoverResponse(devicesTable)
	            	self:printDevices(SSDPResponse)
	            end})
end

-- parse response (ip and uuid) from nuvo device
-- sample response:
-- HTTP/1.1 200 OK
-- Location: http://192.168.80.111:44424/00000000-0000-0000-0000-0025ed1c6473.xml
-- Ext:
-- USN: uuid:00000000-0000-0000-0000-0025ed1c6473::urn:schemas-nuvotechnologies-com:device:Zone:1
-- Server: Linux/2.6.37 UPnP/1.0 GUPnP/0.20.8
-- Cache-Control: max-age=100
-- ST: urn:schemas-nuvotechnologies-com:device:Zone:1
-- Date: Mon, 12 Jan 2015 14:43:45 GMT
-- Content-Length: 0
function Nuvo:parseDiscoverResponse(devicesTable)
	local SSDPResponse = {}
	for key, device in pairs(devicesTable) 
	do
		local ST = device:match("ST: urn:(.*):device:Zone:1")

		if(ST == 'schemas-nuvotechnologies-com')
		then	
			local ip = device:match("http://(.*):%d*/")
			local port = device:match("http://.*:(%d*)/")
			local uuid = device:match("uuid:(.*)::urn")
			local location = device:match("Location: (http://.*xml)")

			table.insert(SSDPResponse, {ip = ip, uuid = uuid, port = port, location = location})
		end
	end

	return SSDPResponse
end

function Nuvo:sendCommandToAllZones(command, value)
	for _, zone in pairs(self.zones)
	do
		zone:sendCommand(command, value)
	end
end

function Nuvo:pauseAll()
	self:sendCommandToAllZones('pause')
end

function Nuvo:playAll()
	self:sendCommandToAllZones('play')
end

function Nuvo:nextAll()
	self:sendCommandToAllZones('next')
end

function Nuvo:previousAll()
	self:sendCommandToAllZones('previous')
end

function Nuvo:muteAll(value)

	self:updateProperty('ui.mute_Switch.value', value)
	for _, zone in pairs(self.zones)
	do
		zone:setMute(value)
	end
end

function Nuvo:refreshGlobalVolume()
	local maxVolume = 0

	for _, zone in pairs(self.zones)
	do
		if maxVolume < (zone.properties['volume'] or 0)
		then
			maxVolume = zone.properties['volume']
		end
	end

	self:updateProperty('volume', maxVolume)
	self:updateProperty('ui.volume_Label.caption', tostring(maxVolume))
	self:updateProperty('ui.volume_Slider.value', maxVolume)

end

function Nuvo:refreshGlobalMute()
	local mute = true

	for _, zone in pairs(self.zones)
	do
		if not zone.properties['ui.mute_Switch.value'] then 
			mute = false
		end
	end

	self:updateProperty('ui.mute_Switch.value', mute)
end

function Nuvo:setGlobalValume(volume)
	local currentVolume = self.properties['volume']
	local volumeRatio = volume - currentVolume

	for _, zone in pairs(self.zones)
	do
		zone:sendCommand('volume', (zone.properties['volume'] + volumeRatio) )	
	end 
end
