require('common.device')
require('common/TCPListener')

class 'Onkyo' (Device)

Onkyo.HEADER_BEGIN = 'ISCP' 

Onkyo.HEADER_SIZE  = string.char( 0x00, 0x00, 0x00, 0x10 )

Onkyo.HEADER_VERSION =  string.char( 0x01, 0x00, 0x00, 0x00 )

Onkyo.PACKET_END = string.char( 0x0d,  0x0a)

Onkyo.COMMANDS = {
	["KEY_ENTER"]	 = '!1OSDENTER',
	["KEY_UP"]= '!1OSDUP',
	["KEY_DOWN"] = '!1OSDDOWN',
	["KEY_RIGHT"] = '!1OSDLEFT',
	["KEY_LEFT"] = '!1OSDRIGHT',
	["KEY_HOME"] = '!1OSDHOME',
	["KEY_RETURN"] = '!1OSDEXIT',
	["KEY_SETUP"] = '!1OSDMENU'
}

Onkyo.INPUT_NAME_TO_CODE = {
	["BD/DVD"] 		= "10",
	["VCR/DVR"]		= "00", 
	["CBL/SAT"] 	= "01",
	["GAME"] 		= "02",
	["PC"] 			= "05",
	["AUX"] 		= "03",
	["TUNER"] 		= "26",
	["TV/CD"] 		= "23",
	["PORT"] 		= "40",
	["NET"] 		= "2B",
	["USB"] 		= "29",
	["PHONO"] 		= "22"
}

Onkyo.INPUT_CODE_TO_NAME = {
	["10"] 		= "BD/DVD",
	["00"]		= "VCR/DVR", 
	["01"]	 	= "CBL/SAT",
	["02"] 		= "GAME",
	["05"] 		= "PC",
	["03"] 		= "AUX",
	["26"] 		= "TUNER",
	["23"] 		= "TV/CD",
	["40"] 		= "PORT",
	["2B"] 		= "NET",
	["29"] 		= "USB",
	["22"] 		= "PHONO"
}

function num2hex(num)
	local hexstr = '0123456789ABCDEF'
	local s = ''
	while num > 0 do
		local mod = math.fmod(num, 16)
		s = string.sub(hexstr, mod+1, mod+1) .. s
		num = math.floor(num / 16)
	end
	if s == '' then s = '0' end
	if s:len() == 1
	then
		s = '0' .. s
	end
	return s
end

function Onkyo:__init(id)
	print('------------- init')

	Device.__init(self, id)
	
	-- create connection

	self.sock = TCPListener({delimiter = string.char( 0x0d,  0x0a)})
    self.sock:registerEvent('dataReceived', function(sock, data) self:onDataReceived(data, id) end)
    self.sock:registerEvent('disconnected', function()  self:onClosed() end)
    self.sock:registerEvent('connected', function() self:onConnected() end)
    self.sock:registerEvent('error', function(sock, error) print("error", error) end)

    if (self.properties.ip ~= '' and self.properties.port ~= '' and 
    	self.properties.ip ~= nil and self.properties.port ~= nil) then  

    	self.sock:connect(self.properties.ip, tonumber(self.properties.port))
    else
        print('No ip or port')
        self:updateProperty('port', 60128)
    end        
end

function Onkyo:onConnected()
	setTimeout(function() self:refreshState(1000) end, 1000)
end

function Onkyo:onClosed()
	setTimeout(function() self.sock:connect(self.properties.ip, tonumber(self.properties.port)) end, 3000)
end

function Onkyo:onDataReceived(data, id)
	print("Onkyo:onDataReceived data: ", data)
	--master volume event
	if string.sub(data, 19, 21) == 'MVL'
	then
		local value = string.sub(data, 22, 23)
		if (not string.find(data, "MVLN/A")) then
			self:updateProperty('volume', tonumber(value, 16))
		end

	--input event
	elseif  string.sub(data, 19, 21) == 'SLI'
	then
		local input = Onkyo.INPUT_CODE_TO_NAME[string.sub(data, 22, 23)]
		if input ~= nil
		then
			self:updateProperty('input', input)
		end

	--power event
	elseif  string.sub(data, 19, 21) == 'PWR'
	then
		local value = string.sub(data, 22, 23)
		if value == '00'
		then
			self:updateProperty('power', false)
		elseif value == '01'
		then
			self:updateProperty('power', true)
		end

	--mute event
	elseif  string.sub(data, 19, 21) == 'AMT'
	then
		local value = string.sub(data, 22, 23)
		if value == '00'
		then
			self:updateProperty('mute', false)
		elseif value == '01'
		then
			self:updateProperty('mute', true)
		end
	end
end


function Onkyo:turnOn()
	self:sendCommand("!1PWR01")
	self:updateProperty("power", true)
end

function Onkyo:turnOff()
	self:sendCommand('!1PWR00')
	self:updateProperty("power", false)
end

function Onkyo:setInput(input)
	local inputCode = Onkyo.INPUT_NAME_TO_CODE[input]
	if inputCode ~= nil
	then 
		cmd = "!1SLI" .. inputCode
		self:sendCommand(cmd)
		self:updateProperty("input", input)
	end
end

function Onkyo:setVolume(vol)
	print("Onkyo:setVolume")
	if (vol >= 0) and (vol <= 100) 
	then
		cmd = "!1MVL" .. num2hex(vol)	
		self:sendCommand(cmd)
		self:updateProperty("volume", vol)
		self:updateProperty("mute", false)
	end
end 

function Onkyo:setMute(mute)
	if mute == true 
	then 
		cmd = "!1AMT01" 
		self:updateProperty("mute", true)
	else
		cmd = "!1AMT00" 
		self:updateProperty("mute", false)
	end

	self:sendCommand(cmd)
end 


function Onkyo:sendCommand(cmd)
	print("-----------------------> sendCommand")
	size = string.len(cmd .. Onkyo.PACKET_END)
	dataSize  = string.char( 0x00, 0x00, 0x00, size)
	self.sock:send(Onkyo.HEADER_BEGIN .. Onkyo.HEADER_SIZE .. dataSize .. Onkyo.HEADER_VERSION  .. cmd .. Onkyo.PACKET_END)
	print("-----------------------> DONE")

end

function Onkyo:refreshState(time)
	--check master volume
	self:sendCommand('!1MVLQSTN')
	--check mute
	self:sendCommand('!1AMTQSTN')
	--check input
	self:sendCommand('!1SLIQSTN')
	--check power
	self:sendCommand('!1PWRQSTN')

	if time ~= nil
	then
		if time > 0 
		then
			setTimeout(function() self:refreshState(1000) end, time)
		end
	end
end
 
function Onkyo:sendKey(key)
	local cmd = Onkyo.COMMANDS[key]
	if cmd ~= nil
	then
		self:sendCommand(cmd)
	end
end