require('common.device')
require('philipsHueLight')

require('net.HTTPClient')

class 'philipsHue' (Device)

--! Constructor - creates philipsHue object. Creates child device objects for all lights connected with the Hue bridge
--!@param id id of this device or table with this device data
function philipsHue:__init(id)
    Device.__init(self, id)

    self.uiBinding = {}
    self.uiBinding["onOff"] = function(event) self:onOff(event); end
    self.uiBinding["hue"] = function(event) self:changeHue(event.values[1]); end
    self.uiBinding["saturation"] = function(event) self:changeSaturation(event.values[1]); end
    self.uiBinding["brightness"] = function(event) self:changeBrightness(event.values[1]); end
    
    self.user = self.properties.userName or ""
    self.devicetype = "HC2"

    self.lights = {}
    local childDevices = plugin.getChildDevices(self.id)

    for _,dev in pairs(childDevices) do
        if (dev.type == 'com.fibaro.philipsHueLight') then
            self.lights[dev.id] = philipsHueLight(self, dev)
        end
    end

    self.controlHeaders = {
        ["content-type"] = 'application/json; charset=UTF-8'
    }
    self.http = net.HTTPClient({timeout=2000})

    self.onPropertyChanged = {}
    self.onPropertyChanged["on"] = function () self:updateProperty('ui.onOff.value', self.properties.on); end

    self:updateProperty('ui.stateLabelValue.caption', 'i18n_stateValueInit')

    self:getFullState()
end

--! Creates user in Hue bridge. This user will be used for all comunication with Hue
--! In order for this process to finish succesfully the button on Hue bridge must be pressed (after pressing it there are 30sec for user creation)
function philipsHue:createUser()
    if (self.properties.ip == "") then
        return
    end

    local requestBody = '{"devicetype":"' .. self.devicetype .. '"}'    
    local controlUrl = 'http://' .. self.properties.ip .. '/api'

    self.http:request(controlUrl, {
        options={
            headers = self.controlHeaders,
            data = requestBody,
            method = 'POST',
            timeout = 5000
        },
        success = function(status)
            local result = json.decode(status.data)
            if result[1] then
                if result[1].error then
                    print ("ERROR")
                    print ("  type: " .. result[1].error.type)
                    print ("  address: " .. result[1].error.address)
                    print ("  description: " .. result[1].error.description)
                    self:updateProperty('ui.stateLabelValue.caption', result[1].error.description)
                elseif result[1].success then
                    self:updateProperty("userName", result[1].success.username)
                    self:updateProperty('ui.stateLabelValue.caption', 'i18n_stateValueCreatingUserOK')
                    self.user = result[1].success.username
                else
                    print ("unknown response structure: ")
                    print(status)
                end
            else
                print ("unknown response structure: ")
                print(status)
            end
        end,
        error = function(error)
            print "ERROR"
            print(error)
            self:updateProperty('ui.stateLabelValue.caption', 'i18n_stateValueConnectionTimeoutError')
        end
    })
end


--!Searches for new lights, and than adds them to the bridge - this process takes >60sec
--!Adds maximum of 15 lights. If more are to be added, this method must be called again
--!Newly found lights would be added automaticly to HC in next polling after searching process is finished
function philipsHue:searchForNewLights()
    if (self.properties.ip == "") then
        return
    end

    local controlUrl = 'http://' .. self.properties.ip .. '/api/' .. self.user .. '/lights'
    
    self.http:request(controlUrl, {
        options={
            headers = self.controlHeaders,
            method = 'POST'
        },
        success = function(status)
            decodedStatus = json.decode(status.data)

            if decodedStatus and decodedStatus[1].error and decodedStatus[1].error.description then
                self:updateProperty('ui.stateLabelValue.caption', decodedStatus[1].error.description)
            elseif decodedStatus and decodedStatus[1].success and decodedStatus[1].success['/lights'] then
                self:updateProperty('ui.stateLabelValue.caption', decodedStatus[1].success['/lights'])
            end
        end,
        error = function(error)
            print "ERROR"
            print(error)
            self:updateProperty('ui.stateLabelValue.caption', 'i18n_stateValueConnectionTimeoutError')
        end
    })
end

--! Gets configuration from Hue bridge. Json with configuration data would be passed as argument to callback function
--!@param callback one parameter function, which will be called after retrieving configuration data. On success resulting json would be passed to it, otherwise error
function philipsHue:getConfiguration(callback)
    if (self.properties.ip == "") then
        return
    end

    if (type(callback) == 'function') then

        url = 'http://' .. self.properties.ip .. '/api/' .. self.user .. '/config'

        self.http:request(url, {
            options={
                headers = self.controlHeaders,
                method = 'GET'
            },
            success = function(status)
                callback(status.data)
            end,
            error = function(error)
                callback(error)
                print "ERROR"
                print(error)
                self:updateProperty('ui.stateLabelValue.caption', 'i18n_stateValueConnectionTimeoutError')
            end
        })
    end
end

--! Gets full state from Hue bridge, and calls philipsHue:processStatus, to analyze response and update all light objects
function philipsHue:getFullState()
    if (self.properties.ip == "") then
        return
    end
    local url = 'http://' .. self.properties.ip .. '/api/' .. self.user .."/"

    local timeout = self.properties.pollingTimeout
    if timeout == nil then
        timeout = 2000
    else
        timeout = tonumber(timeout) * 1000
    end

    self.http:request(url, {
        options={
            headers = self.controlHeaders,
            method = 'GET'
        },
        success = function(status)
            self:processStatus(status.data)
            setTimeout(function() self:getFullState() end , timeout)
        end,
        error = function(error)
            print(error)
            self:updateProperty('ui.stateLabelValue.caption', 'i18n_stateValueConnectionTimeoutError')
            setTimeout(function() self:getFullState() end , timeout)
        end
    })
end

--! Processes Hue bridge json message and updates statuses of all lights
--!@param status json response from Hue bridge with status of all lights
function philipsHue:processStatus(status)
    status = json.decode(status)

    if status and status.lights then
        for key, light in pairs(status.lights) do
            local lightIndex = self:findLight(key)

            if lightIndex>0 then
                local devObject = self.lights[lightIndex]
                devObject:update(key, light)
            else
                local dev = plugin.createChildDevice(self.id, 'com.fibaro.philipsHueLight', key)
                local devObject = philipsHueLight(self, dev)

                devObject:update(key, light)

                self.lights[dev.id] = devObject
            end

            local labelValue = 'Found lights: ' .. tostring(self:countTable(self.lights)) or "0"

            self:updateProperty('ui.stateLabelValue.caption', labelValue)
        end
    end

end


--!Searches in child devices for lights with specified lightId, and returns it's index or 0
--!@param lightId  Hue id (NOT device id)
--!@return if found index of specified light, else 0
function philipsHue:findLight(lightId)
    for key, light in pairs(self.lights) do
        if tonumber(light.properties.lightId) == tonumber(lightId) then
            return key
        end
    end

    return 0
end


--!Changes light state, it calls philipsHue:processLightStateResult to process result of API calls
--!@param lightHueId  Hue id (NOT device id)
--!@param args  table[key=>value] of states to be changed (values must be strings). Supported keys are: on, bri, hue, sat, xy, ct, transitiontime, effect, alert (http://developers.meethue.com/1_lightsapi.html#16_set_light_state)
function philipsHue:setLightState(lightHueId, args)
    if (self.properties.ip == "") then
        return
    end
    if (type(args) == "table") then
        url = 'http://' .. self.properties.ip .. '/api/' .. self.user .. "/lights/" .. lightHueId .. "/state"

        local body = ""
        for key, value in pairs(args) do
            if body ~= "" then
                body = body .. ", "
            end
            body = body .. "\"" .. key .. "\": " .. value
        end
        body = "{" .. body .. "}"

        self.http:request(url, {
            options={
                method = 'PUT',
                headers = self.controlHeaders,
                data = body
            },
            success = function(status)
                self:processLightStateResult(status.data)
            end,
            error = function(error)
                print("ERROR" .. error)
                self:updateProperty('ui.stateLabelValue.caption', 'i18n_stateValueConnectionTimeoutError')
            end
        })
    end
end

--! Processes response to changing lights state from Hue bridge. Update properties of lights based on values from the response.
--!@param resultMessage json with response from bridge
function philipsHue:processLightStateResult(resultMessage)
    local result = json.decode(resultMessage)
    for k, v in pairs(result[1]) do
        if k == "success" then
            for key, value in pairs(v) do
                local lightId = string.match(key, "%d+")
                local property = string.match(key, "[a-z]+$")

                local light = self:findLight(lightId)
                if (property == "xy")then
                    self.lights[light]:updateProperty("x", value[1])
                    self.lights[light]:updateProperty("y", value[2])
                else
                    self.lights[light]:updateProperty(property, value)
                end
            end
        end
    end
end

--! Changes brightness of all lights connected to Hue bridge
--!@param brightness integer in 0-255 range
function philipsHue:changeBrightness(brightness)
    self:updateProperty("ui.brightness.value", brightness)
    for k,v in pairs(self.lights) do
        v:changeBrightness(brightness)
    end
end

--! Changes hue value of all lights connected to Hue bridge
--!@param hue integer in 0-65535 range
function philipsHue:changeHue(hue)
    self:updateProperty("ui.hue.value", hue)
    for k,v in pairs(self.lights) do
        v:changeHue(hue)
    end
end

--! Changes saturation of all lights connected to Hue bridge
--!@param saturation integer in 0-255 range
function philipsHue:changeSaturation(saturation)
    self:updateProperty("ui.saturation.value", saturation)
    for k,v in pairs(self.lights) do
        v:changeSaturation(saturation)
    end
end

--! Changes on/off state of all lights connected to Hue bridge. If light is off it turns on; and if it's on, it becomes off.
function philipsHue:onOff()
    local value = not self.properties.on

    self:updateProperty("on", value)

    if value then
        self:turnOn()
    else
        self:turnOff()
    end
end

--! Turns all lights connected to Hue bridge on
function philipsHue:turnOn()
    for k,v in pairs(self.lights) do
        v:turnOn()
    end
end

--! Turns all lights connected to Hue bridge off
function philipsHue:turnOff()
    for k,v in pairs(self.lights) do
        v:turnOff()
    end
end

function philipsHue:countTable(table)
    local counter = 0

    for k,v in pairs(table or {}) do
        counter = counter + 1
    end

    return counter
end
