require('common.device')
require('colorConverter')

class 'philipsHueLight' (Device)

--! Constructor - creates philipsHueLight object, initialises ui bindings and and onPropertyChanged actions
--!@param parent philipsHue object that is parent of this light
--!@param dev id of this device or table with this device data
function philipsHueLight:__init(parent, dev)
    Device.__init(self, dev)
    self.parent = parent

    self.uiBinding = {}
    self.uiBinding["onOff"] = function(event) self:onOff(event); end
    self.uiBinding["hue"] = function(event) self:changeHue(event.values[1]); end
    self.uiBinding["saturation"] = function(event) self:changeSaturation(event.values[1]); end
    self.uiBinding["brightness"] = function(event) self:changeBrightness(event.values[1]); end

    self.onPropertyChanged = {}
    self.onPropertyChanged["on"] = function () self:updateProperty('ui.onOff.value', self.properties.on); end
    self.onPropertyChanged["bri"] = function () self:updateProperty('ui.brightness.value', self.properties.bri); end
    self.onPropertyChanged["sat"] = function () self:updateProperty('ui.saturation.value', self.properties.sat); end
    self.onPropertyChanged["hue"] = function () self:updateProperty('ui.hue.value', self.properties.hue); end
end


--!@brief Updates lights properties based on information from  hue bridge
--!@param id - hue Id (NOT device Id)
--!@param light - structure received from Hue bridge
function philipsHueLight:update(id, light)
    if (light.state ~= nil) then
        self:updateProperty('lightId', tonumber(id))
        if (light.state.on ~= nil) then
            self:updateProperty('on', light.state.on) 
        end

        if (light.state.bri ~= nil) then
            self:updateProperty('bri', light.state.bri)
        end

        if (light.state.sat ~= nil) then
            self:updateProperty('sat', light.state.sat)
        end

        if (light.state.hue ~= nil) then
            self:updateProperty('hue', light.state.hue)
        end

        if (light.state.effect ~= nil) then
            self:updateProperty('effect', light.state.effect)
        end

        if (light.state.xy ~= nil) then
            self:updateProperty('x', light.state.xy[1])
            self:updateProperty('y', light.state.xy[2])
        end

        if (light.state.ct ~= nil) then
            self:updateProperty('ct', light.state.ct)
        end

        if (light.state.reachable ~= nil) then
            if (light.state.reachable) then
                self:updateProperty('dead', 0)
            else
                self:updateProperty('dead', 1)
            end
        end
    end
end

--! Changes brightness of light
--!@param brightness integer in 0-255 range
function philipsHueLight:changeBrightness(brightness)
    self:updateProperty("ui.brightness.value", brightness)
    self.parent:setLightState(self.properties.lightId, {["bri"]=brightness})
end

--! Changes hue value of light
--!@param hue integer in 0-65535 range
function philipsHueLight:changeHue(hue)
    self:updateProperty("ui.hue.value", hue)
    self.parent:setLightState(self.properties.lightId, {["hue"]=hue})
end

--! Changes saturation of light
--!@param saturation integer in 0-255 range
function philipsHueLight:changeSaturation(saturation)
    self:updateProperty("ui.saturation.value", saturation)
    self.parent:setLightState(self.properties.lightId, {["sat"]=saturation})
end

--! Changes color of light to specified rgb color
--!@param r red component of color, integer in 0-255 range
--!@param g green component of color, integer in 0-255 range
--!@param b blue component of color, integer in 0-255 range
function philipsHueLight:changeRGB(r,g,b)
    local pointXY = colorConverter.rgb2xy(r/255, g/255, b/255)
    local pointXYjson = "[" .. pointXY.x .. ", " .. pointXY.y .. "]"
    self.parent:setLightState(self.properties.lightId, {["xy"]=pointXYjson})
end

--! Changes on/off state of device. If light is off it turns on; and if it's on, it becomes off.
function philipsHueLight:onOff()
    local value = not self.properties.on

    if value then
        self:turnOn()
    else
        self:turnOff()
    end
end


--! Turns the light on
function philipsHueLight:turnOn()
    local hue = self.properties["ui.hue.value"]
    local bri = self.properties["ui.brightness.value"]
    local sat = self.properties["ui.saturation.value"]
    self.parent:setLightState(self.properties.lightId, {["on"]="true", ["hue"]=hue, ["sat"]=sat, ["bri"]=bri})
end

--! Turns the light off
function philipsHueLight:turnOff()
    self.parent:setLightState(self.properties.lightId, {["on"]="false"})
end
