--
-- Tested on:
-- Product: Pioneer AV Receiver
-- Model number: VSX-828-K
-- Software Version: s9294.4004.1066
--

require('common.device')
require('common/TCPListener')

class 'Pioneer' (Device)

Pioneer.INPUT_NAME_TO_CODE = {
	["DVD"] 		= "04",
	["BD"]			= "25",
	["TV/SAT"]		= "05",
	["CBL/SAT"]		= "06",
	["DVD/BDR"]		= "15",
	["VIDEO 1"]		= "10",
	["VIDEO 2"]		= "14",
	["HDMI 1"]		= "19",
	["HDMI 2"]		= "20",
	["HDMI 3"]		= "21",
	["HDMI 4"]		= "22",
	["HDMI 5"]		= "23",
	["HDMI 6"]		= "24",
	["HDMI 7"]		= "34",
	["HOME MEDIA GALLERY"]	= "26",
	["iPod/USB"]	= "17",
	["XM RADIO"]	= "18",
	["CD"]			= "01",
	["CD-R/TAPE"]	= "03",
	["TUNER"]		= "02",
	["PHONO"]		= "00",
	["MULTI CH IN"]	= "12",
	["ADAPTER PORT"]	= "33",
	["SIRIUS"]			= "27",
	["FAVORITE"]		= "45",
	["MEDIA SERVER"]	= "44",
	["INTERNET RADIO"]	= "38"
}

Pioneer.INPUT_CODE_TO_NAME = {
	["04"] = "DVD",
	["25"] = "BD",
	["05"] = "TV/SAT",
	["06"] = "CBL/SAT",
	["15"] = "DVD/BDR",
	["10"] = "VIDEO 1",
	["14"] = "VIDEO 2",
	["19"] = "HDMI 1",
	["20"] = "HDMI 2",
	["21"] = "HDMI 3",
	["22"] = "HDMI 4",
	["23"] = "HDMI 5",
	["24"] = "HDMI 6",
	["34"] = "HDMI 7",
	["26"] = "HOME MEDIA GALLERY",
	["17"] = "iPod/USB",
	["18"] = "XM RADIO",
	["01"] = "CD",
	["03"] = "CD-R/TAPE",
	["02"] = "TUNER",
	["00"] = "PHONO",
	["12"] = "MULTI CH IN",
	["33"] = "ADAPTER PORT",
	["27"] = "SIRIUS",
	["45"] = "FAVORITE",
	["44"] = "MEDIA SERVER",
	["38"] = "INTERNET RADIO"
}

function Pioneer:__init(id)
	Device.__init(self, id)
	-- create connection
	self.sock = TCPListener({delimiter = string.char( 0x0D , 0x0A ), readTimeout = 10000, connectTimeout=10000})
    self.sock:registerEvent('dataReceived', function(sock, data) self:onDataReceived(data, id) end)
    self.sock:registerEvent('disconnected', function()  self:onClosed() end)
    self.sock:registerEvent('connected', function() self:onConnected() end)
    self.sock:registerEvent('error', function(sock, error) print("error", error) end)

    if (self.properties.ip ~= '' and self.properties.ip ~= nil) then  
    	self.sock:connect(self.properties.ip, 8102)
    else
        print('No ip or port')
    end

    volumeUncalculated = 0
end

function Pioneer:onConnected()
	Device.updateProperty(self, 'dead', false)
	self:refreshState()
end

function Pioneer:onClosed()
	setTimeout(function() self.sock:connect(self.properties.ip, 8102) end, 3000)
end

function Pioneer:onDataReceived(data, id)
	print("Pioneer:onDataReceived: ", data)
	--
	-- power on & power off
	--
	if (data == "PWR0") then
		self:updateProperty('power', true)
		self.sock:send("?V\r") -- get new volume report
	elseif (data == "PWR1") then
		self:updateProperty('power', false)
	--
	-- volume changed
	--
	elseif (string.sub(data, 1, 3) == "VOL") then
		volumeUncalculated = tonumber(string.sub(data, 4, 6))
		volume = tonumber(string.sub(data, 4, 6))
		volume = volume * 0.5 - 80.5
		self:updateProperty('volume', volume)
	--
	-- mute
	--
	elseif (data == "MUT0") then
		self:updateProperty('mute', true)
	elseif (data == "MUT1") then
		self:updateProperty('mute', false)
		self.sock:send("?V\r") -- get new volume report
	--
	-- input changed
	--
	elseif (string.sub(data, 1, 2) == "FN") then
		print('report received')
		input = string.sub(data, 3, 4)
		input = self.INPUT_CODE_TO_NAME[input]
		self:updateProperty('input', input)
	end	
end

--
--
-- actions
--
-- 
function Pioneer:turnOn()
	self.sock:send("PO\r")
end

function Pioneer:turnOff()
	self.sock:send("PF\r")
end

function Pioneer:setVolume(command)

		if (command == 'VD') then
			volumeUncalculated = volumeUncalculated - 2
		else
			volumeUncalculated = volumeUncalculated + 2
		end

		if (((volumeUncalculated ) * 0.5) - 80.5 <= 0 and ((volumeUncalculated ) * 0.5) - 80.5 >= -80.0 and volumeUncalculated ~= nil) then
			self:updateProperty('volume', ((volumeUncalculated ) * 0.5) - 80.5)
			print( ((volumeUncalculated ) * 0.5) - 80.5)
		end 

		self.sock:send(command.."\r")
end

function Pioneer:setMute(state)
	if (state) then
		self.sock:send("MO\r")
	else
		self.sock:send("MF\r")
	end
end

function Pioneer:setInput(input)
	print('setInput to ' .. input)

	local code = self.INPUT_NAME_TO_CODE[input]

	if (code) then
		local cmd = code .. "FN\r"
		self.sock:send(cmd)
	end
end

function Pioneer:refreshState()
	self.sock:send("?P\r")	-- power
	self.sock:send("?V\r")	-- volume
	self.sock:send("?M\r")	-- mute
	self.sock:send("?F\r")	-- input
end