require('RussoundController')
require('common.device')
require('common.TCPListener')
require('net.HTTPClient')

 
class 'RussoundMCA88X' (Device) -- controller class
 

function RussoundMCA88X:__init(id)
	Device.__init(self, id)	

	self.network = api.get('/settings/network')
	self.http = net.HTTPClient({ timeout = 10000 }) 
	self.zones = {}
	self.controllers = {}
	self.port = 9621
	self.delimiter = '\r\n'

	self:sync()
end

function RussoundMCA88X:sync()
	self.zones = {}
	self.controllers = {}
	self.socketsTable = {}
	self.retryCounter = 0

	self:discover()
end

function RussoundMCA88X:discover()
	self.retryCounter = self.retryCounter + 1
	local url = 'http://127.0.0.1:11111/api/networkDiscovery/arp'

	self:updateProperty('ui.playersLabelInfo.caption', 'Zones:\n---\n\nZones in Party Mode:\n---')
 	self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelDiscovering')

	self.http:request(url, {	
		options={
			headers = {['X-Fibaro-Version'] = '2'},
			method = 'POST',
			data = '{}',
		}, 
		success = function(response)
			if response and response.status == 200 then
				self:parseDiscoverResponse(response.data)
			else
				print('network error')
			end
		end,
		error = function (error)
			print('*******************')
			print('* ERROR: '.. error)
			print('*******************')	
			self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelRetryExhausted')		
		end
	})
end

function RussoundMCA88X:parseDiscoverResponse(data)
	local ok, jsonData = pcall(json.decode, data)

	if ok and jsonData then
		for _,arpScanEntry in pairs(jsonData) do
			if arpScanEntry.vendor and arpScanEntry.vendor == 'Russound' then
				self.controllers[arpScanEntry.ip] = {
					['status'] = 'notVerified',
					['vendor'] = arpScanEntry.vendor,
					['mac'] = arpScanEntry.mac
				}
			end
		end
	end

	self:verifyIps(self.controllers)
end

function RussoundMCA88X:verifyIps()
	if self:countTableElements(self.controllers) > 0 then
		self:checkcontrollers()

		for ip,_ in pairs(self.controllers) do
			self:startSocket(ip)
		end
	else
		self:retryDiscover()
	end
end

function RussoundMCA88X:retryDiscover()
	if self.retryCounter <= 5 then
		self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelDiscoveringRetry')
		setTimeout(function() self:discover() end, 3000)
	else
		self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelRetryExhausted')		
	end
end

function RussoundMCA88X:startSocket(ip)  -- We starting sockets here (in this class) because of need to test that  Ip Devices (controllers) are reachable and can connect to it.
	local ip = tostring(ip)

    self.socketsTable[ip] = TCPListener({delimiter = string.char( 0x0D , 0x0A ), readTimeout = 10000, connectTimeout = 10000})
    self.socketsTable[ip]:registerEvent('dataReceived', function(sock, data) self:onDataReceived(sock, data) end)
    self.socketsTable[ip]:registerEvent('disconnected', function()  self:onClosed(ip) end)
    self.socketsTable[ip]:registerEvent('connected', function() self:onConnected(ip) end)
    self.socketsTable[ip]:registerEvent('error', function(error) print("error", error, ip) end)
 
    self.socketsTable[ip]:connect(ip, self.port)     
end

function RussoundMCA88X:onConnected(ip)
	self.controllers[ip].status = true
end

-- For log purpose
function RussoundMCA88X:onDataReceived(sock,  data)
	-- print('onDataReceived', data)   
end

function RussoundMCA88X:onClosed(ip)
	if self.controllers[ip] and self.controllers[ip].status then
		self.controllers[ip].status = false
	end	
end


function RussoundMCA88X:checkcontrollers()
	if self:mustRetry() then
		setTimeout(function() self:checkcontrollers() end, 5000)
	else
		self:checkForActiveControllers()
	end
end    
 
function RussoundMCA88X:mustRetry()
	if self:countTableElements(self.controllers) > 0 then
		for _, data in pairs(self.controllers) do
			if data.status == 'notVerified' then
				return true
			end
		end

		return false
	else
		return true
	end
end

function RussoundMCA88X:checkForActiveControllers()
	local onlineControllers = {}

	for ip,data in pairs(self.controllers) do
		if data.status == true then
			onlineControllers[ip] = data
		end
	end

	if self:countTableElements(onlineControllers) > 0 then
		self.controllers = self:getControllerFromExistingHC2Model() -- We doing this way in case if some one delete zone by cross button. Then we do not have notification (won't update internal model) but the model should be always updates if sync button is pressed.
		self:synchronizeControllers(onlineControllers, self.controllers)
	else
		self:retryDiscover()
	end
end

function RussoundMCA88X:updateSystemLabels()
	local labelText = ''

	if self:countTableElements(self.zones) > 0 then
		self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelConnected')

		labelText = 'Zones:\n'

		for controllerId,zones in pairs(self.zones) do
			for _,zone in pairs(zones) do
				labelText = labelText .. zone.name ..'\n'
			end			
		end
	else
		self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelNotMCA88XDevices')
		labelText = 'Zones:\n---\n'
	end

	if self:countTableElements(self.zones) > 0 then
		local isParty = false
		labelText = labelText .. '\n\n Zones in Party Mode:\n'

		for controllerId,zones in pairs(self.zones) do
			for _,zone in pairs(zones) do
				if zone.properties and zone.properties.partyMode == true then
					labelText = labelText .. zone.name ..'\n'
					isParty = true
				end
			end
		end

		if not isParty then
			labelText = labelText .. '---\n'
		end
	else
		labelText =  labelText .. '\n\nZones in Party Mode:\n---'
	end

	self:updateProperty('ui.playersLabelInfo.caption', labelText)
end

function RussoundMCA88X:synchronizeControllers(onlineControllers, hcControllers)
	local onlineControllersVolume = self:countTableElements(onlineControllers)
	local hcControllersVolume = self:countTableElements(hcControllers)
	local createdControllers = {}

	if onlineControllersVolume > 0 then
		for onlineControllerIp, onlineControllerDetails in pairs(onlineControllers) do   
			local hcController = self:getControllerByIpFromLocalControllers(hcControllers, onlineControllerIp)  
        	if not hcController then -- if onlne controller from LAN is not in local hcControllers we will create it
    			local id, controller = self:createControllerDevice(onlineControllerDetails, onlineControllerIp)
    			createdControllers[id] = controller
    		end
    	end
    end
    	
    if hcControllersVolume > 0 then 
    	for _,hcController in pairs(hcControllers) do   
			local controller = self:getControllerByIpFromOnlineControllers(onlineControllers, hcController.properties.ip)       
        	if not controller then -- if zone from HC is not in online zones we will remove it
    			self:removeControllerDevice(hcController) 
    		end
    	end
    end
 
    if onlineControllersVolume > 0 then
    	for onlineControllerIp, onlineControllerDetails in pairs(onlineControllers) do    
			local hcController = self:getControllerByIpFromLocalControllers(hcControllers, onlineControllerIp)         
        	if hcController then -- if zone from local LAN is in local zones we will update same local id zone with datra form online zone
    			hcController:updateSettings(onlineControllerIp, self.socketsTable[onlineControllerIp])
    		end
    	end
    end

    self.controllers = self:mergeTables(hcControllers, createdControllers) -- to prevent double zone update in createZone and in loop above

	self:cleanAfterSync()
end 

function RussoundMCA88X:cleanAfterSync()
	setTimeout(function() self:updateSystemLabels() end, 5000) -- wait for zones update from socket (update is an async task)
end

function RussoundMCA88X:mergeTables(table1, table2)
	for k,v in pairs(table2) do
		table1[k] = v
	end

	return table1
end

function RussoundMCA88X:getControllerByIpFromLocalControllers(hcControllers, ip)
	if type(hcControllers) == 'table' then
		for _,controller in pairs(hcControllers) do
			if tostring(controller.properties.ip) == tostring(ip) then
				return controller
			end
		end
	end
	
	return nil
end

function RussoundMCA88X:getControllerByIpFromOnlineControllers(onlineControllers, ip)
	if type(onlineControllers) == 'table' then
		for controllerIp, controllerDetails in pairs(onlineControllers) do
			if tostring(controllerIp) == tostring(ip) then
				return controllerDetails
			end
		end
	end
	
	return nil
end

function RussoundMCA88X:createControllerDevice(onlineControllerDetails, ip)
	if onlineControllerDetails and ip then
		local controller = plugin.createChildDevice(self.id, 'com.fibaro.russoundController', 'Controller ' .. onlineControllerDetails.vendor)
	    local controllerObject = RussoundController(self, controller)
	    controllerObject:updateSettings(ip, self.socketsTable[ip]) -- tu dopisać, że ma też połączyć się z tą zoną i zciągnąć z niej informację o nazwie itp i kontuować jej update
	    return controller.id, controllerObject   
    end                     
end

function RussoundMCA88X:removeControllerDevice(controller)
	self.controllers[controller.properties.ip] = nil
    plugin.deleteDevice(controller.id)
end

function RussoundMCA88X:getControllerFromExistingHC2Model()
	local controllers = {}

	local childDevices = plugin.getChildDevices(self.id)

    for _,dev in pairs(childDevices) do
        if (dev.type == 'com.fibaro.russoundController') then
            controllers[dev.properties.ip] = RussoundController(self, dev)
        end
    end

    return controllers
end

----------------------------------- HELPER METHODS 

function RussoundMCA88X:countTableElements(table)
	local counter = 0

	if table and type(table) == 'table' then
		for _,_ in pairs(table) do
			counter = counter + 1
		end	
	end

	return counter
end








