require('RussoundX5Zone')
require('common.device')
require('common.TCPListener')
require('net.HTTPClient')

 
class 'RussoundX5' (Device) -- controller class
 

function RussoundX5:__init(id)
	Device.__init(self, id)	

	self.network = api.get('/settings/network')
	self.http = net.HTTPClient({ timeout = 10000 }) 
	self.zones = {} 
	self.port = 9621
	self.delimiter = '\r\n'

	self:sync()
end

function RussoundX5:sync()
	self.ipQueue = {}
	self.socketsTable = {}
	self.retryCounter = 0

	self:discover()
end

function RussoundX5:discover()
	self.retryCounter = self.retryCounter + 1
	local url = 'http://127.0.0.1:11111/api/networkDiscovery/arp'

	self:updateProperty('ui.playersLabelInfo.caption', 'Zones:\n---\n\nZones in Party Mode:\n---')
 	self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelDiscovering')

	self.http:request(url, {	
		options={
			headers = {['X-Fibaro-Version'] = '2'},
			method = 'POST',
			data = '{}',
		}, 
		success = function(response)
			if response and response.status == 200 then
				self:parseDiscoverResponse(response.data)
			else
				print('network error')
			end
		end,
		error = function (error)
			print('*******************')
			print('* ERROR: '.. error)
			print('*******************')	
			self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelRetryExhausted')		
		end
	})
end

function RussoundX5:parseDiscoverResponse(data)
	local ok, jsonData = pcall(json.decode, data)

	if ok and jsonData then
		for _,arpScanEntry in pairs(jsonData) do
			if arpScanEntry.vendor and arpScanEntry.vendor == 'Russound' then
				self.ipQueue[arpScanEntry.ip] = {
					['status'] = 'notVerified',
					['vendor'] = arpScanEntry.vendor,
					['mac'] = arpScanEntry.mac
				}
			end
		end
	end

	self:verifyIps(self.ipQueue)
end

function RussoundX5:verifyIps()
	if self:countTableElements(self.ipQueue) > 0 then
		self:checkIpQueue()

		for ip,_ in pairs(self.ipQueue) do
			self:startSocket(ip)
		end
	else
		self:retryDiscover()
	end
end

function RussoundX5:retryDiscover()
	if self.retryCounter <= 5 then
		self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelDiscoveringRetry')
		setTimeout(function() self:discover() end, 3000)
	else
		self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelRetryExhausted')		
	end
end

function RussoundX5:startSocket(ip)  
	local ip = tostring(ip)

    self.socketsTable[ip] = TCPListener({delimiter = string.char( 0x0D , 0x0A ), readTimeout = 10000, connectTimeout = 10000})
    self.socketsTable[ip]:registerEvent('dataReceived', function(sock, data) self:onDataReceived(sock, data) end)
    self.socketsTable[ip]:registerEvent('disconnected', function()  self:onClosed(ip) end)
    self.socketsTable[ip]:registerEvent('connected', function() self:onConnected(ip) end)
    self.socketsTable[ip]:registerEvent('error', function(error) print("error", error, ip) end)
 
    self.socketsTable[ip]:connect(ip, self.port)     
end

function RussoundX5:onConnected(ip)
	self.ipQueue[ip].status = true
end

-- For log purpose
function RussoundX5:onDataReceived(sock,  data)
	print('onDataReceived', data)   
end

function RussoundX5:onClosed(ip)
	self.ipQueue[ip].status = false
end


function RussoundX5:checkIpQueue()
	if self:mustRetry() then
		setTimeout(function() self:checkIpQueue() end, 5000)
	else
		self:checkForActiveZones()
	end
end    
 
function RussoundX5:mustRetry()
	if self:countTableElements(self.ipQueue) > 0 then
		for _, data in pairs(self.ipQueue) do
			if data.status == 'notVerified' then
				return true
			end
		end

		return false
	else
		return true
	end
end

function RussoundX5:checkForActiveZones()
	local onlineZones = {}

	for ip,data in pairs(self.ipQueue) do
		if data.status == true then
			onlineZones[ip] = data
		end
	end

	if self:countTableElements(onlineZones) > 0 then
		self.zones = self:getZonesFromExistingHC2Model() -- We doing this way in case if some one delete zone by cross button. Then we do not have notification (won't update internal model) but the model should be always updates if sync button is pressed.
		self:synchronizeZones(onlineZones, self.zones)
	else
		self:retryDiscover()
	end
end

function RussoundX5:updateSystemLabels()
	local labelText = ''

	if self:countTableElements(self.zones) > 0 then
		self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelConnected')

		labelText = 'Zones:\n'

		for _,zone in pairs(self.zones) do
			labelText = labelText .. zone.name ..'\n'
		end
	else
		self:updateProperty('ui.systemStatusLabelValue.caption', 'i18n_systemStatusLabelNotX5Devices')
		labelText = 'Zones:\n---\n'
	end

	if self:countTableElements(self.zones) > 0 then
		local isParty = false
		labelText = labelText .. '\n\n Zones in Party Mode:\n'

		for _,zone in pairs(self.zones) do
			if zone.properties.partyMode == true then
				labelText = labelText .. zone.name ..'\n'
				isParty = true
			end
		end

		if not isParty then
			labelText = labelText .. '---\n'
		end
	else
		labelText =  labelText .. '\n\nZones in Party Mode:\n---'
	end

	self:updateProperty('ui.playersLabelInfo.caption', labelText)
end

function RussoundX5:synchronizeZones(onlineZones, hcZones)
	local onlineZonesVolume = self:countTableElements(onlineZones)
	local hcZonesVolume = self:countTableElements(hcZones)
	local createdZones = {}

	if onlineZonesVolume > 0 then
		for onlineZoneIp,onlineZoneDetails in pairs(onlineZones) do   
			local hcZone = self:getZoneByIpFromLocalZones(hcZones, onlineZoneIp)  
        	if not hcZone then -- if onlne zone from LAN is not in local hcZones we will create it
    			local id, zone = self:createZoneDevice(onlineZoneDetails, onlineZoneIp)
    			createdZones[id] = zone
    		end
    	end
    end
    	
    if hcZonesVolume > 0 then 
    	for _,hcZone in pairs(hcZones) do   
			local cloudZoneDetails = self:getZoneByIpFromOnlineZones(onlineZones, hcZone.properties.ip)       
        	if not cloudZoneDetails then -- if zone from HC is not in online zones we will remove it
    			self:removeZoneDevice(hcZone) 
    		end
    	end
    end
 
    if onlineZonesVolume > 0 then
    	for onlineZoneIp,onlineZoneDetails in pairs(onlineZones) do    
			local hcZone = self:getZoneByIpFromLocalZones(hcZones, onlineZoneIp)         
        	if hcZone then -- if zone from local LAN is in local zones we will update same local id zone with datra form online zone
    			hcZone:updateZoneSettings(onlineZoneDetails, onlineZoneIp, self.socketsTable[onlineZoneIp])
    		end
    	end
    end

    self.zones = self:mergeTables(hcZones, createdZones) -- to prevent double zone update in createZone and in loop above

	self:cleanAfterSync()
end 

function RussoundX5:cleanAfterSync()
	setTimeout(function() self:updateSystemLabels() end, 5000) -- wait for zones update from socket (update is an async task)
end


function RussoundX5:mergeTables(table1, table2)
	for k,v in pairs(table2) do
		table1[k] = v
	end

	return table1
end

function RussoundX5:getZoneByIpFromLocalZones(hcZones, ip)
	if type(hcZones) == 'table' then
		for _,zoneDevice in pairs(hcZones) do
			if tostring(zoneDevice.properties.ip) == tostring(ip) then
				return zoneDevice
			end
		end
	end
	
	return nil
end

function RussoundX5:getZoneByIpFromOnlineZones(onlineZones, ip)
	if type(onlineZones) == 'table' then
		for zoneIp,zoneDetails in pairs(onlineZones) do
			if tostring(zoneIp) == tostring(ip) then
				return zoneDetails
			end
		end
	end
	
	return nil
end

function RussoundX5:createZoneDevice(onlineZoneDetails, ip)
	if onlineZoneDetails and ip then
		local zoneDevice = plugin.createChildDevice(self.id, 'com.fibaro.russoundX5Zone', 'Zone ' .. onlineZoneDetails.vendor)
	    local zoneDeviceObject = RussoundX5Zone(self, zoneDevice)
	    zoneDeviceObject:updateZoneSettings(onlineZoneDetails, ip, self.socketsTable[ip] ) -- tu dopisać, że ma też połączyć się z tą zoną i zciągnąć z niej informację o nazwie itp i kontuować jej update
	    return zoneDevice.id, zoneDeviceObject   
    end                     
end

function RussoundX5:removeZoneDevice(zone)
	self.zones[zone.id] = nil
    plugin.deleteDevice(zone.id)
end

function RussoundX5:getZonesFromExistingHC2Model()
	local zones = {}

	local childDevices = plugin.getChildDevices(self.id)

    for _,dev in pairs(childDevices) do
        if (dev.type == 'com.fibaro.russoundX5Zone') then
            zones[dev.id] = RussoundX5Zone(self, dev)
        end
    end

    return zones
end

----------------------------------- HELPER METHODS 

function RussoundX5:countTableElements(table)
	local counter = 0

	if table then
		for _,_ in pairs(table) do
			counter = counter + 1
		end	
	end

	return counter
end







