require('common.device')
require('common.queue')
require('satelZone')
require('satelPartition')
require('satelOutput')
require('satelResponseHandler')

require('common/TCPListener')

require('net.TCPSocket')

function tableSize(tab)
    local count = 0
    for _,_ in pairs(tab) do
        count = count + 1
    end

    return count
end

--
-- SatelAlarm
--
class 'SatelAlarm' (Device)

function SatelAlarm:__init(id)
    Device.__init(self, id)

    self.zones = {}
    self.partitions = {}
    self.outputs = {}

    self.maxZones = 128
    self.maxPartitions = 32
    self.maxOutputs = 128
    
    self.numberOfAllSteps = (self.maxZones + self.maxPartitions + self.maxOutputs) * 2
    self.stepsToFinishAddingPlugin = (self.maxZones + self.maxPartitions + self.maxOutputs) * 2
    self.lastJobPercent = 0
    self.notificationId = 0

    self.queue = queue()

    self.currentMessage = nil
    self.connectTimeoutValue = 360000
    self.closeTimeoutID = nil
    self.canPollForChanges = false

    self.sock = net.TCPSocket()

    self:initializeChildDevices()

    self:connect()

    self.responseHandler = SatelResponseHandler(self)
end

function SatelAlarm:calculatePercentOfPluginBuildProcess()
    if self.stepsToFinishAddingPlugin > 0 then
        self.stepsToFinishAddingPlugin = self.stepsToFinishAddingPlugin -1
        local percentToFinish = (self.stepsToFinishAddingPlugin * 100) / self.numberOfAllSteps
        local percentOfFinishedJob = 100 - math.floor(percentToFinish)

        if self:canSendJobPercent(percentOfFinishedJob) then
            if self.stepsToFinishAddingPlugin == 0 then
                local putBody = {["canBeDeleted"] = true, ["data"] = {["deviceId"]  = self.id, ["progress"] = percentOfFinishedJob}}
                api.put("/notificationCenter/"..self.notificationId, putBody)
                api.delete("/notificationCenter/"..self.notificationId)
            else
                local putBody = {["canBeDeleted"] = false, ["data"] = {["deviceId"]  = self.id, ["progress"] = percentOfFinishedJob -1}}
                api.put("/notificationCenter/"..self.notificationId, putBody)
            end 
        end
    end
end

function SatelAlarm:createProgressNotification()
    local postBody = {["type"] = "SatelConfigurationNotification", ["canBeDeleted"] = false, ["data"] = {["deviceId"]  = self.id, ["progress"] = 0}}
    local postResponse = api.post("/notificationCenter", postBody)

    if postResponse then
        self.notificationId = postResponse.id
    end
end

function SatelAlarm:canSendJobPercent(percentOfFinishedJob)
    if (self.lastJobPercent ~= percentOfFinishedJob) or self.stepsToFinishAddingPlugin == 0  then
        self.lastJobPercent = percentOfFinishedJob
        return true
    else
        return false
    end
end

function SatelAlarm:connect()
    if (self.properties.ip ~= nil and self.properties.port ~= nil) then
        self.sock:connect(self.properties.ip, tonumber(self.properties.port), {
            success = function() self:onConnected() 
                self:sendNextMessage()                
            end,
            error = function(err) print(err) self:onClosed() end,
            timeout = self.connectTimeoutValue
        })
    end
end

function SatelAlarm:initializeChildDevices()
    local childDevices = plugin.getChildDevices(self.id)
    self.currentMessage = true

    if (tableSize(childDevices) == 0) then
        self:createProgressNotification()

        for i=1, self.maxZones do
            local dev = plugin.createChildDevice(self.id, 'com.fibaro.satelZone', 'Zone ' .. i)
            local devObject = SatelZone(self, dev)

            self.childDevices[dev.id] = devObject
            devObject:updateProperty('zoneId', i)
            devObject:updateProperty('disabled', true)
            devObject:setVisible(false)
            devObject:reload()

            self:calculatePercentOfPluginBuildProcess()
        end

        for i=1, self.maxOutputs do
            local dev = plugin.createChildDevice(self.id, 'com.fibaro.satelOutput', 'Output ' .. i)
            local devObject = SatelOutput(self, dev)

            self.childDevices[dev.id] = devObject
            devObject:updateProperty('outputId', i)
            devObject:updateProperty('disabled', true)
            devObject:setVisible(false)
            devObject:reload()

            self:calculatePercentOfPluginBuildProcess()
        end

        for i=1, self.maxPartitions do
            local dev = plugin.createChildDevice(self.id, 'com.fibaro.satelPartition', 'Partition ' .. i)
            local devObject = SatelPartition(self, dev)

            self.childDevices[dev.id] = devObject
            devObject:updateProperty('partitionId', i)
            devObject:updateProperty('disabled', true)
            devObject:setVisible(false)
            devObject:reload()

            self:calculatePercentOfPluginBuildProcess()
        end
    else
        self.stepsToFinishAddingPlugin = 0 -- Sets stepsToFinishAddingPlugin to 0 becouse at this point plugin is already added.

        for _,dev in pairs(childDevices) do
            if (dev.type == 'com.fibaro.satelZone') then
                self.childDevices[dev.id] = SatelZone(self, dev)
            elseif (dev.type == 'com.fibaro.satelPartition') then
                self.childDevices[dev.id] = SatelPartition(self, dev)
            elseif (dev.type == 'com.fibaro.satelOutput') then
                self.childDevices[dev.id] = SatelOutput(self, dev)
            end
        end
    end

    for id, dev in pairs(self.childDevices) do
        if (dev.properties.zoneId) then
            self.zones[tonumber(dev.properties.zoneId)] = dev
        elseif (dev.properties.partitionId) then
            self.partitions[tonumber(dev.properties.partitionId)] = dev
        elseif (dev.properties.outputId) then
            self.outputs[tonumber(dev.properties.outputId)] = dev
        else
            --print('[WARNING] Wrong child found (id = ' .. id .. ')')
        end
    end
end

function SatelAlarm:onConnected()
    logGeneral(os.date() .. ">> SATEL CONNECTION OPENED \n")
    if (self.pollTimeout) then
        clearTimeout(self.pollTimeout)
    end
    self.canPollForChanges = true
    self:pollForChanges()
    self:readLoop()
end

function SatelAlarm:readLoop()
    self.sock:read({
        success = function(data) self:onDataReceived(data) self:readLoop() end,
        error = function(err) print(err) self:onClosed() end
    })
end

function SatelAlarm:onClosed()
    logGeneral(os.date() .. ">> SATEL CONNECTION CLOSED \n")

    self.canPollForChanges = false

    if self.closeTimeoutID then
        clearTimeout(self.closeTimeoutID)
    end

    self.closeTimeoutID = setTimeout(function() self:connect() end, 3000)
end

function SatelAlarm:pollForChanges()
    if self.canPollForChanges then
        self:sendData(string.char(0x7f))
        self.pollTimeout = setTimeout(function() self:pollForChanges() end, 100)
    end
end

function SatelAlarm:onDataReceived(data)
    logGeneral(os.date() .. ">> RECV: " .. self:getAsHexString(data))
    self.responseHandler:handle(data)
    logGeneral("\n")
    self.currentMessage = nil
    self:sendNextMessage()
end

--! Commands that requires pin code, uses its 8bytes long representation padded by 0xff
--! e.g.: "1234" ==> 0x12 0x34 0xff 0xff 0xff 0xff 0xff 0xff
--! @return 8bytes representation of pin code
function SatelAlarm:getPinAsBytes()
    local pin = tostring(self.properties.userCode)
    local i=1
    local result = ""

    if (pin:len()%2 == 1) then
        pin = pin .. "F"
    end

    while i<pin:len() and i<16 do
        result = result .. string.char(tonumber(pin:sub(i, i+1), 16))
        i = i+2
    end

    i = result:len() + 1
    while (i<=8) do
        result = result .. string.char(0xff)
        i=i+1
    end

    return result
end

--! pushes message to message queue for later execution
function SatelAlarm:sendData(data)
    local message = string.char(0xFE) .. string.char(0xFE) .. self:__escapeFE(data .. self:crc2string(self:calculateCrc(data))) .. string.char(0xFE) .. string.char(0x0D)

    -- check if there is already the same queued message
    local found = false
    for k, q in pairs(self.queue.queue) do
        if q == message then
            found = true
        end
    end

    if found==false then
        self.queue:pushBack(message)

        if (self.currentMessage == nil) then
            self:sendNextMessage()
        end
    end
end

--! Sends to Satel next message from queue
function SatelAlarm:sendNextMessage()
    --print("QUEUE size=", #self.queue.queue)

    self.currentMessage = self.queue:popFront()
    if self.currentMessage ~= nil then

        self:calculatePercentOfPluginBuildProcess()

        logGeneral(os.date() .. ">> SEND: " .. self:getAsHexString(self.currentMessage) .. "           queue(" .. #self.queue.queue .. ")")
        self.sock:write(self.currentMessage, {
            error = function(err) print(err) self:onClosed() end
        })
    end
end

--! In all satel communication byte 0xFE has special meaning, so if it apears in message, it should be replaced by 0xFEF0
--!@param data string to be checked for 0xFE occurences
--!@return data with 0xFE replaced by 0xFEFO
function SatelAlarm:__escapeFE(data)
    return data:gsub(string.char(0xfe), string.char(0xfe, 0xf0))
end

function SatelAlarm:__highByte(integer)
    return bit32.rshift(bit32.band(integer, 0x0000FF00), 8)
end

function SatelAlarm:__lowByte(integer)
    return bit32.band(integer, 0x000000FF)
end

function SatelAlarm:crc2string(crc)
    local result = ""
    result = result .. string.char(self:__highByte(crc)) .. string.char(self:__lowByte(crc))
    return result
end

function SatelAlarm:calculateCrc(data)
    local crc = 0x147A
    
    for i = 1, data:len() do
        crc = bit32.lshift(crc, 1) -- works on 32 bits, so rotation have to be made manually
        if (bit32.band(crc, 0x00010000) == 0x00010000) then -- there was 1 shifted to bit 17
            crc = bit32.band(crc, 0x0000FFFF) -- cut to 16bits
            crc = bit32.bor(crc, 0x00000001) -- put 1 in bit 1
        end
        crc = bit32.bxor(crc, 0x0000FFFF)
        crc = crc + self:__highByte(crc) + data:byte(i)
        crc = bit32.band(crc, 0x0000FFFF) -- cut to 16bits
    end

    return crc
end

function SatelAlarm:printAsHex(data)
    local count = string.len(tostring(data))
    local i = 1
    local result = ""
    
    print("Got " .. count .. " bytes:")
    while count > 0 do
        local byte = string.byte(data, i)

        result = result .. string.format("0x%.2x ", byte)

        i = i+1
        count = count-1
    end
    print(result)
    print("")
end

function SatelAlarm:getAsHexString(data)
    local count = string.len(tostring(data))
    local i = 1
    local result = ""
    
    while count > 0 do
        local byte = string.byte(data, i)

        result = result .. string.format("0x%.2x ", byte)

        i = i+1
        count = count-1
    end
    return result
end

function SatelAlarm:getZoneById(id)
    for k,zone in pairs(self.zones) do
        if (tonumber(zone.properties.zoneId) == id) then
            return zone
        end
    end
    return nil
end

function SatelAlarm:getPartitionById(id)
    for k,partition in pairs(self.partitions) do
        if (tonumber(partition.properties.partitionId) == id) then
            return partition
        end
    end
    return nil
end

function SatelAlarm:getOutputById(id)
    for k,output in pairs(self.outputs) do
        if (tonumber(output.properties.outputId) == id) then
            return output
        end
    end
    return nil
end

-- typeId 0/1/2 - partition/input/output
-- each object is disabled/hidden and reloaded separately (so this function is recursively called in setTimeout)
-- this is done so, because disabling and hiding all devices at once (before queuing reload orders) took to much time and during this process satel broke connection
function SatelAlarm:reloadConfiguration(typeId, number)
    if typeId == nil then
      typeId = 0
      number = 1
    end
    
    -- get version at start of reconfiguration
    if typeId == 0 and number == 1 then
        self:getVersion()
    end

    local object = nil
    local newType = typeId
    local newNumber = number


    -- set type and number for next device
    if (typeId == 0) then
        object = self.partitions[number]
        newNumber = newNumber + 1
        if newNumber > self.maxPartitions then
            newNumber = 1
            newType = newType + 1
        end
    elseif (typeId == 1) then
        object = self.zones[number]
        newNumber = newNumber + 1
        if newNumber > self.maxZones then
            newNumber = 1
            newType = newType + 1
        end
    elseif (typeId == 2) then
        object = self.outputs[number]
        newNumber = newNumber + 1
        if newNumber > self.maxOutputs then
            newNumber = 1
            newType = newType + 1
        end
    end

    object:updateProperty("disabled", true)
    object:setVisible(false)
    object:reload()
  
    if (newType <=2) then
        -- reload next device
        setTimeout(function() self:reloadConfiguration(newType, newNumber) end, 10)
    end
end

function SatelAlarm:to128bits(id)
    local result = ""
    local byte = math.floor(tonumber(id-1)/8)
    local bitInByte = tonumber(id) - 8*byte
    if (bitInByte == 0) then bitInByte = 8 end
    local bitMask = math.pow(2, bitInByte-1)

    result = string.rep(string.char(0x00), byte) .. string.char(bit32.bor(0x00, bitMask)) .. string.rep(string.char(0x00), 15-byte)
    return result
end

function SatelAlarm:to32bits(id)
    local result = ""
    local byte = math.floor(tonumber(id-1)/8)
    local bitInByte = tonumber(id) - 8*byte
    if (bitInByte == 0) then bitInByte = 8 end
    local bitMask = math.pow(2, bitInByte-1)

    result = string.rep(string.char(0x00), byte) .. string.char(bit32.bor(0x00, bitMask)) .. string.rep(string.char(0x00), 3-byte)
    return result
end

function SatelAlarm:getVersion()
    self:sendData(string.char(0x7e))
end

function SatelAlarm:armAll()
    self:sendData(string.char(0x80) .. self.parent:getPinAsBytes() .. string.char(0xff, 0xff, 0xff, 0xff))
end

function SatelAlarm:disarmAll()
    self:sendData(string.char(0x84) .. self.parent:getPinAsBytes() .. string.char(0xff, 0xff, 0xff, 0xff))
end

function SatelAlarm:clearAlarmAll()
    self:sendData(string.char(0x85) .. self.parent:getPinAsBytes() .. string.char(0xff, 0xff, 0xff, 0xff))
end

function SatelAlarm:setNetworkSettings(ip, port)
    self:updateProperty("ip", ip)
    self:updateProperty("port", port)

    plugin.restart()
end

function SatelAlarm:setNetworkSettings(code)
    self:updateProperty("userCode", code)
end

function SatelAlarm:clearAlarm(partitionList)
    local partitions = self:idListTo32Bits(partitionList)
    self:sendData(string.char(0x85) .. self.parent:getPinAsBytes() .. partitions)
end

function SatelAlarm:arm(partitionList)
    local partitions = self:idListTo32Bits(partitionList)
    self:sendData(string.char(0x80) .. self.parent:getPinAsBytes() .. partitions)
end

function SatelAlarm:disarm(partitionList)
    local partitions = self:idListTo32Bits(partitionList)
    self:sendData(string.char(0x84) .. self.parent:getPinAsBytes() .. partitions)
end

function SatelAlarm:setOutputsOn(partitionList)
    local outputs = self:idListTo128Bits(partitionList)
    self:sendData(string.char(0x88) .. self.parent:getPinAsBytes() .. outputs)
end

function SatelAlarm:setOutputsOn(partitionList)
    local outputs = self:idListTo128Bits(partitionList)
    self:sendData(string.char(0x89) .. self.parent:getPinAsBytes() .. outputs)
end

--! Transforms list of ids to 16byte string with apropriate bits set
--!@param idsList coma separeted string with ids (e.g.: "1,3,5,31")
--!@return 16bytes each bit represents one id
function SatelAlarm:idListTo128Bits(partitionList)
    local startMatch
    local endMatch
    local list = {}

    while true do
        startMatch,endMatch = string.find(partitionList, "%d+")
        
        if (startMatch ~= nil) then
            table.insert(list, partitionList:sub(startMatch, endMatch))
            partitionList = partitionList:sub(endMatch+1)
        else
            break
        end
    end

    local result = string.char(0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00)

    for k,v in pairs(list) do
        local partitionBitMask = self:to128bits(v)
        result = string.char(bit32.bor(result:byte(1), partitionBitMask:byte(1)), 
                             bit32.bor(result:byte(2), partitionBitMask:byte(2)), 
                             bit32.bor(result:byte(3), partitionBitMask:byte(3)), 
                             bit32.bor(result:byte(4), partitionBitMask:byte(4)), 
                             bit32.bor(result:byte(5), partitionBitMask:byte(5)), 
                             bit32.bor(result:byte(6), partitionBitMask:byte(6)), 
                             bit32.bor(result:byte(7), partitionBitMask:byte(7)), 
                             bit32.bor(result:byte(8), partitionBitMask:byte(8)), 
                             bit32.bor(result:byte(9), partitionBitMask:byte(9)), 
                             bit32.bor(result:byte(10), partitionBitMask:byte(10)), 
                             bit32.bor(result:byte(11), partitionBitMask:byte(11)), 
                             bit32.bor(result:byte(12), partitionBitMask:byte(12)), 
                             bit32.bor(result:byte(13), partitionBitMask:byte(13)), 
                             bit32.bor(result:byte(14), partitionBitMask:byte(14)), 
                             bit32.bor(result:byte(15), partitionBitMask:byte(15)), 
                             bit32.bor(result:byte(16), partitionBitMask:byte(16)))
    end

    return result
end

--! Transforms list of ids to 4byte string with apropriate bits set
--!@param idsList coma separeted string with ids (e.g.: "1,3,5,31")
--!@return 4bytes each bit represents one id
function SatelAlarm:idListTo32Bits(idsList)
    local startMatch
    local endMatch
    local list = {}

    while true do
        startMatch,endMatch = string.find(idsList, "%d+")
        
        if (startMatch ~= nil) then
            table.insert(list, idsList:sub(startMatch, endMatch))
            idsList = idsList:sub(endMatch+1)
        else
            break
        end
    end

    local result = string.char(0x00, 0x00, 0x00, 0x00)

    for k,v in pairs(list) do
        local partitionBitMask = self:to32bits(v)
        result = string.char(bit32.bor(result:byte(1), partitionBitMask:byte(1)), 
                             bit32.bor(result:byte(2), partitionBitMask:byte(2)), 
                             bit32.bor(result:byte(3), partitionBitMask:byte(3)), 
                             bit32.bor(result:byte(4), partitionBitMask:byte(4)))
    end

    return result
end

--! Arms partition if all zones within it are armed - otherwise disarms it
--!@param partitionId Satel partition id (not device id)
function SatelAlarm:armPartitionBasedOnZonesState(partitionId)
    local armedCount = 0
    local allCount = 0

    for key, zone in pairs(self.zones) do
        if zone.properties.disabled == false and zone.properties.satelPartitionID==partitionId then
            allCount = allCount + 1
            if (zone.properties.armed == true) then
                armedCount = armedCount + 1
            end
        end
    end

    if armedCount ~= allCount and  self.partitions[partitionId].properties.armed == true then
        self.partitions[partitionId]:disarm()
    elseif allCount > 0 and armedCount == allCount and  self.partitions[partitionId].properties.armed == false then
        self.partitions[partitionId]:arm()
    end
end
