require ('characterEncoder')

--! Class resposnible for processing satel messages
class 'SatelResponseHandler'

--! Constructor
--!@param satelAlarm Satel alarm object
function SatelResponseHandler:__init(satelAlarm)

    self.resultMessages = {}
    self.resultMessages[0x00] = "ok"
    self.resultMessages[0x01] = "requesting user code not found"
    self.resultMessages[0x02] = "no access"
    self.resultMessages[0x03] = "selected user does not exist"
    self.resultMessages[0x04] = "selected user already exists"
    self.resultMessages[0x05] = "wrong code or code already exists"
    self.resultMessages[0x06] = "telephone code already exists"
    self.resultMessages[0x08] = "other error"
    self.resultMessages[0x80] = "other error"
    self.resultMessages[0x81] = "other error"
    self.resultMessages[0x82] = "other error"
    self.resultMessages[0x83] = "other error"
    self.resultMessages[0x84] = "other error"
    self.resultMessages[0x85] = "other error"
    self.resultMessages[0x86] = "other error"
    self.resultMessages[0x87] = "other error"
    self.resultMessages[0x88] = "other error"
    self.resultMessages[0x89] = "other error"
    self.resultMessages[0xFF] = "function accepted (i.e. data length and crc ok), will be processed"

    self.satelAlarm  = satelAlarm

    self.lastRead = ""

    self.responsHandlers = {}
    self.responsHandlers[0x7e] = function(data) self:handleVersionAnswer(data) end

    self.responsHandlers[0xef] = function(data) self:handleResultAnswer(data) end

    self.responsHandlers[0x7f] = function(data) self:handlePollAnswer(data) end

    self.responsHandlers[0x00] = function(data) self:handleZonesViolationAnswer(data) end
    self.responsHandlers[0x01] = function(data) self:handleZonesTamperAnswer(data) end
    self.responsHandlers[0x02] = function(data) self:handleZonesAlarmAnswer(data) end
    self.responsHandlers[0x03] = function(data) self:handleZonesTamperAlarmAnswer(data) end
    self.responsHandlers[0x04] = function(data) self:handleZonesAlarmMemoryAnswer(data) end
    self.responsHandlers[0x05] = function(data) self:handleZonesTamperAlarmMemoryAnswer(data) end
    self.responsHandlers[0x07] = function(data) self:handleZonesNoViolationAnswer(data) end
    self.responsHandlers[0x08] = function(data) self:handleZonesLongViolationAnswer(data) end

    self.responsHandlers[0x13] = function(data) self:handlePartitionAlarmAnswer(data) end
    self.responsHandlers[0x14] = function(data) self:handlePartitionFireAlarmAnswer(data) end
    self.responsHandlers[0x15] = function(data) self:handlePartitionAlarmMemoryAnswer(data) end
    self.responsHandlers[0x16] = function(data) self:handlePartitionFireAlarmMemoryAnswer(data) end
    self.responsHandlers[0x0E] = function(data) self:handlePartitionEntryTimeAnswer(data) end

    self.responsHandlers[0x09] = function(data) self:handlePartitionArmAnswer(data) end
    self.responsHandlers[0x0A] = function(data) self:handlePartitionArmAnswer(data) end

    self.responsHandlers[0x0f] = function(data) self:handlePartitionExitTimeMoreThen10Answer(data) end
    self.responsHandlers[0x10] = function(data) self:handlePartitionExitTimeLessThen10Answer(data) end

    self.responsHandlers[0x17] = function(data) self:handlePartitionOutputsAnswer(data) end

    self.responsHandlers[0xEE] = function(data) self:handleReadDeviceNameAnswer(data) end
end

--! Handles all Satel messages. Checks for correctness and dispatches interpretation to apropriate methods
--!@param data message received from Satel
function SatelResponseHandler:handle(data)
    self.lastRead = self.lastRead .. data

    local errorMessage = ""
    local message
    local fefePos

    --check if message starts with FEFE
    if (self.lastRead:byte(1) ~= 0xfe or self.lastRead:byte(2) ~= 0xfe) then
        --find first occurence of FEFE
        fefePos = self.lastRead:find(string.char(0xfe, 0xfe)) 
        
        if (fefePos ~= nil) then
            -- FEFE found - part of message incorrect - discard it
            self.errorMessage = self.lastRead:sub(1, fefePos-1)
            self.lastRead = self.lastRead:sub(fefePos)
            print("ERROR: part of message incorrect ")
            self.satelAlarm:printAsHex(errorMessage)
        end
    end

    -- find end of message (FEF0)
    local j = 1
    local fe0dPos = 1
    fefePos = self.lastRead:find(string.char(0xfe, 0xfe), 3) 
    fe0dPos = self.lastRead:find(string.char(0xfe, 0x0d), 3) 

    if (fefePos ~= nil and fe0dPos ~= nil and fefePos < fe0dPos) then
        -- FEFE encountered before FEF0 - reset message start to new FEFE
        print("ERROR: part of message incorrect ")
        self.satelAlarm:printAsHex(self.lastRead:sub(1, fefePos-1))
        data = self.lastRead:sub(fefePos)
        self.lastRead = ""
        self:handle(data)
        return    
    end

    if (fe0dPos ~= nil) then
        -- end of message found
        message = self.lastRead:sub(1, fe0dPos+1)
        self.lastRead = self.lastRead:sub(fe0dPos+2)
        message, errorMessage = self:isMessageCorrect(message)

        if errorMessage ~= nil then
            print(errorMessage)
        else
            if (type(self.responsHandlers[message:byte(1)]) == 'function') then
                self.responsHandlers[message:byte(1)](message)
            else
                print("unhandled message: ", message:byte(1))
            end
        end

        if (self.lastRead ~= "") then
            -- still some data in buffer - try to proccess it
            self:handle("")
        end
    end
end

--! Checks whether message starts with FEFE and ands with FE0D; checks if crc is correct
--!@param data satelMessage
--@return string - message without frame prefix and sufix, and without crc. On error pair (nil, errorMessage) is returned
function SatelResponseHandler:isMessageCorrect(data)
    if (data:len() < 2) then
        return nil, "ERROR in frame received from Satel. Message length < 2"
    end
    if (data:byte(1) ~= 0xfe or data:byte(2) ~= 0xfe) then
        return nil, "ERROR in frame received from Satel. Message start != FEFE"
    end
    if (data:byte(data:len()-1) ~= 0xfe or data:byte(data:len()) ~= 0x0d) then
        return nil, "ERROR in frame received from Satel. Message end != FE0D"
    end

    local rawMessage = data:sub(3, data:len()-2)
    rawMessage = rawMessage:gsub(string.char(0xfe, 0xf0), string.char(0xfe))

    if (rawMessage:len() < 1) then
        return nil, "ERROR in frame received from Satel. Empty message."
    end

    local messageCrc = rawMessage:sub(rawMessage:len()-1)
    rawMessage = rawMessage:sub(0, rawMessage:len()-2)

    local crc = self.satelAlarm:crc2string(self.satelAlarm:calculateCrc(rawMessage))

    if (crc ~= messageCrc) then
        return nil, "Crc incorrect"
    end

    return rawMessage
end

--! handles response to polling for list of new states
function SatelResponseHandler:handlePollAnswer(data)
    if (data:len() ~= 6) then
        print("ERROR in poll answer")
        self.satelAlarm:printAsHex(data)
        return
    end

    local bitNo
    local byteNo
    local commandNo = 0

    for byteNo=1,5,1 do
        for bitNo=1,8,1 do
            if bit32.band(data:byte(byteNo + 1), math.pow(2, bitNo-1)) ~= 0 then
                if (type(self.responsHandlers[commandNo]) == 'function') then
                    self.satelAlarm:sendData(string.char(commandNo))
                end
            end
            commandNo = commandNo+1
        end
    end
end

--! handles answer with zones violation state
--!@param 5bytes - each bit represents one zone. 
function SatelResponseHandler:handleZonesViolationAnswer(data)
    self:updateZonesProperty(data:sub(2), "value")
end

--! handles answer with zones tamper state
--!@param 5bytes - each bit represents one zone. 
function SatelResponseHandler:handleZonesTamperAnswer(data) 
    self:updateZonesProperty(data:sub(2), "tamper")
end

--! handles answer with zones alarm state
--!@param 5bytes - each bit represents one zone. 
function SatelResponseHandler:handleZonesAlarmAnswer(data) 
    self:updateZonesProperty(data:sub(2), "alarm")
end

--! handles answer with zones tamper alarm state
--!@param 5bytes - each bit represents one zone. 
function SatelResponseHandler:handleZonesTamperAlarmAnswer(data) 
    self:updateZonesProperty(data:sub(2), "tamperAlarm")
end

--! handles answer with zones alarm memory state
--!@param 5bytes - each bit represents one zone. 
function SatelResponseHandler:handleZonesAlarmMemoryAnswer(data) 
    self:updateZonesProperty(data:sub(2), "alarmMemory")
end

--! handles answer with zones tamper alarm memory state
--!@param 5bytes - each bit represents one zone. 
function SatelResponseHandler:handleZonesTamperAlarmMemoryAnswer(data) 
    self:updateZonesProperty(data:sub(2), "tamperAlarmMemory")
end

--! handles answer with zones no violation state
--!@param 5bytes - each bit represents one zone. 
function SatelResponseHandler:handleZonesNoViolationAnswer(data) 
    self:updateZonesProperty(data:sub(2), "noViolationTrouble")
end

--! handles answer with zones long violation state
--!@param 5bytes - each bit represents one zone. 
function SatelResponseHandler:handleZonesLongViolationAnswer(data) 
    self:updateZonesProperty(data:sub(2), "longViolationTrouble")
end


function SatelResponseHandler:handlePartitionAlarmAnswer(data) 
    self:updatePartitionProperty(data:sub(2), "alarm")
end

function SatelResponseHandler:handlePartitionFireAlarmAnswer(data) 
    self:updatePartitionProperty(data:sub(2), "fireAlarm")
end

function SatelResponseHandler:handlePartitionAlarmMemoryAnswer(data) 
    self:updatePartitionProperty(data:sub(2), "alarmMemory")
end

function SatelResponseHandler:handlePartitionFireAlarmMemoryAnswer(data) 
    self:updatePartitionProperty(data:sub(2), "fireAlarmMemory")
end

function SatelResponseHandler:handlePartitionEntryTimeAnswer(data) 
    self:updatePartitionProperty(data:sub(2), "entryTime")
end

function SatelResponseHandler:handlePartitionArmAnswer(data) 
    self:updatePartitionProperty(data:sub(2), "armed")
end

function SatelResponseHandler:handlePartitionExitTimeMoreThen10Answer(data) 
    self:updatePartitionProperty(data:sub(2), "exitTimeMoreThan10")
end

function SatelResponseHandler:handlePartitionExitTimeLessThen10Answer(data) 
    self:updatePartitionProperty(data:sub(2), "exitTimeLessThan10")
end

-- Updates zones property based on bitmask received from Satel
--@data 5 bytes - each bit represents one zone. Property for each zone is set to value of it's corresponding bit.
--@property name of property to be updated
function SatelResponseHandler:updateZonesProperty(data, property)
    local bitNo
    local byteNo
    local zoneNo = 1

    for byteNo=1,16,1 do
        for bitNo=1,8,1 do
            local bitValue = bit32.band(data:byte(byteNo), math.pow(2, bitNo-1))

            local value
            if (bitValue == 0) then 
                value=false 
            else 
                value=true
            end

            local zone = self.satelAlarm:getZoneById(zoneNo)
            if (zone ~= nil) then
                zone:updateProperty(property, value)
            end

            zoneNo = zoneNo+1
        end
    end 
end

-- Updates partitions property based on bitmask received from Satel
--@data 16 bytes - each bit represents one partition. Property for each partition is set to value of it's corresponding bit.
--@property name of property to be updated
function SatelResponseHandler:updatePartitionProperty(data, property)
    local bitNo
    local byteNo
    local partitionNo = 1
    for byteNo=1,4,1 do
        for bitNo=1,8,1 do
            local bitValue = bit32.band(data:byte(byteNo), math.pow(2, bitNo-1))
        
            local value
            if (bitValue == 0) then 
                value=false
            else 
                value=true
            end

            local partition = self.satelAlarm:getPartitionById(partitionNo)

            if (partition ~= nil) then
                partition:updateProperty(property, value)
            end

            partitionNo = partitionNo+1
        end
    end 
end

function SatelResponseHandler:handlePartitionOutputsAnswer(data)
    self:updateOutputProperty(data:sub(2), "value")
end

-- Updates outputs property based on bitmask received from Satel
--@data 16 bytes - each bit represents one output. Property for each output is set to value of it's corresponding bit.
--@property name of property to be updated
function SatelResponseHandler:updateOutputProperty(data, property)
    local bitNo
    local byteNo
    local outputNo = 1

    for byteNo=1,16,1 do
        for bitNo=1,8,1 do
            local bitValue = bit32.band(data:byte(byteNo), math.pow(2, bitNo-1))

            local value
            if (bitValue == 0) then 
                value=false
            else 
                value=true
            end
            
            local output = self.satelAlarm:getOutputById(outputNo)
            if (output ~= nil) then
                output:updateProperty(property, value)
            end

            outputNo = outputNo+1
        end
    end 
end

function SatelResponseHandler:handleReadDeviceNameAnswer(data)
    logUpdates(os.date() .. ">> UPDATE: " .. self.satelAlarm:getAsHexString(data))
    if (data:len()>=20) then
        local deviceType = data:byte(2)
        local deviceId = data:byte(3)
        local typeFunction = data:byte(4)
        local name = data:sub(5, 20)
        local i = 1
        while i<name:len() and name:byte(i)==0x00 do
            i = i+1
        end
        name = name:sub(i)

        name = characterEncoder:convertToUtf8(name)

        name = name:gsub("^(.-)%s*$", "%1")
        logUpdates("    " .. name .. "(" .. deviceType .. ")")

        if (deviceType == 0x00) then
            self.satelAlarm.partitions[deviceId]:updateProperty("satelName", name)
            self.satelAlarm.partitions[deviceId]:setName(name)
            self.satelAlarm.partitions[deviceId]:updateProperty("satelType", typeFunction)
            self.satelAlarm.partitions[deviceId]:updateProperty("disabled", false)
            self.satelAlarm.partitions[deviceId]:setVisible(true)
        elseif(deviceType == 0x05) then
            self.satelAlarm.zones[deviceId]:updateProperty("satelName", name)
            self.satelAlarm.zones[deviceId]:setName(name)
            self.satelAlarm.zones[deviceId]:updateProperty("satelType", typeFunction)
            self.satelAlarm.zones[deviceId]:updateProperty("disabled", false)
            self.satelAlarm.zones[deviceId]:setVisible(true)
            self.satelAlarm.zones[deviceId]:updateProperty("satelPartitionID", data:byte(21))
        elseif(deviceType == 0x04) then
            self.satelAlarm.outputs[deviceId]:updateProperty("satelName", name)
            self.satelAlarm.outputs[deviceId]:setName(name)
            self.satelAlarm.outputs[deviceId]:updateProperty("satelType", typeFunction)
            -- 0 means 'not used'
            if (typeFunction ~= 0) then
                self.satelAlarm.outputs[deviceId]:setVisible(true)
                self.satelAlarm.outputs[deviceId]:updateProperty("disabled", false)
            end
        end
    end
    logUpdates("\n")
end

function SatelResponseHandler:handleVersionAnswer(data)
    local model
    local softwareVersion

    if data:byte(2) == 0 then
        model = "INTEGRA 24";
    elseif data:byte(2) == 1 then
        model = "INTEGRA 32";
    elseif data:byte(2) == 2 then
        model = "INTEGRA 64";
    elseif data:byte(2) == 3 then
        model = "INTEGRA 128";
    elseif data:byte(2) == 4 then
        model = "INTEGRA 128-WRL SIM300";
    elseif data:byte(2) == 66 then
        model = "INTEGRA 64 PLUS";
    elseif data:byte(2) == 67 then
        model = "INTEGRA 128 PLUS";
    elseif data:byte(2) == 132 then
        model = "INTEGRA 128-WRL LEON";
    else
        model = "UNKNOWN";
    end

    softwareVersion = data:sub(3,3) .. "." .. data:sub(4,5) .. " " .. data:sub(6, 9) .. "-" .. data:sub(10,11) .. "-" .. data:sub(12,13)
    self.satelAlarm:updateProperty("model", model);
    self.satelAlarm:updateProperty("version", softwareVersion);
end

function SatelResponseHandler:handleResultAnswer(data)
    local message = self.resultMessages[data:byte(2)] or "unknown answer"
    logGeneral(os.date() .. ">> ANSWER: " .. message)
end
