require('common.device')
require('net.HTTPClient')
require('common.unitsConverter')
local xmlParser = require("common.xmlSimple").newParser()
local lustrous = require('common.lustrous')

class 'yrWeather' (Device)

function yrWeather:__init(id)
    Device.__init(self, id) -- Calls super class (Device) constructor.
    self.http = net.HTTPClient({ timeout = 10000 })  -- Creates an instance of HTTPClient class.
    self.lustrous = lustrous.init({})
    self.pollingCommandId = nil
    self.pollingForStatusTime = 5* 60 * 1000 -- 5 minutes
    self.pollingWaitOnErrorTime = 1000

    self.symbolsMap = {
        [1] = 32,    --Sun
        [2] = 30,    --LightCloud
        [3] = 28,    --PartlyCloud
        [4] = 26,    --Cloud
        [5] = 45,    --LightRainSun
        [6] = 37,    --LightRainThunderSun
        [7] = 45,    --SleetSun
        [8] = 42,    --SnowSun
        [9] = 10,    --LightRain
        [10] = 40,   --Rain
        [11] = 4,    --RainThunder
        [12] = 5,    --Sleet
        [13] = 16,   --Snow
        [14] = 15,   --SnowThunder
        [15] = 20,   --Fog
        [20] = 45,   --SleetSunThunder
        [21] = 43,   --SnowSunThunder
        [22] = 4,    --LightRainThunder
        [23] = 46,   --SleetThunder
        [24] = 37,   --DrizzleThunderSun
        [25] = 46,   --RainThunderSun
        [26] = 37,   --LightSleetThunderSun
        [27] = 38,   --HeavySleetThunderSun
        [28] = 42,   --LightSnowThunderSun
        [29] = 43,   --HeavySnowThunderSun
        [30] = 4,    --DrizzleThunder
        [31] = 4,    --LightSleetThunder
        [32] = 3,    --HeavySleetThunder
        [33] = 42,   --LightSnowThunder
        [34] = 43,   --HeavySnowThunder
        [40] = 9,    --DrizzleSun
        [41] = 10,   --RainSun
        [42] = 6,    --LightSleetSun
        [43] = 7,    --HeavySleetSun
        [44] = 14,   --LightSnowSun
        [45] = 15,   --HeavysnowSun
        [46] = 9,    --Drizzle
        [47] = 5,    --LightSleet
        [48] = 18,   --HeavySleet
        [49] = 14,   --LightSnow
        [50] = 15   --HeavySnow
    }


    self:getYrWeather()
end

function yrWeather:urlencode(str)
  if (str) then
    str = string.gsub (str, "\n", "\r\n")
    str = string.gsub (str, "([^%w %-%_%.%~])",
        function (c) return string.format ("%%%02X", string.byte(c)) end)
    str = string.gsub (str, " ", "+")
  end
  return str
end

function yrWeather:getYrWeather()
    local settingsLocation = api.get('/settings/location')
    local locationLatitude = settingsLocation.latitude
    local locationLongitude = settingsLocation.longitude

    if locationLatitude ~= nil and locationLongitude~= nil then
        local params = {
            url = 'http://api.met.no/weatherapi/locationforecast/1.9/?lat=' .. locationLatitude .. ';lon=' .. locationLongitude,
            callback =  function(response) self:onDataReceivedFromGetYrWeather(response) end,
        }
        self:httpRequest(params)
    end
end

function yrWeather:onDataReceivedFromGetYrWeather(response)
    local xmlData = {}
    local ok, xmlData = pcall(function() return xmlParser:ParseXmlText(response.data) end)

    if xmlData ~= nil and ok then
        pcall(function()
            local conditionCode = self:convertYRnoSymbolToYahooSymbol( tonumber(xmlData['weatherdata']['product']['time'][2]['location']['symbol']['@number']), self:getTimeOfTheDay())
            self:updateProperty('ConditionCode', conditionCode)
            self:updateProperty('ConditionCodeConverted', conditionCode)
            self:updateProperty('Humidity', tonumber(xmlData['weatherdata']['product']['time'][1]['location']['humidity']['@value']))
            self:updateProperty('Pressure', tonumber(xmlData['weatherdata']['product']['time'][1]['location']['pressure']['@value']))
            self:updateProperty('Temperature',  tonumber(xmlData['weatherdata']['product']['time'][1]['location']['temperature']['@value']))
            self:updateProperty('WeatherCondition', self:convertYahooWeatherCondition(conditionCode))
            self:updateProperty('Wind', tonumber(xmlData['weatherdata']['product']['time'][1]['location']['windSpeed']['@mps']) * 3.6)
        end)
    end

    self:setCurentWeatherLabel()
    if self.pollingCommandId then
        clearTimeout(self.pollingCommandId)
    end
    self.pollingCommandId = setTimeout(function() self:getYrWeather(); end, self.pollingForStatusTime)
end

function yrWeather:onErrorReceivedFromGetYrWeather()
    if self.pollingCommandIdOnErrorReceived then
        clearTimeout(self.pollingCommandIdOnErrorReceived)
    end
    self.pollingCommandIdOnErrorReceived = setTimeout(function() self:getYrWeather(); end, self.pollingWaitOnErrorTime)
end

function yrWeather:getTimezoneOffset(ts)
    local utcdate   = os.date("!*t", ts)
    local localdate = os.date("*t", ts)
    localdate.isdst = false -- this is the trick
    return os.difftime(os.time(localdate), os.time(utcdate))
end

function yrWeather:getTimeOfTheDay()
    local settingsLocation = api.get('/settings/location')
    local ts = os.time()
    local offset = self:getTimezoneOffset(ts)
    offset = offset / 3600 -- convert to hours
    local params = {
        lat =  settingsLocation.latitude,
        lon =  settingsLocation.longitude,
        offset =  offset,
        zenith = 90.83,
    }
    local  timeOfTheDay = lustrous.get_time(params)
    return timeOfTheDay
end


function yrWeather:httpRequest(params)
    self.http:request(params.url,{
        options = {
            method = params.method
        },
        success = function(response)
            if (params.callback) then
                params.callback(response)
            end
        end,
        error = function(error) self:onErrorReceivedFromGetYrWeather() end,
    })
end

function yrWeather:round(num, idp)
    idp = idp or 1
    if idp~=0 and num%1~=0 then
        local mult = 10^idp
        r= math.floor(num * mult + 0.5) / mult
        if r%1==0 then
            return math.floor(r)
        else
            return r
        end
    else
        return math.floor(tonumber(num) + 0.5)
    end
end

 -- Convert YR.no weather symbols to Yahoo symbols
 -- Documentatnion of YR.no weather icons http://api.met.no/weatherapi/weathericon/1.1/documentation
 -- Yahoo symbols http://somethingnew2.com/blog/images/conditionCode.pdf

function yrWeather:convertYRnoSymbolToYahooSymbol(symbol, timeOfTheDay)

    local symbol = self.symbolsMap[symbol]
    if symbol == nil then
        symbol=99
    end

    -- Change code if it is a night.
    -- YR.no API don't send different codes for this conditions at night
    if timeOfTheDay == 'night' then
        if symbol == 32 then symbol = 31        -- sun
        elseif symbol == 30 then symbol = 29    -- lightCloud
        elseif symbol == 28 then symbol = 27    -- partlyCloud
        end
    end

    return symbol
end

function yrWeather:convertYahooWeatherCondition(code)
    converted = ''

    if ( code == 6 or code == 8 or code == 9 or code == 10 or code == 11 or code == 12 or code == 17 or code == 18 or code == 35 or code == 40 ) then converted = 'rain';
    elseif ( code == 5 or code == 7 or code == 13 or code == 14 or code == 15 or code == 16 or code == 41 or code == 42 or code == 43 or code == 46 ) then converted = 'snow';
    elseif ( code == 0 or code == 1 or code == 2 or code == 3 or code == 4 or code == 37 or code == 38 or code == 39 or code == 45 or code == 47 ) then converted = 'storm';
    elseif ( code == 23 or code == 24 or code == 25 or code == 31 or code == 32 or code == 33 or code == 34 or code == 36 ) then converted = 'clear';
    elseif ( code == 26 or code == 27 or code == 28 or code == 29 or code == 30 or code == 44 ) then converted = 'cloudy';
    elseif ( code == 19 or code == 20 or code == 21 or code == 22 ) then converted = 'fog';
    end

    return converted
end

function yrWeather:setCurentWeatherLabel()
    local settingsLocation = api.get('/settings/location')
    local settingsInfo = api.get('/settings/info')
    local humidityValueLabel = tostring(self:round(self.properties.Humidity)) .. "%"
    local tempValueLabel = ""
    local windValueLabel = ""

    if settingsLocation.temperatureUnit == "F" then
        tempValueLabel = tostring(self:round(unitsConverter.celsius2farenheit(self.properties.Temperature))) .. '°F'
    else
        tempValueLabel = tostring(self:round(self.properties.Temperature)) .. '°C'
    end

    if settingsLocation.windUnit == "mph" then
        windValueLabel = tostring(self:round(unitsConverter.km2mile(self.properties.Wind))) .. 'mph'
    else
        windValueLabel = tostring(self:round(self.properties.Wind)) .. 'km/h'
    end
    
    if settingsInfo.defaultLanguage == 'pl' then
        self:updateProperty('ui.Current_Weather_Label.caption',  'Temp.: ' .. tempValueLabel .. '\n\nWilg.: '  .. humidityValueLabel .. '\n\nWiatr: '   .. windValueLabel)
    else
        self:updateProperty('ui.Current_Weather_Label.caption',  'Temp.: ' .. tempValueLabel .. '\n\nHumidity: '  .. humidityValueLabel .. '\n\nWind: '   .. windValueLabel)
    end
end
