require('actionHandlers.HTTPAction')
require('net.HTTPClient')
require('common.device')

class 'SchedulerOne' (HTTPAction)


function SchedulerOne:__init()
    HTTPAction.__init(self)

    self.hcNumber = self.settingsInfo.serialNumber
    self.defaultLanguage = self.settingsInfo.defaultLanguage
    self.weatherDeviceId = 3

    self.actions = {
        ['User tracking']   = function() self:handleUserTracking() end,
        ['Cooling panels']  = function() self:handleCoolingPanels() end,
        ['Humidity panels'] = function() self:handleHumidityPanels() end,
        ['SMS management']  = function() self:handleSMSManagement() end
    }
end

function SchedulerOne:getHomeDevices()
    local controlHeaders = {
        ['X-Fibaro-Version'] = '2',
    }

   local httpRequestParams = {
        url = 'http://127.0.0.1:11111/api/devices',
        method = 'GET',
        headers = controlHeaders,
        callback =  function(response) self:onDataReceivedFromGetHomeDevices(response) end,
    }

    self:httpRequest(httpRequestParams)
end

function SchedulerOne:onDataReceivedFromGetHomeDevices(response)
    local ok, devices = pcall(json.decode, response.data)

    if not ok then
        logger.error('Cannot get HC devices: ' .. (devices or '<empty>'))
    else
        self.devicesList = devices
        for functionName,func in pairs(self.actions) do
            logger.info('SchedulerOne: ' .. functionName)
            local ok, returnCode = pcall(function() func() end)
            if not ok then
                logger.error('Unexpected error in ' .. functionName .. ': ' .. (returnCode or 'nil'))
            end 
            logger.info('SchedulerOne: ' .. functionName .. ' finished')
        end
    end 
end

function SchedulerOne:execute()
    self:getHomeDevices()
end

function SchedulerOne:onDataReceivedFromExecute(response)
    print('****response***  onDataReceivedFromExecute:')
    tprint(response)
end

--[[
------------------------------- USERS LOCATION
--]]

function SchedulerOne:handleUserTracking()
    local users = api.get('/users')
    local anyTrackingUser = false
    for _,user in pairs(users) do
        if tonumber(user.tracking) ~= 0 then
            anyTrackingUser = true
            break
        end
    end

    if anyTrackingUser then
        self:getUsersLocationFromServer()
        self:sendLocationIntervals()
    end

end

function SchedulerOne:getUsersLocationFromServer()
    -- TODO: refactor to user request from super class
    self.http:request('https://ml.fibaro.com/get_ml_json.php?PK_AccessPoint=' .. self.hcNumber .. '&HW_Key=' .. self.hcKey, {
        success = function(params)
            if params.status == 200 then
                local ok, data = pcall(json.decode, params.data)
  
                if ok and data then
                    self:saveUsersLocation(data)
                else
                    logger.error('Unexpected error: can not decode JSON from ML service')
                end
                
            else
                logger.error('Unexpected error: can not read from ML service')
            end
        end,
        error = function()
            logger.error('Unexpected error: couldn\'t connect to ML service')
        end
    })
end

function SchedulerOne:saveUsersLocation(data)
    local users = api.get('/devices?type=HC_user')
    for _,location in pairs(data) do
        for _,hcUser in pairs(users) do
            if tonumber(hcUser.id) == tonumber(location.user_id) then
                local userDevice = Device(hcUser.id)
                userDevice:updateProperty('hasGPS', 1)
                userDevice:updateProperty('PreviousLatitude', tonumber(hcUser.properties.Latitude))
                userDevice:updateProperty('PreviousLongitude', tonumber(hcUser.properties.Longitude))
                userDevice:updateProperty('PreviousLocationTimestamp', tonumber(hcUser.properties.LocationTimestamp))
                userDevice:updateProperty('PreviousLocationTime', hcUser.properties.LocationTime)
                userDevice:updateProperty('PreviousLocation', hcUser.properties.Location)

                userDevice:updateProperty('Latitude', tonumber(location.latitude))
                userDevice:updateProperty('Longitude', tonumber(location.longitude))
                userDevice:updateProperty('LocationTimestamp', tonumber(location.timestamp))
                userDevice:updateProperty('LocationTime', location.date)
                userDevice:updateProperty('Location', location.latitude .. ';' .. location.longitude)

                local postBody = {
                    ["userId"] = hcUser.id,
                    ["longitude"] = tonumber(location.longitude),
                    ["latitude"] = tonumber(location.latitude),
                    ["timestamp"] = tonumber(location.timestamp),
                }
                api.post('/panels/family', postBody)    
            end
        end
    end
end

function SchedulerOne:getScheduledDataByUserId(userID, schedules)
    for _,scheduleData in pairs(schedules) do
        for key,element in pairs(scheduleData) do
            if key == 'id' and tonumber(element) == userID then
                return scheduleData
            end
        end
    end

    return nil 
end

function SchedulerOne:sendLocationIntervals()
    local schedules = api.get('/trackingSchedules')
    local users = api.get('/users')
    for _,user in pairs(users) do
        local interval = 0
        if (user.tracking ~= -1) then 
            self:sendLocationIntervalToServer(user.id, user.name, user.tracking)
        else
            local uName = user.name
            local scheduleData = self:getScheduledDataByUserId(user.id, schedules)
            if scheduleData then                
                local day = os.date('%A')
                local timeOfDay = {'Morning', 'Day', 'Evening', 'Night'}
                local hour = ''
                local hourTo = ''
                local time = ''
                local track = ''
                local locationIntervalFound = false

                for _,tod in pairs(timeOfDay) do
                    hour = scheduleData[day .. tod .. 'Hour']
                    hourTo = scheduleData[day .. tod .. 'HourTo']
                    times = scheduleData[day .. tod .. 'Time']
                    if tonumber(hour) < tonumber(hourTo) then
                        if tonumber(os.date('%H')) >= tonumber(hour) and tonumber(os.date('%H')) < tonumber(hourTo) then -- start tracking
                            self:sendLocationIntervalToServer(user.id, uName, times)
                            locationIntervalFound = true
                        end
                    elseif tonumber(hour) > tonumber(hourTo) then
                        if tonumber(os.date('%H')) >= tonumber(hour) or tonumber(os.date('%H')) < tonumber(hourTo) then -- start tracking
                            self:sendLocationIntervalToServer(user.id, uName, times)
                            locationIntervalFound = true
                        end
                    end
                end
                if not locationIntervalFound then
                    self:sendLocationIntervalToServer(user.id, uName, 0)
                end
            end
        end
    end
end

function SchedulerOne:sendLocationIntervalToServer(userId, userName, interval)
    local controlHeaders = {
            ['Content-Type'] = 'application/x-www-form-urlencoded',
    }
    local postData = 'PK_AccessPoint=' .. self.hcNumber .. '&HW_Key=' .. self.hcKey .. '&userID=' .. userId .. '&userName=' .. userName .. '&interval=' .. interval

    -- TODO: refactor to use request from super class
    self.http:request('https://ml.fibaro.com/saveMLUser.php', {
        options = {
            headers = controlHeaders,
            method = 'POST',
            data = postData,
        }, 
        success = function(params)
            if params.status == 200 then
                -- do nothing for now
            else
                logger.error('Unexpected error: Error saving user interval')
            end
        end,
        error = function()
            logger.error('Unexpected error: couldn\'t connect to ML service 2')
        end
    })
end

function SchedulerOne:getCoolingZones()
    local result = {}    
    local zones = api.get('/panels/cooling?detailed=true')

    for _,zone in pairs(zones) do
        result[zone.id] = zone
    end

    return result
end

function SchedulerOne:handleCoolingPanels()
    local coolingZones = self:getCoolingZones()
    local ts = os.time()

    if self:tablelength(coolingZones) > 0 then
        for hzID,hz in pairs(coolingZones) do
            local day = os.date('%A')
            local prefix = string.lower(day)            
            local hzDay = hz.properties[string.lower(day)]            
            local temperature = 0
            local handTimestamp = tonumber(hz.properties.handTimestamp)

            if handTimestamp ~= 0 and ts <= handTimestamp then 
                local handTemperature = tonumber(hz.properties.handTemperature)
                if handTemperature > 0 then
                    temperature = handTemperature
                end
            end 

            if temperature == 0 then
                local vacationTemperature = tonumber(hz.properties.vacationTemperature)
                if vacationTemperature > 0 then
                    temperature = vacationTemperature
                end
            end

            if temperature == 0 then
                local partOfDay = ''  

                local tsmorning = os.time({ year = os.date('%Y'), month = os.date('%m'), day = os.date('%d'), hour = hzDay.morning.hour, min = hzDay.morning.minute, sec = os.date('%S')})    
                local tsday = os.time({ year = os.date('%Y'), month = os.date('%m'), day = os.date('%d'), hour = hzDay.day.hour, min = hzDay.day.minute, sec = os.date('%S')}) 
                if (tsday <= tsmorning) then 
                    tsday = tsday + 86400
                end

                local tsevening = os.time({ year = os.date('%Y'), month = os.date('%m'), day = os.date('%d'), hour = hzDay.evening.hour, min = hzDay.evening.minute, sec = os.date('%S')}) 
                if tsevening <= tsday then
                    tsevening = tsevening + 86400
                end

                local tsnight = os.time({ year = os.date('%Y'), month = os.date('%m'), day = os.date('%d'), hour = hzDay.night.hour, min = hzDay.night.minute, sec = os.date('%S')}) 
                if tsnight <= tsevening then
                    tsnight = tsnight + 86400
                end
                            
                if ts < tsmorning then
                    hzDay = hz.properties[prefix]
                    partOfDay = hzDay.night
                elseif ts >= tsmorning and ts < tsday then
                    partOfDay = hzDay.morning            
                elseif ts >= tsday and ts < tsevening then
                    partOfDay = hzDay.day
                elseif ts >= tsevening and ts < tsnight then
                    partOfDay = hzDay.evening
                else 
                    partOfDay = hzDay.night
                end
                if partOfDay ~= '' then
                    temperature = partOfDay.temperature
                end
            end

            if temperature > 0  then
                if self:tablelength(hz.properties.rooms) > 0 then
                    for _,roomID in pairs(hz.properties.rooms) do
                        local devicesFromHome = self.devicesList
                        local conns = api.get('/linkedDevices')
                        if self:tablelength(conns) > 0 then
                            for k,connData in pairs(conns) do
                                if connData.type == 'cooling' and connData.roomID == roomID and connData.deviceID > 0 then
                                    local device = self:getDeviceById(devicesFromHome, connData.deviceID)
                                    local level = device.properties.value
                                    if level then
                                        if level > temperature then
                                            if self:tablelength(connData.devices) > 0 then
                                                for key,connDevice in pairs(connData.devices) do
                                                    local connDevID = connDevice.id
                                                    local device = self:getDeviceById(devicesFromHome, connDevID)                            
                                                    local connLevel = device.properties.value
                                                    if connLevel == false or (tonumber(connLevel) or 1) == 0 then                                                 
                                                        api.get("/callAction?deviceID=" .. device.id .. "&name=" .. 'turnOn')
                                                    end
                                                end
                                            end
                                        elseif level < temperature then
                                            if self:tablelength(connData.devices) > 0 then
                                                for key,connDevice in pairs(connData.devices) do
                                                    local connDevID = connDevice.id
                                                    local device = self:getDeviceById(devicesFromHome, connDevID)                                                                                
                                                    local connLevel = device.properties.value
                                                    if connLevel == true or (tonumber(connLevel) or 0) > 0 then  
                                                        api.get("/callAction?deviceID=" .. device.id .. "&name=" .. 'turnOff')
                                                    end

                                                end
                                            end
                                        end
                                    end
                                end
                            end
                        end
                    end
                end
            end
        end
    else
        return
    end
end

function SchedulerOne:getHumidityZones()
    local result = {}    
    local zones = api.get('/panels/humidity?detailed=true')

    for _,zone in pairs(zones) do
        result[zone.id] = zone
    end

    return result
end

function SchedulerOne:handleHumidityPanels()
    local humidityZones = self:getHumidityZones()
    local ts = os.time()

    if self:tablelength(humidityZones) > 0 then
        for hzID,hz in pairs(humidityZones) do
            local day = os.date('%A')
            local prefix = string.lower(day)            
            local hzDay = hz.properties[string.lower(day)]            
            local humidity = 0
            local handTimestamp = tonumber(hz.properties.handTimestamp)

            if handTimestamp ~= 0 and ts <= handTimestamp then 
                local handHumidity = tonumber(hz.properties.handHumidity)
                if handHumidity > 0 then
                    humidity = handHumidity
                end
            end 

            if humidity == 0 then
                local vacationHumidity = tonumber(hz.properties.vacationHumidity)
                if vacationHumidity > 0 then
                    humidity = vacationHumidity
                end
            end

            if humidity == 0 then
                local partOfDay = ''  

                local tsmorning = os.time({ year = os.date('%Y'), month = os.date('%m'), day = os.date('%d'), hour = hzDay.morning.hour, min = hzDay.morning.minute, sec = os.date('%S')})    
                local tsday = os.time({ year = os.date('%Y'), month = os.date('%m'), day = os.date('%d'), hour = hzDay.day.hour, min = hzDay.day.minute, sec = os.date('%S')}) 
                if (tsday <= tsmorning) then 
                    tsday = tsday + 86400
                end

                local tsevening = os.time({ year = os.date('%Y'), month = os.date('%m'), day = os.date('%d'), hour = hzDay.evening.hour, min = hzDay.evening.minute, sec = os.date('%S')}) 
                if tsevening <= tsday then
                    tsevening = tsevening + 86400
                end

                local tsnight = os.time({ year = os.date('%Y'), month = os.date('%m'), day = os.date('%d'), hour = hzDay.night.hour, min = hzDay.night.minute, sec = os.date('%S')}) 
                if tsnight <= tsevening then
                    tsnight = tsnight + 86400
                end
                            
                if ts < tsmorning then
                    hzDay = hz.properties[prefix]
                    partOfDay = hzDay.night
                elseif ts >= tsmorning and ts < tsday then
                    partOfDay = hzDay.morning            
                elseif ts >= tsday and ts < tsevening then
                    partOfDay = hzDay.day
                elseif ts >= tsevening and ts < tsnight then
                    partOfDay = hzDay.evening
                else 
                    partOfDay = hzDay.night
                end
                if partOfDay ~= '' then
                    humidity = partOfDay.humidity
                end
            end

            if humidity > 0  then
                if self:tablelength(hz.properties.rooms) > 0 then
                    for _,roomID in pairs(hz.properties.rooms) do
                        local devicesFromHome = self.devicesList
                        local conns = api.get('/linkedDevices')
                        if self:tablelength(conns) > 0 then
                            for k,connData in pairs(conns) do
                                if connData.type == 'humidity' and connData.roomID == roomID and connData.deviceID > 0 then
                                    local device = self:getDeviceById(devicesFromHome, connData.deviceID)
                                    local level = device.properties.value
                                    if level then
                                        if level > humidity then
                                            if self:tablelength(connData.devices) > 0 then
                                                for key,connDevice in pairs(connData.devices) do
                                                    local connDevID = connDevice.id
                                                    local device = self:getDeviceById(devicesFromHome, connDevID)                            
                                                    local connLevel = device.properties.value
                                                    if connLevel == true or (tonumber(connLevel) or 0) > 0 then  
                                                        api.get("/callAction?deviceID=" .. device.id .. "&name=" .. 'turnOff')
                                                    end
                                                end
                                            end
                                        elseif level < humidity then
                                            if self:tablelength(connData.devices) > 0 then
                                                for key,connDevice in pairs(connData.devices) do
                                                    local connDevID = connDevice.id
                                                    local device = self:getDeviceById(devicesFromHome, connDevID)                                                                                
                                                    local connLevel = device.properties.value
                                                    if connLevel == false or (tonumber(connLevel) or 1) == 0 then                                                 
                                                        api.get("/callAction?deviceID=" .. device.id .. "&name=" .. 'turnOn')
                                                    end

                                                end
                                            end
                                        end
                                    end
                                end
                            end
                        end
                    end
                end
            end
        end
    else
        return
    end
end

function SchedulerOne:getDeviceById(devices, id)
    for _,device in pairs(devices) do
        if device.id == id then 
            return device
        end
    end

    return nil 
end

--[[
------------------------------- SMS MANAGEMENT
--]]

function SchedulerOne:handleSMSManagement()
    if self.settingsInfo.smsManagement then
        -- self:sendRoomsListToServer()
        self:getSMSCommands()
    end
end

function SchedulerOne:sendRoomsListToServer()
    local rooms = api.get('/rooms');
    local roomsToSend = {}
    for _,room in pairs(rooms) do
        local r = {
            ['id'] = room.id,
            ['name'] = room.name
        }
        table.insert(roomsToSend, r)
    end
    local jsonToSend = json.encode(roomsToSend)

    local httpRequestParams = {
        url = 'https://sms.fibaro.com/saveRooms.php?PK_AccessPoint=' .. self.hcNumber .. '&HW_Key=' .. self.hcKey,
        method = 'POST',
        data = jsonToSend,
        callback =  function(response) self:test(response) end
    }

    self:httpRequest(httpRequestParams)

end

function SchedulerOne:getSMSCommands()
    local httpRequestParams = {
        url = 'https://sms.fibaro.com/getCommands.php?PK_AccessPoint=' .. self.hcNumber .. '&HW_Key=' .. self.hcKey,
        method = 'GET',
        callback =  function(response) self:executeSMSCommands(response) end
    }

    self:httpRequest(httpRequestParams)
end

function SchedulerOne:inTable(value, table)
    -- TODO: move it to common tools
    if type(table) == 'table' then
        for _,val in pairs(table) do
            if (val == value) then
                return true
            end
        end
    end

    return false
end

function SchedulerOne:executeSMSCommands(response)
    if (response.status == 200) then
        local ok, data = pcall(json.decode, response.data)

        if ok and data then
            for _,command in pairs(data) do
                local jobId = command.job_id;
                if command.command_type == 'swiatlo' then
                    if command.command_value == 'on' or command.command_value == 'off' then
                        local devices = api.get('/devices?roomId=' .. command.room_id)
                        for _,device in pairs(devices) do
                            if (device.type == 'com.fibaro.binarySwitch' or device.baseType == 'com.fibaro.binarySwitch' or device.baseType == 'com.fibaro.multilevelSwitch')
                            and self:inTable('light', device.interfaces) then
                                if command.command_value == 'on' and (device.properties.value == false or (tonumber(device.properties.value) or 1) == 0) then
                                    api.get("/callAction?deviceID=" .. device.id .. "&name=" .. 'turnOn')
                                elseif command.command_value == 'off' and (device.properties.value == true or (tonumber(device.properties.value) or 0) > 0) then
                                    api.get("/callAction?deviceID=" .. device.id .. "&name=" .. 'turnOff')
                                end
                            end
                        end
                    elseif command.command_value == 'stan' then
                        local response = {
                            ['phone_number'] = command.phone_number,
                            ['type'] = 'lights',
                            ['status'] = {}
                        }
                        local devices = self.devicesList
                        for _,device in pairs(devices) do
                            if (device.type == 'com.fibaro.binarySwitch' or device.baseType == 'com.fibaro.binarySwitch' or device.baseType == 'com.fibaro.multilevelSwitch')
                            and self:inTable('light', device.interfaces) then
                                local s = {}
                                if device.properties.value == false or (tonumber(device.properties.value) or 1) == 0 then
                                    local s = {
                                        ['device_id'] = device.id,
                                        ['room_id'] = device.roomID,
                                        ['state'] = false
                                    }
                                    table.insert(response.status, s)
                                elseif device.properties.value == true or (tonumber(device.properties.value) or 0) > 0 then
                                    local s = {
                                        ['device_id'] = device.id,
                                        ['room_id'] = device.roomID,
                                        ['state'] = true
                                    }
                                    table.insert(response.status, s)
                                end
                            end
                        end
                        self:sendHouseStatus(response)
                    end
                elseif command.command_type == 'rolety' then
                    if command.command_value == 'otworz' or command.command_value == 'zamknij' then
                        local devices = api.get('/devices?roomId=' .. command.room_id)
                        for _,device in pairs(devices) do
                            if (device.type == 'com.fibaro.rollerShutter' or device.baseType == 'com.fibaro.rollerShutter' or device.type == 'com.fibaro.FGRM222') then
                                if command.command_value == 'otworz' and tonumber(device.properties.value) == 0 then
                                    api.get("/callAction?deviceID=" .. device.id .. "&name=" .. 'open')
                                elseif command.command_value == 'zamknij' and tonumber(device.properties.value) > 0 then
                                    api.get("/callAction?deviceID=" .. device.id .. "&name=" .. 'close')
                                end
                            end
                        end
                    elseif command.command_value == 'stan' then
                        local response = {
                            ['phone_number'] = command.phone_number,
                            ['type'] = 'blinds',
                            ['status'] = {}
                        }
                        local devices = self.devicesList
                        for _,device in pairs(devices) do
                            if (device.type == 'com.fibaro.rollerShutter' or device.baseType == 'com.fibaro.rollerShutter' or device.type == 'com.fibaro.FGRM222') then
                                local s = {}
                                if tonumber(device.properties.value) == 0 then
                                    local s = {
                                        ['device_id'] = device.id,
                                        ['room_id'] = device.roomID,
                                        ['state'] = false
                                    }
                                    table.insert(response.status, s)
                                elseif tonumber(device.properties.value) > 0 then
                                    local s = {
                                        ['device_id'] = device.id,
                                        ['room_id'] = device.roomID,
                                        ['state'] = true
                                    }
                                    table.insert(response.status, s)
                                end
                            end
                        end
                        self:sendHouseStatus(response)
                    end
                end
                self:sendSMSJobStatus(jobId)
            end
        else
            logger.error('Unexpected error: Error decoding JSON from SMS server')
        end
    else
        logger.error('Unexpected error: Error reading from SMS server')
    end
end

function SchedulerOne:sendHouseStatus(data)
    local jsonToSend = json.encode(data)

    local httpRequestParams = {
        url = 'https://sms3.fibaro.com/sendHouseStatus.php?PK_AccessPoint=' .. self.hcNumber .. '&HW_Key=' .. self.hcKey,
        method = 'POST',
        data = jsonToSend,
        callback =  function(response) self:test(response) end
    }

    self:httpRequest(httpRequestParams)
end

function SchedulerOne:sendSMSJobStatus(jobId)
    local httpRequestParams = {
        url = 'https://sms.fibaro.com/setSMSJobDone.php?PK_AccessPoint=' .. self.hcNumber .. '&HW_Key=' .. self.hcKey .. '&job_id=' .. jobId,
        method = 'GET',
        callback =  function(response) self:test(response) end
    }

    self:httpRequest(httpRequestParams)
end

function SchedulerOne:test(response)
    if type(response) == 'table' then
        tprint(response)
    else
        print(response)
    end
end



