require('actionHandlers.HTTPAction')
require('actionHandlers.SendEmailAction')
require('common.device')

local lustrous = require('common.lustrous')

class 'SchedulerThirty' (HTTPAction)


function SchedulerThirty:__init()
    HTTPAction.__init(self)

    self.hcNumber = self.settingsInfo.serialNumber
    self.defaultLanguage = self.settingsInfo.defaultLanguage
    self.lustrous = lustrous.init({})
    self.settingsLocation = api.get('/settings/location')
    self.masterController = Device(1)
end

function SchedulerThirty:getHomeDevices()
    local controlHeaders = {
        ['X-Fibaro-Version'] = '2',
    }

   local httpRequestParams = {
        url = 'http://127.0.0.1:11111/api/devices',
        method = 'GET',
        headers = controlHeaders,
        callback =  function(response) self:onDataReceivedFromGetHomeDevices(response) end,
    }

    self:httpRequest(httpRequestParams)
end

function SchedulerThirty:onDataReceivedFromGetHomeDevices(response)
    local ok, devices = pcall(json.decode, response.data)

    if not ok then
        logger.error('Cannot get HC devices: ' .. (devices or '<empty>'))
    else
        self.devicesList = devices
        self:setSunHours()
        self:sendBatteryLowNotification()
    end 
end

function SchedulerThirty:execute()    
    self:getHomeDevices()
end

function SchedulerThirty:onDataReceivedFromExecute(response)
    print('****response***  onDataReceivedFromExecute:')
    tprint(response)
end

--[[
------------------------------- SUNRISE / SUNSET HOURS
--]]

function SchedulerThirty:getTimezoneOffset(ts)
    local utcdate   = os.date("!*t", ts)
    local localdate = os.date("*t", ts)
    localdate.isdst = false -- this is the trick
    return os.difftime(os.time(localdate), os.time(utcdate))
end

function SchedulerThirty:getSunHours()
    local ts = os.time()
    local offset = self:getTimezoneOffset(ts)
    offset = offset / 3600 -- convert to hours

    local params = {
        lat =  self.settingsLocation.latitude,
        lon =  self.settingsLocation.longitude,
        offset =  offset,
        zenith = 90.83,
    }

    local  timeOfTheDay, sunrise, sunset = lustrous.get_time(params)
    
    if (type(sunrise) == 'number') then
        sunrise = os.date("%H:%M", sunrise)
    end

    if (type(sunset) == 'number') then
        sunset = os.date("%H:%M", sunset)
    end

    return sunrise, sunset
end

function SchedulerThirty:setSunHours()
    local sunriseHour, sunsetHour = self:getSunHours()

    self.masterController:updateProperty('sunriseHour', sunriseHour)
    self.masterController:updateProperty('sunsetHour', sunsetHour)
end

--[[
------------------------------- BATTERY LOW NOTIFICATIONS
--]]

function SchedulerThirty:sendBatteryLowNotification()
    local hcDevices = self.devicesList
    for _,dev in pairs(hcDevices) do
        if type(dev.properties.batteryLowNotification) == 'boolean' and dev.properties.batteryLowNotification 
            and type(dev.properties.batteryLevel) == 'number' and dev.properties.batteryLevel == 255
            and type(dev.visible) == 'boolean' and dev.visible then
            
            local subject = self:getBatteryLowSubjectTranslation()
            local message = self:getBatteryLowMessageTranslation(dev.name, self:getRoomNameById(dev.roomID))

            local emailParams = {
                    ['args'] = { subject, message},
                    ['actionName'] = 'Email...'   ,
                }

            local sendEmailAction = SendEmailAction(2, emailParams)
            sendEmailAction:execute()
        end
    end 
end

function SchedulerThirty:getRoomNameById(id)
    local rooms = api.get('/rooms')
    for _,room in pairs(rooms) do
        if tonumber(room.id) == tonumber(id) then
            return room.name
        end
    end

    return self:getUnassignedTranslation()
end

function SchedulerThirty:getUnassignedTranslation()
    local transTable = {
        ['br'] = 'Sem correspondência',
        ['de'] = 'nicht zugeordnet',
        ['en'] = 'Unassigned',
        ['et'] = 'Määramata',
        ['fr'] = 'Non assigné',
        ['it'] = 'Non Assegnato',
        ['lv'] = 'Nepiesķirts',
        ['nl'] = 'Niet toegewezen',
        ['pl'] = 'Nieprzypisane',
        ['pt'] = 'Sem correspondencia',
        ['ro'] = 'Nealocat',
        ['sv'] = 'Ej tilldelad',
        ['cn'] = '未分配',
        ['ru'] = 'Нераспределенные',
        ['fi'] = 'Määrittämätön'
    }

    return transTable[self.defaultLanguage] or transTable['en']
end

function SchedulerThirty:getBatteryLowSubjectTranslation()
    local transTable = {
        ['en'] = 'Fibaro System - Notification of the battery status',
        ['it'] = 'Fibaro System - Notificazione dello stato di batteria',
        ['lv'] = 'Fibaro System - Brīdinājums par bateriju stāvokli',
        ['nl'] = 'Batterij status melding',
        ['pl'] = 'Fibaro System - Powiadomienie o stanie baterii'
    }

    return transTable[self.defaultLanguage] or transTable['en']
end

function SchedulerThirty:getBatteryLowMessageTranslation(deviceName, roomName)
    local transTable = {
        ['en'] = 'Warning, device ' .. deviceName .. ' in room ' .. roomName .. ' has a low battery. Replace or charge the battery in this device. If you do not want to receive notifications of this type, disable them in the configuration.',
        ['it'] = 'Attenzione, il dispositivo ' .. deviceName .. ' in room ' .. roomName .. ' ha una batteria di livello basso. E\' fortemente consigliato di sostituire le batterie nel dispositivo. Se non vuoi ricevere informazioni di questo tipo puoi disabilitare la funzione nella configurazione.',
        ['lv'] = 'Brīdināt, ja istabā ' .. roomName .. ' kādai ierīcei ' .. deviceName .. ' ir tukša baterija. Rekomendējam ierīcei nomainīt baterijas. Ja Jūs nevēlaties saņemt paziņojumus, tad šo iespēju var atslēgt šajā konfigurācijā.',
        ['nl'] = 'Waarschuwing, het apparaat ' .. deviceName .. ' in de kamer ' .. roomName .. ' heeft een lage batterij. Het wordt aanbevolen om de batterijen in dit apparaat te vervangen. Indien u geen meldingen meer wilt ontvangen van dit type, dan kan dit worden uitgeschakeld bij de configuratie.',
        ['pl'] = 'Uwaga, urządzenie ' .. deviceName .. ' w pomieszczeniu ' .. roomName .. ' ma niski stan baterii.  Wymień lub naładuj baterię w tym urządzeniu. Jeżeli nie chcesz otrzymywać tego typu powiadomień, wyłącz je w konfiguracji urządzenia.'
    }

    return transTable[self.defaultLanguage] or transTable['en']
end



