require('net.WebSocketClient') 
require('net.WebSocketClientTls') 
require('common/SyncQueue')
require('common.Object')


class 'JsonRpcOverWebsocket' (Object)

function JsonRpcOverWebsocket:__init(options)
    Object.__init(self)

    options = options or {}
    self.callTimeout = options.callTimeout or 5000
    self.syncQueue = SyncQueue() 
    self.rpcCallId = 0

    if (options.secure) then
        self.sock = net.WebSocketClientTls()
    else
        self.sock = net.WebSocketClient()
    end

    self.sock:registerEvent("connected", function() self:emitEvent('connected') end)
    self.sock:registerEvent("disconnected", function() self:emitEvent('disconnected') end)
    self.sock:registerEvent("error", function(err) self:emitEvent('connectionError', err) end)
    self.sock:registerEvent("dataReceived", function(data) self:onDataReceived(data) end)
end

function JsonRpcOverWebsocket:sendJsonRpcCall(method, params, callbacks)
  self:generateRpcCallId()

  local JSONRequestArray = {
    ["jsonrpc"] = "2.0",
    ["method"] = method,
    ["params"] = params,
    id = self.rpcCallId,
  }

  local ok, jsonRequest = pcall(json.encode, JSONRequestArray)
  
  if ok and jsonRequest then
    local rpcRequestTable = {[JSONRequestArray.id] = callbacks}
    self.syncQueue:pushBack(rpcRequestTable)
    self.sock:send(jsonRequest)
    setTimeout(function() self:applyTimeout(JSONRequestArray.id) end, self.callTimeout) 
  else 
    error("invalid params from sendJsonRpcCall", 2)
  end
end

function JsonRpcOverWebsocket:applyTimeout(id)
  local rpcRequestTable = self.syncQueue:popCommand(id)
    self:callErrorCallbackFunction(rpcRequestTable, '{"jsonrpc": "2.0", "error": "Call Timeout"}, "id": '..id..'}')
end

function JsonRpcOverWebsocket:generateRpcCallId()
  if self.rpcCallId > 100000000 then 
    self.rpcCallId = 0 
  else 
    self.rpcCallId = self.rpcCallId + 1  
  end
end

function JsonRpcOverWebsocket:onDataReceived(data)
    local ok, dataTable = pcall(json.decode, data)

    if ok and dataTable then -- data is valid json object
      if type(dataTable.id) == 'number' and dataTable.id then -- id is not null so there should be related rpcCallRequest in self.asyncQueue and id exist so this is not a notification
        local rpcRequestTable = self.syncQueue:popCommand(dataTable.id)
        if rpcRequestTable then -- if there is related id in asyncQueue
          if dataTable.error then -- error occurred
            self:callSuccessCallbackFunction(rpcRequestTable, {["error"] = dataTable.error})
          elseif dataTable.result then 
            self:callSuccessCallbackFunction(rpcRequestTable, {["result"] = dataTable.result})
          else
            print("Silent Error: no result or error in response. Message was:", data)
          end
        else
          print("Silent Error: no related id in command queue - it might be a timeout issue")
        end
      else -- there is no id so it might be notification or some kind of error (there should not be any related rpcCallRequest in self.asyncQueue)       
        if dataTable.error then -- error occurred
            print("Undefined Error:", dataTable.error)
        else -- this is a notification
          self:handleNotification(data)
        end
      end
    end
end

function JsonRpcOverWebsocket:callSuccessCallbackFunction(rpcRequestTable, data)
  if type(rpcRequestTable) == "table" then
    for _,callbacks in pairs(rpcRequestTable) do
        callbacks.success(data)
    end
  end
end

function JsonRpcOverWebsocket:callErrorCallbackFunction(rpcRequestTable, data)
  if type(rpcRequestTable) == "table" then
    for _,callbacks in pairs(rpcRequestTable) do
      callbacks.error(data)
    end
  end
end

function JsonRpcOverWebsocket:handleNotification(data)
  self:emitEvent('notificationFromServer', data)
end

function JsonRpcOverWebsocket:connect(uri)
  local ok = pcall (function() self.sock:connect(uri) end)
  if not ok then  error("invalid uri", 2)
 end
end 

function JsonRpcOverWebsocket:close()
  local ok = pcall (function() self.sock:close() end)
  if not ok then  error("JsonRpcOverWebsocket:close", 2)
 end
end



