
class 'SyncQueue'

function SyncQueue:__init()
    self.queue = {}
end

--! Adds command to back of the queue
--!@param command command string to be added
function SyncQueue:pushBack(command)
    table.insert(self.queue, command)
end

--! Adds command at the front of the queue (rest of commands are shifted)
--!@param command command string to be added
function SyncQueue:pushFront(command)
    table.insert(self.queue, 1, command)
end

--! Gets element from queue front. Element is removed from queue.
--!@return command string
function SyncQueue:popFront()
	if #self.queue > 0 then
    	return table.remove(self.queue, 1)
    else
    	return nil
    end
end

--! Gets element from queue front. 
--!@return command string
function SyncQueue:getFrontElement()
	if #self.queue > 0 then
    	return self.queue[1]
    else
    	return nil
    end
end

--! Gets element from queue back. Element is removed from queue.
--!@return command string
function SyncQueue:popBack()
	if #self.queue > 0 then
    	return table.remove(self.queue)
    else
    	return nil
    end
end

--! Searches for given command in queue and returns index (or 0 if not found) of first occurence of this command in queue.
--! Search would be performed by matching commandToFind with actual commands from queue (commands may be longer - matches only from begining of string)
--!@param commandToFind string with command
--!@return if found index of first occurence of command, else 0
function SyncQueue:findCommand(commandToFind)
	for index,table in pairs(self.queue) do 
		for identifier,commandFunction in pairs(table) do
			if (identifier == commandToFind) then
				return index
			end
		end
	end

	return 0
end

--! Gets command from queue.
--!@param commandToGet.
--!@return command from queue.
function SyncQueue:getCommand(command)
	return self.queue[self:findCommand(command)] 
end

--! Pops command from queue.
--!@param commandToPop.
--!@return command from queue.
function SyncQueue:popCommand(command)
	local commandToReturn =  self.queue[self:findCommand(command)] 

	if commandToReturn then
		self:removeCommand(command)
		return commandToReturn
	else
		return nil
	end
end

--! Removes all command occuriences from queue. Compares commandToRemove with start of commands strings
--!@param commandToRemove start of command string that will be searched for removal
function SyncQueue:removeCommand(commandToRemove)
	local commandIndex = self:findCommand(commandToRemove)
	while commandIndex>0 do
		table.remove(self.queue, commandIndex)
		commandIndex = self:findCommand(commandToRemove)
	end
end


--! Adds command at queue front, and removes all other occurences of this command from queue. 
--! E.g. pushFrontAndRemoveOthers('command(123)', 'command') adds 'command(123)' and removes everything starting with 'command' (e.g. 'command(a)', 'command(111)', 'command()', ...)
--!@param commandToSend command to be added
--!@param commandToFind command (start of command string) to find and remove from queue
function SyncQueue:pushFrontAndRemoveOthers(commandToSend, commandToFind)
	self:removeCommand(commandToFind)
	self:pushFront(commandToSend)
end

--! Clears queue. It becomes empty.
function SyncQueue:clear()
	self.queue = {}
end

--! Returns queue table.
--@return queue table
function SyncQueue:getQueueTable()
	if #self.queue > 0 then
    	return self.queue
    else
    	return nil
    end
end

 