require('common.Object')
require('net.TCPSocket')

--! Handles asynchronous read of data from TCP socket. Provides interaction with events:
--!  - error : called when error occures
--!  - connected : when socket is succesfully connected
--!  - disconnected : when socked is disconnected
--!  - dataReceived : when same data were succesfully read from socket
--! Events can be registered with registerEvent(eventName, handler) method
--! Upon succesfull connection it starts to listen on socket
class 'TCPListener' (Object)

--! Constructor
--! @param options array(key => value) with options. Currently followingg options are handled:
--!   - delimiter: data received will be sliced by delimiter. If not provided dataReceived event would be rised after every succesfull read
--!   - connectTimeout
--!   - readTimeout
function TCPListener:__init(options)
    Object.__init(self)

    options = options or {}

    self.delimiter = options.delimiter or ''
    self.connectTimeout = options.connectTimeout or 0
    self.readTimeout = options.readTimeout or 0

    self.sock = net.TCPSocket()
end

--! Tries to connect on the tcp socket. Starts reading upon success.
--! @param address ip adress
--! @param port
function TCPListener:connect(address, port)
    self.sock:connect(address, port, {
        success = function()
            self:emitEvent('connected')
            self:__readLoop()
        end,
        error = function(err)
            self:emitEvent('error', err)
            self:emitEvent('disconnected')
        end,
        timeout=self.connectTimeout
    })
end

--! Closes tcp connection
function TCPListener:close()
    self.sock:close()
end

--! Sends data to socket
--! @param data string with data to be sent
function TCPListener:send(data)
    self.sock:write(data, {
        error = function(err)
            self:emitEvent('error', err)
            self:emitEvent('disconnected')
        end
    })
end

function TCPListener:__success(data)
    self:emitEvent('dataReceived', data)
    self:__readLoop()
end

function TCPListener:__error(err)
    self:emitEvent('error', err)

    if (self.sock:isOpen()) then
        self:__readLoop()
    else
        self:emitEvent('disconnected')
    end
end


function TCPListener:__readLoop()
    if (self.delimiter == '') then
        self.sock:read({
            success = function(data) self:__success(data) end,
            error = function(err) self:__error(err) end
        })
    else
        self.sock:readUntil(self.delimiter, {
            success = function(data) self:__success(data) end,
            error = function(err) self:__error(err) end
        })
    end
end
