require('net.HTTPClient')

local cosmo = require("cosmo.cosmo")

class 'UPnP'

UPnP.ENVELOPE = '<?xml version="1.0" encoding="utf-8"?><s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/"><s:Body>$body</s:Body></s:Envelope>'

UPnP.PLAY_ACTION = '"urn:schemas-upnp-org:service:AVTransport:1#Play"'
UPnP.PLAY_BODY = '<u:Play xmlns:u="urn:schemas-upnp-org:service:AVTransport:1"><InstanceID>0</InstanceID><Speed>1</Speed></u:Play>'

UPnP.PAUSE_ACTION = '"urn:schemas-upnp-org:service:AVTransport:1#Pause"'
UPnP.PAUSE_BODY = '<u:Pause xmlns:u="urn:schemas-upnp-org:service:AVTransport:1"><InstanceID>0</InstanceID><Speed>1</Speed></u:Pause>'

UPnP.STOP_ACTION = '"urn:schemas-upnp-org:service:AVTransport:1#Stop"'
UPnP.STOP_BODY = '<u:Stop xmlns:u="urn:schemas-upnp-org:service:AVTransport:1"><InstanceID>0</InstanceID><Speed>1</Speed></u:Stop>'

UPnP.NEXT_ACTION = '"urn:schemas-upnp-org:service:AVTransport:1#Next"'
UPnP.NEXT_BODY = '<u:Next xmlns:u="urn:schemas-upnp-org:service:AVTransport:1"><InstanceID>0</InstanceID><Speed>1</Speed></u:Next>'

UPnP.PREV_ACTION = '"urn:schemas-upnp-org:service:AVTransport:1#Previous"'
UPnP.PREV_BODY = '<u:Previous xmlns:u="urn:schemas-upnp-org:service:AVTransport:1"><InstanceID>0</InstanceID><Speed>1</Speed></u:Previous>'

UPnP.SET_MUTE_ACTION = '"urn:schemas-upnp-org:service:RenderingControl:1#SetMute"'
UPnP.SET_MUTE_BODY_TEMPLATE = '<u:SetMute xmlns:u="urn:schemas-upnp-org:service:RenderingControl:1"><InstanceID>0</InstanceID><Channel>Master</Channel><DesiredMute>$mute</DesiredMute></u:SetMute>'

UPnP.GET_MUTE_ACTION = '"urn:schemas-upnp-org:service:RenderingControl:1#GetMute"'
UPnP.GET_MUTE_BODY = '<u:GetMute xmlns:u="urn:schemas-upnp-org:service:RenderingControl:1"><InstanceID>0</InstanceID><Channel>Master</Channel></u:GetMute>'

UPnP.SET_VOLUME_ACTION = '"urn:schemas-upnp-org:service:RenderingControl:1#SetVolume"'
UPnP.SET_VOLUME_BODY_TEMPLATE = '<u:SetVolume xmlns:u="urn:schemas-upnp-org:service:RenderingControl:1"><InstanceID>0</InstanceID><Channel>Master</Channel><DesiredVolume>$volume</DesiredVolume></u:SetVolume>'

UPnP.GET_VOLUME_ACTION = '"urn:schemas-upnp-org:service:RenderingControl:1#GetVolume"'
UPnP.GET_VOLUME_BODY = '<u:GetVolume xmlns:u="urn:schemas-upnp-org:service:RenderingControl:1"><InstanceID>0</InstanceID><Channel>Master</Channel></u:GetVolume>'

UPnP.GET_TRANSPORT_INFO_ACTION = '"urn:schemas-upnp-org:service:AVTransport:1#GetTransportInfo"'
UPnP.GET_TRANSPORT_INFO_BODY = '<u:GetTransportInfo xmlns:u="urn:schemas-upnp-org:service:AVTransport:1"><InstanceID>0</InstanceID></u:GetTransportInfo>'

UPnP.GET_POSITION_INFO_ACTION = '"urn:schemas-upnp-org:service:AVTransport:1#GetPositionInfo"'
UPnP.GET_POSITION_INFO_BODY = '<u:GetPositionInfo xmlns:u="urn:schemas-upnp-org:service:AVTransport:1"><InstanceID>0</InstanceID></u:GetPositionInfo>'

UPnP.SET_AV_TRANSPORT_ACTION = '"urn:schemas-upnp-org:service:AVTransport:1#SetAVTransportURI"'
UPnP.SET_AV_TRANSPORT_TEMPLATE = [[<u:SetAVTransportURI xmlns:u="urn:schemas-upnp-org:service:AVTransport:1">
         <InstanceID>0</InstanceID>
         <CurrentURI>$uri</CurrentURI>
         <CurrentURIMetaData></CurrentURIMetaData>
      </u:SetAVTransportURI>]]

UPnP.SET_MAIN_TV_SOURCE_ACTION = '"urn:samsung.com:service:MainTVAgent2:1#SetMainTVSource"'
UPnP.SET_MAIN_TV_SOURCE_TEMPLATE = [[<u:SetMainTVSource xmlns:u="urn:samsung.com:service:MainTVAgent2:1">
         <Source>$source</Source>
         <ID>$id</ID>
         <UiID>0</UiID>
      </u:SetMainTVSource>]]



function UPnP:__init(config)
	self.ip = config.ip
	self.port = config.port

	self.AVTRANSPORT_URI = config.AVTransport
	self.RENDERING_CONTROL_URI = config.RenderingControl
	if config.MainTVAgent ~= nil
	then
		self.MAIN_TV_AGENT_URI = config.MainTVAgent
	end 

	self.http = net.HTTPClient({ timeout  = config.timeout or 3000 })
	self.xmlParser = require("common.xmlSimple").newParser()
end

function UPnP:play()
	self:__sendRequest(self.AVTRANSPORT_URI, UPnP.PLAY_ACTION, UPnP.PLAY_BODY)
end

function UPnP:pause()
	self:__sendRequest(self.AVTRANSPORT_URI, UPnP.PAUSE_ACTION, UPnP.PAUSE_BODY)
end

function UPnP:stop()
	self:__sendRequest(self.AVTRANSPORT_URI, UPnP.STOP_ACTION, UPnP.STOP_BODY)
end

function UPnP:next()
	self:__sendRequest(self.AVTRANSPORT_URI, UPnP.NEXT_ACTION, UPnP.NEXT_BODY)
end

function UPnP:prev()
	self:__sendRequest(self.AVTRANSPORT_URI, UPnP.PREV_ACTION, UPnP.PREV_BODY)
end

function UPnP:setMute(mute)
	local body = cosmo.fill(UPnP.SET_MUTE_BODY_TEMPLATE, { mute = mute })
	self:__sendRequest(self.RENDERING_CONTROL_URI, UPnP.SET_MUTE_ACTION, body)
end

function UPnP:getMute(callback, errorCallback)
	local success = function(data)
		local ok = pcall(function()
			local xmlData = self.xmlParser:ParseXmlText(data.data)
			local mute = xmlData['s:Envelope']['s:Body']['u:GetMuteResponse']['CurrentMute']:value()
			
			if (mute == '1') then
				callback(true)
			else
				callback(false)
			end
		end)

		if (not ok) then
			errorCallback('xml parse error')
		end
	end

	self:__sendRequest(self.RENDERING_CONTROL_URI, UPnP.GET_MUTE_ACTION, UPnP.GET_MUTE_BODY, success, errorCallback)
end	

function UPnP:setVolume(volume)
	local body = cosmo.fill(UPnP.SET_VOLUME_BODY_TEMPLATE, { volume = volume })
	self:__sendRequest(self.RENDERING_CONTROL_URI, UPnP.SET_VOLUME_ACTION, body)
end

function UPnP:getVolume(callback, errorCallback)
	local success = function(data)
		local ok = pcall(function()
			local xmlData = self.xmlParser:ParseXmlText(data.data)
			local volume = xmlData['s:Envelope']['s:Body']['u:GetVolumeResponse']['CurrentVolume']:value()

			callback(tonumber(volume))
		end)

		if (not ok) then
			errorCallback('xml parse error')
		end
	end

	self:__sendRequest(self.RENDERING_CONTROL_URI, UPnP.GET_VOLUME_ACTION, UPnP.GET_VOLUME_BODY, success, errorCallback)
end

function UPnP:getTransportInfo(callback, errorCallback)
	local success = function(data)
		local ok = pcall(function()
			local xmlData = self.xmlParser:ParseXmlText(data.data)
			local state = xmlData['s:Envelope']['s:Body']['u:GetTransportInfoResponse']['CurrentTransportState']:value()

			callback(state)
		end)

		if (not ok) then
			errorCallback('xml parse error')
		end
	end

	self:__sendRequest(self.AVTRANSPORT_URI, UPnP.GET_TRANSPORT_INFO_ACTION, UPnP.GET_TRANSPORT_INFO_BODY, success, errorCallback)
end

function UPnP:getPositionInfo(callback, errorCallback)
	local success = function(data)
		local ok = pcall(function()
			local info = {
				relTime = '0:00:00',
				totalTime = '0:00:00',
				title = '',
				creator = '',
				albumArtURI = ''
			}

			local xmlData = self.xmlParser:ParseXmlText(data.data)

			local metadata = xmlData['s:Envelope']['s:Body']['u:GetPositionInfoResponse']['TrackMetaData']:value()
			local xmlMetadata = self.xmlParser:ParseXmlText(metadata)
			
			info = {
				relTime = xmlData['s:Envelope']['s:Body']['u:GetPositionInfoResponse']['RelTime']:value(),
				totalTime = xmlData['s:Envelope']['s:Body']['u:GetPositionInfoResponse']['TrackDuration']:value()
			}

			if (xmlMetadata['DIDL']['item']['dc:title']) then
				info.title = xmlMetadata['DIDL']['item']['dc:title']:value() or ''
			end

			if (xmlMetadata['DIDL']['item']['dc:creator']) then
				info.creator = xmlMetadata['DIDL']['item']['dc:creator']:value() or ''
			end

			if (xmlMetadata['DIDL']['item']['upnp:albumArtURI']) then
				info.albumArtURI = xmlMetadata['DIDL']['item']['upnp:albumArtURI']:value() or ''
			end

			callback(info)
		end)

		if (not ok) then
			errorCallback('xml parse error')
		end		
	end

	self:__sendRequest(self.AVTRANSPORT_URI, UPnP.GET_POSITION_INFO_ACTION, UPnP.GET_POSITION_INFO_BODY, success, errorCallback)
end

function UPnP:setAVTransportUri(uri)
	local body = cosmo.fill(UPnP.SET_AV_TRANSPORT_TEMPLATE, { uri = uri })
	self:__sendRequest(self.AVTRANSPORT_URI, UPnP.SET_AV_TRANSPORT_ACTION, body)
end

--samsung TV function
function UPnP:setMainTVSource(source, id)
	local body = cosmo.fill(UPnP.SET_MAIN_TV_SOURCE_TEMPLATE, { source = source, id = id})
	self:__sendRequest(self.MAIN_TV_AGENT_URI, UPnP.SET_MAIN_TV_SOURCE_ACTION, body)
end

function UPnP:__sendRequest(uri, action, body, successCallback, errorCallback)
    data = cosmo.fill(UPnP.ENVELOPE, { body = body })

	self.http:request('http://' .. self.ip .. ':' .. tostring(self.port) .. uri, {
		options = {		
        	headers = {
                ['Content-Type'] = 'text/xml',
                ['SOAPACTION'] = action
        	},
        	data = cosmo.fill(UPnP.ENVELOPE, { body = body }),
        	method = 'POST'
        },
        success = successCallback,
		error = errorCallback             
	})
end